/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *  
 *  $Id: library-thread.c,v 1.9 2002/04/01 14:49:22 jbaayen Exp $
 */

#include <fileinfo.h>
#include <libgnomevfs/gnome-vfs-uri.h>
#include <libgnomevfs/gnome-vfs-directory.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-ops.h>
#include <glib.h>

#include "library.h"
#include "library-private.h"

static void set_node_to_fileinfo (Node *song, FileInfo *info);
static void update_song (Library *l, Node *song);

/**
 * get_mtime: return mtime for @node
 */
static time_t
get_mtime (Node *node)
{
	GnomeVFSFileInfo *info = gnome_vfs_file_info_new ();
	time_t ret;
	gnome_vfs_get_file_info (node_get_string_property (node, "uri"),
				 info, GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
	ret = info->mtime;
	gnome_vfs_file_info_unref (info);
	return ret;
}

/**
 * Main loop for the updating thread
 */
gpointer
library_thread_main (gpointer data)
{
	Library *l = LIBRARY (data);
	GList *node;

	while (TRUE)
	{
		//printf ("Thread: waiting on lock run_thread\n");
		g_mutex_lock (l->priv->mutex->run_thread);
		g_mutex_unlock (l->priv->mutex->run_thread);
		//printf ("Thread: through lock run_thread\n");

		//printf ("Thread: waiting on lock songs_to_update\n");
		g_mutex_lock (l->priv->mutex->songs_to_update); 
		//printf ("Thread: through lock songs_to_update\n");
		if (l->priv->songs_to_update == NULL) {
			//printf ("Thread: no items, locking run_thread\n");
			g_mutex_lock (l->priv->mutex->run_thread);
			g_mutex_unlock (l->priv->mutex->songs_to_update);
			continue;
		}
		//printf ("Thread: examining node, queue length is %d\n", g_list_length (l->priv->songs_to_update));
		node = l->priv->songs_to_update;
		l->priv->songs_to_update = g_list_remove_link (l->priv->songs_to_update, l->priv->songs_to_update);
		g_mutex_unlock (l->priv->mutex->songs_to_update);
		update_song (l, node->data);
		//printf ("Thread: Done updating song\n");
		g_list_free (node);
	}
}

static void
update_song (Library *l, Node *song)
{
	Node *album = NULL, *artist = NULL;
	FileInfo *info;
	int id;
	const char *uri;

	uri = node_get_string_property (song, SONG_PROPERTY_URI);

	g_assert (uri != NULL);

	/* get fileinfo */
	info = fileinfo_get (uri, TRUE);

	if (info == NULL)
	{
		/* invalid mimetype, do not add */
		return;
	}

	if (info->artist == NULL)
		info->artist = g_strdup ("");

	if (info->album == NULL)
		info->album = g_strdup ("");

	artist = library_private_add_artist_if_needed (l, info->artist);
		
	album = library_private_add_album_if_needed (l, info->album, artist);

	id = library_private_build_id (l);
	node_set_int_property (song, "id", id);
	
	library_private_remove_song (l, song);
	set_node_to_fileinfo (song, info->polished);
	library_private_add_song (l, song, artist, album);
	
	fileinfo_free (info);

	library_private_append_node_signal (l, song, NODE_CREATED);
}

/**
 * set_node_to_fileinfo: set @song to @info
 */
static void
set_node_to_fileinfo (Node *song, FileInfo *info)
{
	if (info->title)     node_set_string_property (song, "name",      g_strdup (info->title));
	if (info->date)      node_set_string_property (song, "date",      g_strdup (info->date));
	if (info->genre)     node_set_string_property (song, "genre",     g_strdup (info->genre));
	if (info->comment)   node_set_string_property (song, "comment",   g_strdup (info->comment));
	if (info->codecinfo) node_set_string_property (song, "codecinfo", g_strdup (info->codecinfo));
	if (info->tracknum)  node_set_string_property (song, "tracknum",  g_strdup (info->tracknum));
	node_set_int_property (song, "bitrate",  info->bitrate);
	node_set_int_property (song, "filesize", info->filesize);
	node_set_int_property (song, "length",   info->length);
	node_set_int_property (song, "mtime",    get_mtime (song));
}
