/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: library-ui.c,v 1.101 2002/04/13 21:54:29 kristian Exp $
 */

#include <gtk/gtktreemodel.h>
#include <libgnome/gnome-i18n.h>
#include <string.h>

#include "library-ui.h"
#include "misc.h"
#include "rbtreemodelfilter.h"
#include "queue.h"
#include "node-utils.h"
#include "treeview-utils.h"
#include "songs-view.h"
#include "state.h"
#include "icons.h"
#include "state-manager.h"
#include "eel-gconf-extensions.h"

#define CONF_SHOW_BROWSER "/apps/rhythmbox/LibraryUI/show_browser"

/* columns of the lists */
typedef enum
{
	ALBUM_COL_NAME     = 0,
	ALBUM_COL_NODE     = 1,
	ALBUM_COL_VISIBLE  = 2,
	ALBUM_COL_PRIORITY = 3,
	ALBUM_COL_COUNT    = 4
} AlbumRow;

typedef enum
{
	ARTIST_COL_NAME     = 0,
	ARTIST_COL_NODE     = 1,
	ARTIST_COL_PRIORITY = 2,
	ARTIST_COL_COUNT    = 3
} ArtistRow;

struct _LibraryUIPrivate
{
	SourceManager *manager;   /* the associaated SourceManager */

	Library *library;         /* the associated Library */

	Queue *queue;             /* the associated Queue */

	StateManager *sm;         /* the associated StateManager */

	UI *ui;                   /* the associated UI */

	GtkWidget *paned;         /* the paned between the artist/album browser and the
				     songs */

	GtkWidget *hbox;          /* contains the browsing treeviews (artists, albums) */

	/* artistview-related stuff */
	GtkWidget *artists;
	GtkListStore *artists_liststore;
	GHashTable *artists_node_to_iter;

	/* albumsview-related stuff */
	GtkWidget *albums;
	GtkListStore *albums_liststore;
	GtkTreeModel *albums_filtermodel;
	GtkTreeModel *albums_sortmodel;
	GHashTable *albums_node_to_iter;
	
	SongsView *songs;                /* the list of songs */

	Node *selected_artist;           /* the currently selected artist node */
	Node *selected_album;            /* the currently selected album node */

	gint id;                         /* our source ID */

	gboolean active;                 /* are we the selected source ? */

	gboolean querying;		 /* arew e running a search ? */

	GList *action_widgets;           /* list of actionbuttons */
};

/* all the different signals */
enum
{
	START_PLAYING,
	STOP_PLAYING,
	LAST_SIGNAL
};

/* GObject properties */
enum
{
	PROP_0,
	PROP_SOURCE_MANAGER
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint library_ui_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void library_ui_class_init (LibraryUIClass *klass);
static void library_ui_init (LibraryUI *l);
static void library_ui_finalize (GObject *object);
static void library_ui_set_property (GObject *object,
			             guint prop_id,
			             const GValue *value,
			             GParamSpec *pspec);
static void library_ui_get_property (GObject *object,
			             guint prop_id,
			             GValue *value,
			             GParamSpec *pspec);

/* signal prototypes */
static void node_created_cb (Library *l, Node *node, LibraryUI *lui);
static void node_changed_cb (Library *l, Node *node, LibraryUI *lui);
static void node_deleted_cb (Library *l, Node *node, LibraryUI *lui);
static void artists_selection_changed_cb (GtkTreeSelection *selection, LibraryUI *l);
static void albums_selection_changed_cb (GtkTreeSelection *selection, LibraryUI *l);
static void songs_row_activated_cb (SongsView *v, Node *node, LibraryUI *l);
static void queue_request_next_cb (Queue *q, LibraryUI *l);
static void ui_source_selected_cb (UI *ui, gint id, LibraryUI *l);
static void ui_copy_cb (UI *ui, LibraryUI *l);
static void ui_rewind_cb (UI *ui, LibraryUI *l);
static void ui_forward_cb (UI *ui, LibraryUI *l);
static void songs_view_rewind_available_cb (SongsView *v, gboolean avail,
						     LibraryUI *l);
static void songs_view_forward_available_cb (SongsView *v, gboolean avail,
						     LibraryUI *l);
static void songs_view_selection_available_cb (SongsView *v, gboolean avail, LibraryUI *l);
static void browser_cb (GtkWidget *widget, LibraryUI *l);
static void ui_play_cb (UI *ui, LibraryUI *l);
static void ui_stop_cb (UI *ui, LibraryUI *l);
static void ui_query_cb (UI *ui, gchar *text, LibraryUI *l);
static void ui_pause_cb (UI *ui, LibraryUI *l);
static void songs_row_remove_request_cb (SongsView *v, GList *nodes, LibraryUI *l);
static void ui_set_shuffle_cb (UI *ui, gboolean shuffle, LibraryUI *l);
static void ui_set_repeat_cb (UI *ui, gboolean repeat, LibraryUI *l);
static void ui_clear_cb (UI *ui, LibraryUI *l);
static void ui_select_all_cb (UI *ui, LibraryUI *l);
static void ui_select_none_cb (UI *ui, LibraryUI *l);
static void ui_cut_cb (UI *ui, LibraryUI *l);
static void source_manager_stop_playing_cb (SourceManager *m, gpointer exception, LibraryUI *l);
static void state_manager_start_playing_cb (StateManager *m, LibraryUI *l);
static void state_manager_stop_playing_cb (StateManager *m, LibraryUI *l);

/* local functions */
static void add_artist (LibraryUI *l, Node *node);
static void remove_artist (LibraryUI *l, Node *node);
static void add_album (LibraryUI *l, Node *node);
static void remove_album (LibraryUI *l, Node *node);
static void select_albums_all (LibraryUI *l);
static void select_artists_all (LibraryUI *l);
static void check_album_all_count (LibraryUI *l);
static void check_artist_all_count (LibraryUI *l);
static void update_statusbar (LibraryUI *l);
static void remove_nodes (LibraryUI *l, GList *sel);
static void library_ui_show_browser (LibraryUI *l, gboolean show);
static void library_ui_set_manager (LibraryUI *lui, SourceManager *m);

/**
 * library_ui_get_type: get the GObject type of the LibraryUI
 */
GType
library_ui_get_type (void)
{
	static GType library_ui_type = 0;

  	if (library_ui_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (LibraryUIClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) library_ui_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (LibraryUI),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) library_ui_init
      		};

      		library_ui_type = g_type_register_static (G_TYPE_OBJECT,
                				          "LibraryUI",
                                           	          &our_info, 0);
    	}

	return library_ui_type;
}

/**
 * library_ui_class_init: initialize the LibraryUI class
 */
static void
library_ui_class_init (LibraryUIClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = library_ui_finalize;

	object_class->set_property = library_ui_set_property;
	object_class->get_property = library_ui_get_property;

	/* init properties */
	g_object_class_install_property (object_class,
					 PROP_SOURCE_MANAGER,
					 g_param_spec_object ("source-manager",
						 	      "SourceManager",
							      "Associated SourceManager object",
							      source_manager_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

	/* init signals */
	library_ui_signals[START_PLAYING] =
   		g_signal_new ("start_playing",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (LibraryUIClass, start_playing),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	library_ui_signals[STOP_PLAYING] =
   		g_signal_new ("stop_playing",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (LibraryUIClass, stop_playing),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
}

/**
 * library_ui_init: intialize the LibraryUI object
 */
static void
library_ui_init (LibraryUI *l)
{
	GtkWidget *sw;
	GtkTreeView *current;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;

	l->priv = g_new0 (LibraryUIPrivate, 1);

	/* create the paned */
	l->priv->paned = gtk_vpaned_new ();
	l->priv->hbox = gtk_hbox_new (TRUE, 5);
	gtk_container_set_border_width (GTK_CONTAINER (l->priv->hbox), 5);
	gtk_paned_add1 (GTK_PANED (l->priv->paned), l->priv->hbox);

	/* artists view */
	sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
					     GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
					GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	l->priv->artists = gtk_tree_view_new (); 
	gtk_container_add (GTK_CONTAINER (sw), l->priv->artists);
	gtk_box_pack_start (GTK_BOX (l->priv->hbox), sw, TRUE, TRUE, 0);
	current = GTK_TREE_VIEW (l->priv->artists);
	l->priv->artists_liststore = gtk_list_store_new (ARTIST_COL_COUNT,
						         G_TYPE_STRING,
						         G_TYPE_POINTER,
							 G_TYPE_INT);
	gtk_tree_view_set_model (current, GTK_TREE_MODEL (l->priv->artists_liststore));
	gtk_tree_view_set_rules_hint (current, FALSE);
	gtk_tree_view_set_headers_visible (current, TRUE);

	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Artist",
						           renderer,
						           "text", ARTIST_COL_NAME,
						           NULL);
	gtk_tree_view_insert_column (current, column, ARTIST_COL_NAME);
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (l->priv->artists_liststore),
					 ARTIST_COL_NAME,
					 priority_sort_func,
					 GINT_TO_POINTER (ARTIST_COL_PRIORITY),
					 NULL);
	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (l->priv->artists_liststore),
					      ARTIST_COL_NAME,
					      GTK_SORT_ASCENDING);
	gtk_tree_view_column_set_clickable (column, FALSE);
	selection = gtk_tree_view_get_selection (current);
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
	g_signal_connect (G_OBJECT (selection), "changed",
			  G_CALLBACK (artists_selection_changed_cb), l);

	l->priv->artists_node_to_iter = g_hash_table_new (NULL, NULL);
	
	/* albums view */
	sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
					     GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
					GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	l->priv->albums = gtk_tree_view_new ();
	gtk_container_add (GTK_CONTAINER (sw), l->priv->albums);
	gtk_box_pack_start (GTK_BOX (l->priv->hbox), sw, TRUE, TRUE, 0);
	current = GTK_TREE_VIEW (l->priv->albums);
	l->priv->albums_liststore = gtk_list_store_new (ALBUM_COL_COUNT,
						    G_TYPE_STRING,
						    G_TYPE_POINTER,
						    G_TYPE_BOOLEAN,
						    G_TYPE_INT);
	l->priv->albums_filtermodel = rb_tree_model_filter_new_with_model 
		(GTK_TREE_MODEL (l->priv->albums_liststore), ALBUM_COL_VISIBLE, NULL);
	l->priv->albums_sortmodel = gtk_tree_model_sort_new_with_model
		(l->priv->albums_filtermodel);
	gtk_tree_view_set_model (current, l->priv->albums_sortmodel);
	gtk_tree_view_set_rules_hint (current, FALSE);
	gtk_tree_view_set_headers_visible (current, TRUE);

	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Album",
						            renderer,
						            "text", ALBUM_COL_NAME,
						            NULL);
	gtk_tree_view_insert_column (current, column, ALBUM_COL_NAME);
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (l->priv->albums_sortmodel),
					 ALBUM_COL_NAME,
					 priority_sort_func,
					 GINT_TO_POINTER (ALBUM_COL_PRIORITY),
					 NULL);
	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (l->priv->albums_sortmodel),
					      ALBUM_COL_NAME,
					      GTK_SORT_ASCENDING);
	gtk_tree_view_column_set_clickable (column, FALSE);
	selection = gtk_tree_view_get_selection (current);
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
	g_signal_connect (G_OBJECT (selection), "changed",
			  G_CALLBACK (albums_selection_changed_cb), l);

	l->priv->albums_node_to_iter = g_hash_table_new (NULL, NULL);

	/* songs view */
	l->priv->songs = SONGS_VIEW (songs_view_new ());
	gtk_paned_add2 (GTK_PANED (l->priv->paned), GTK_WIDGET (l->priv->songs));
	g_signal_connect (G_OBJECT (l->priv->songs), "row_remove_request",
			  G_CALLBACK (songs_row_remove_request_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "row_activated",
			  G_CALLBACK (songs_row_activated_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "rewind_available",
			  G_CALLBACK (songs_view_rewind_available_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "forward_available",
			  G_CALLBACK (songs_view_forward_available_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "selection_available",
			  G_CALLBACK (songs_view_selection_available_cb), l);

	/* show most bits */
	gtk_widget_show_all (l->priv->hbox);
	gtk_widget_show (GTK_WIDGET (l->priv->songs));

	/* nullify */
	l->priv->active = FALSE;
	l->priv->querying = FALSE;
	l->priv->selected_artist = NULL;
	l->priv->selected_album = NULL;
	l->priv->ui = NULL;
}

/**
 * library_ui_finalize: finalize the LibraryUI object
 */
static void
library_ui_finalize (GObject *object)
{
	LibraryUI *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_LIBRARY_UI (object));
	
   	s = LIBRARY_UI (object);

	g_return_if_fail (s->priv != NULL);

	/* disconnect signals */
	g_signal_handlers_disconnect_matched (G_OBJECT (s->priv->ui),
					      G_SIGNAL_MATCH_DATA,
					      0,
					      0,
					      NULL,
					      NULL,
					      object);
	g_signal_handlers_disconnect_matched (G_OBJECT (s->priv->library),
					      G_SIGNAL_MATCH_DATA,
					      0,
					      0,
					      NULL,
					      NULL,
					      object);
	g_signal_handlers_disconnect_matched (G_OBJECT (s->priv->queue),
					      G_SIGNAL_MATCH_DATA,
					      0,
					      0,
					      NULL,
					      NULL,
					      object);

	/* save states */
	state_save_pane_pos (s->priv->paned, "library_paned");

	/* cleanup */
	free_node_to_iter_hash (s->priv->albums_node_to_iter);
	free_node_to_iter_hash (s->priv->artists_node_to_iter);

	/* get rid of songsview */
	gtk_widget_destroy (GTK_WIDGET (s->priv->songs));

	/* get rid of actionbuttons */
	g_list_foreach (s->priv->action_widgets, (GFunc) gtk_widget_destroy, NULL);
	g_list_free (s->priv->action_widgets);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * library_ui_new: create a new LibraryUI object
 */
LibraryUI *
library_ui_new (SourceManager *manager)
{
 	LibraryUI *l;

	l = LIBRARY_UI (g_object_new (TYPE_LIBRARY_UI,
				      "source-manager", manager,
				      NULL));

	g_return_val_if_fail (l->priv != NULL, NULL);

	return l;
}

/**
 * library_ui_set_manager: set the SourceManager for this LibraryUI
 */
static void
library_ui_set_manager (LibraryUI *lui, SourceManager *m)
{
	GdkPixbuf *library_pixbuf;
	GtkWidget *button;
	GdkPixbuf *icon;

	lui->priv->manager = m;
	lui->priv->queue   = source_manager_get_queue (m);
	lui->priv->library = source_manager_get_library (m);
	lui->priv->ui      = source_manager_get_ui (m);

	/* connect manager signals */
	g_signal_connect (G_OBJECT (lui->priv->manager),
			  "stop_playing", G_CALLBACK (source_manager_stop_playing_cb), lui);

	/* connect signal to feed new songs to the queue when it wants them */
	g_signal_connect (G_OBJECT (lui->priv->queue),
			  "request_next", G_CALLBACK (queue_request_next_cb), lui);

	/* hook up to library */
	g_signal_connect (G_OBJECT (lui->priv->library), "node_created",
			  G_CALLBACK (node_created_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->library), "node_changed",
			  G_CALLBACK (node_changed_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->library), "node_deleted",
			  G_CALLBACK (node_deleted_cb), lui);

	/* add to the list of sources */
	lui->priv->id = ui_append_source (lui->priv->ui, lui->priv->paned);

	/* load the pixbuf used as the source icon */
	library_pixbuf = gtk_widget_render_icon (lui->priv->ui->notebook, RHYTHMBOX_STOCK_LIBRARY,
						 GTK_ICON_SIZE_MENU, NULL);

	/* set source properties */
	ui_source_set_name (lui->priv->ui, lui->priv->id, _("Library"));
	ui_source_set_icon (lui->priv->ui, lui->priv->id, library_pixbuf);
	ui_source_set_editable (lui->priv->ui, lui->priv->id, FALSE);	
	ui_source_set_priority (lui->priv->ui, lui->priv->id, 1);

	g_object_unref (G_OBJECT (library_pixbuf));

	/* connect ui signals */
	g_signal_connect (G_OBJECT (lui->priv->ui), "source_selected",
			  G_CALLBACK (ui_source_selected_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "cut",
			  G_CALLBACK (ui_cut_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "copy",
			  G_CALLBACK (ui_copy_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "clear",
			  G_CALLBACK (ui_clear_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "select_all",
			  G_CALLBACK (ui_select_all_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "select_none",
			  G_CALLBACK (ui_select_none_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "rewind",
			  G_CALLBACK (ui_rewind_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "forward",
			  G_CALLBACK (ui_forward_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "play",
			  G_CALLBACK (ui_play_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "stop",
			  G_CALLBACK (ui_stop_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "query",
			  G_CALLBACK (ui_query_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "pause",
			  G_CALLBACK (ui_pause_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "set_shuffle",
                          G_CALLBACK (ui_set_shuffle_cb), lui);
        g_signal_connect (G_OBJECT (lui->priv->ui), "set_repeat",
                          G_CALLBACK (ui_set_repeat_cb), lui);

	/* add state manager */
	lui->priv->sm = state_manager_new (lui->priv->queue, lui->priv->ui, lui->priv->songs, lui->priv->id);

	g_signal_connect (G_OBJECT (lui->priv->sm), "start_playing",
			  G_CALLBACK (state_manager_start_playing_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->sm), "stop_playing",
			  G_CALLBACK (state_manager_stop_playing_cb), lui);

	/* create "All" rows */
	add_artist (lui, library_get_all_albums (lui->priv->library));
	add_album (lui, library_get_all_songs (lui->priv->library));

	/* select "All" nodes */
	select_artists_all (lui);

	/* load state */
	state_load_pane_pos (lui->priv->paned, "library_paned");
	songs_view_set_shuffle (lui->priv->songs, eel_gconf_get_boolean (CONF_SHUFFLE));
	songs_view_set_repeat (lui->priv->songs, eel_gconf_get_boolean (CONF_REPEAT));
	library_ui_show_browser (lui, eel_gconf_get_boolean (CONF_SHOW_BROWSER));

	/* create actionbuttons */
	icon = gtk_widget_render_icon (GTK_WIDGET (lui->priv->songs),
				       RHYTHMBOX_STOCK_ACTION_LIBRARY,
				       RHYTHMBOX_ICON_SIZE_ACTION, NULL);
	button = ui_new_action_button (lui->priv->ui, _("Browse"),
				       _("Toggle browser visibility"),
				       icon, TRUE);
	g_object_unref (G_OBJECT (icon));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
				      eel_gconf_get_boolean (CONF_SHOW_BROWSER));
	g_signal_connect (G_OBJECT (button), "toggled",
			  G_CALLBACK (browser_cb), lui);

	lui->priv->action_widgets = NULL;
	lui->priv->action_widgets = g_list_append (lui->priv->action_widgets, button);
	
	/* show */
	gtk_widget_show (lui->priv->paned);

	/* notify object */
	g_object_notify (G_OBJECT (lui), "source-manager");
}

/**
 * library_ui_get_source_id: return the source id of the library
 */
int
library_ui_get_source_id (LibraryUI *l)
{
	return l->priv->id;
}

/**
 * node_created_cb: a node in the library was created, update the ui
 */
static void
node_created_cb (Library *l, Node *node, LibraryUI *lui)
{
	LibraryNodeType type;

	type = node_get_int_property (node, "type");	

	switch (type)
	{
	case LIBRARY_NODE_SONG:
		/* add the node to the songsview */
		songs_view_add_node (lui->priv->songs, node);
		
		/* update ui settings */
		check_album_all_count (lui);
		update_statusbar (lui);
		break;
	case LIBRARY_NODE_ARTIST:
		add_artist (lui, node);
		break;
	case LIBRARY_NODE_ALBUM:
		add_album (lui, node);
		break;
	default:
		break;
	}
}

/**
 * node_changed_cb: a node in the library was changed, update the ui
 */
static void
node_changed_cb (Library *l, Node *node, LibraryUI *lui)
{
	songs_view_update_node (lui->priv->songs, node);
}

/**
 * node_deleted_cb: a node was deleted from the library, update the ui
 */
static void
node_deleted_cb (Library *l, Node *node, LibraryUI *lui)
{
	LibraryNodeType type = node_get_int_property (node, "type");	

	switch (type)
	{
	case LIBRARY_NODE_SONG:
		/* remove from the songsview */
		songs_view_remove_node (lui->priv->songs, node);
		if (node == state_manager_get_playing_node (lui->priv->sm))
		{
			/* if this node was playing, stop that */
			state_manager_set_playing_node (lui->priv->sm, NULL);
		}

		/* check ui settings */
		check_album_all_count (lui);
		break;
	case LIBRARY_NODE_ARTIST:
		remove_artist (lui, node);
		break;
	case LIBRARY_NODE_ALBUM:
		if (g_list_length (node_get_children (node)) == 0)
		{
			/* no children left - delete row */
			remove_album (lui, node);
		}
		else
		{
			/* there are nodes left, but we caught a delete
			 * since all songs of one artist were removed from
			 * the album. So:
			 *
			 * - check whether album is visible
			 * - if so, check whether the selected artist (if not NULL)
			 *   is in the list of child songs
			 * - if so, return
			 * - if not, hide row
			 */

			GList *l;
			GValue val = {0, };
			GtkTreeIter *iter = g_hash_table_lookup (lui->priv->albums_node_to_iter, node);
		
			gtk_tree_model_get_value
				(GTK_TREE_MODEL (lui->priv->albums_liststore), iter,
				 ALBUM_COL_VISIBLE, &val);

			if (g_value_get_boolean (&val) == FALSE) return;

			for (l = node_get_children (node); l != NULL; l = g_list_next (l))
			{
				Node *n = l->data;

				if ((lui->priv->selected_artist == library_get_all_albums (lui->priv->library)) ||
				    (lui->priv->selected_artist == song_get_artist (n)))
				{
					return;
				}
			}

			gtk_list_store_set (lui->priv->albums_liststore, iter,
			    		    ALBUM_COL_VISIBLE, FALSE, -1);
		}
		break;
	default:
		break;
	}
}

/**
 * add_artist: Add an artist to the LibraryUI
 */
static void
add_artist (LibraryUI *l, Node *node)
{
	GtkTreeIter *iter;
	const char *name = NULL;
	int priority;
	gboolean is_all = (node == library_get_all_albums (l->priv->library));

	/* if we have a node, but it doesnt have a name, ignore it */
	if (is_all == FALSE)
	{
		name = node_get_string_property (node, NODE_PROPERTY_NAME);
		if (name == NULL || strlen (name) < 1) return;
		priority = 0;
	}
	else
	{
		name = g_strdup (_("All"));
		priority = 1;
	}

	/* append the item to the store */
	iter = g_new0 (GtkTreeIter, 1);
	gtk_list_store_append (l->priv->artists_liststore, iter);
	
	/* set row values */
	gtk_list_store_set (l->priv->artists_liststore, iter,
			    ARTIST_COL_NAME,     name,
			    ARTIST_COL_NODE,     node,
			    ARTIST_COL_PRIORITY, priority,
			    -1);

	/* set row data in node */
	g_hash_table_insert (l->priv->artists_node_to_iter, node, iter);
}

/**
 * remove_artist: Remove an artist from the LibraryUI
 */
static void
remove_artist (LibraryUI *l, Node *node)
{
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->artists_node_to_iter, node);
	GtkTreePath *path;
	GtkTreeSelection *selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW (l->priv->artists));

	/* store path */
	path = gtk_tree_model_get_path (GTK_TREE_MODEL (l->priv->artists_liststore), iter);

	/* remove it */
	gtk_list_store_remove (l->priv->artists_liststore, iter);
	g_hash_table_remove (l->priv->artists_node_to_iter, node);

	/* check if anything is still selected */
	if (gtk_tree_selection_get_selected (selection, NULL, iter) == FALSE)
	{
		/* select next or prev row */
		if (!gtk_tree_model_get_iter (GTK_TREE_MODEL (l->priv->artists_liststore),
					      iter, path))
		{
			gtk_tree_path_prev (path);
		}

		gtk_tree_selection_select_path (selection, path);
	}

	gtk_tree_iter_free (iter);
}

/**
 * add_album: Add an album to the LibraryUI
 */
static void
add_album (LibraryUI *l, Node *node)
{
	GtkTreeIter *iter;
	const char *name = NULL;
	gboolean vis = TRUE;
	int priority;
	gboolean is_all = (node == library_get_all_songs (l->priv->library));
	
	/* if an artist is selected, check whether this album belongs to it, if
	 * so we show it - otherwise it is hidden by default */
	if (l->priv->selected_artist != NULL)
		vis = node_has_child (l->priv->selected_artist, node);

	/* if we have a node, but it doesnt have a name, ignore it */
	if (is_all == FALSE)
	{
		name = node_get_string_property (node, NODE_PROPERTY_NAME);
		if (name == NULL || strlen (name) < 1) return;
		priority = 0;
	}
	else
	{
		name = g_strdup (_("All"));
		priority = 1;
	}

	/* append to the store */
	iter = g_new0 (GtkTreeIter, 1);
	gtk_list_store_append (l->priv->albums_liststore, iter);

	/* set store values */
	gtk_list_store_set (l->priv->albums_liststore, iter,
			    ALBUM_COL_NAME,     name,
			    ALBUM_COL_NODE,     node,
			    ALBUM_COL_VISIBLE,  vis,
			    ALBUM_COL_PRIORITY, priority,
			    -1);

	/* set row data in node */
	g_hash_table_insert (l->priv->albums_node_to_iter, node, iter);

	/* update number of items in artist "All" node */
	check_artist_all_count (l);
}

/**
 * remove_album: remove an album from the LibraryUI
 */
static void
remove_album (LibraryUI *l, Node *node)
{
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->albums_node_to_iter, node);
	GtkTreePath *path;
	GtkTreeSelection *selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW (l->priv->albums));

	/* store path */
	path = gtk_tree_model_get_path (GTK_TREE_MODEL (l->priv->albums_liststore), iter);

	/* remove from the store */
	gtk_list_store_remove (l->priv->albums_liststore, iter);
	g_hash_table_remove (l->priv->albums_node_to_iter, node);

	/* update All node in the artist list */
	check_artist_all_count (l);

	/* check if anything is still selected */
	if (gtk_tree_selection_get_selected (selection, NULL, iter) == FALSE)
	{
		/* select next or prev row */
		path = store_path_to_sort_path (l->priv->albums_filtermodel,
						l->priv->albums_sortmodel,
						path);

		if (path == NULL) goto done;
		
		if (!gtk_tree_model_get_iter (GTK_TREE_MODEL (l->priv->albums_liststore),
					      iter, path))
		{
			gtk_tree_path_prev (path);
		}

		gtk_tree_selection_select_path (selection, path);
	}
	
done:
	gtk_tree_iter_free (iter);
}

/**
 * artists_selection_changed_cb: selection in the artists treeview chaned
 */
static void
artists_selection_changed_cb (GtkTreeSelection *selection, LibraryUI *l)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GValue val = {0, };
	GList *iters = NULL, *li;
	Node *node;

	/* get the selected node */
	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) return;
	gtk_tree_model_get_value (model, &iter, ARTIST_COL_NODE, &val);
	l->priv->selected_artist = g_value_get_pointer (&val);

	/* run the filter */
	iters = g_list_append (iters, g_hash_table_lookup (l->priv->albums_node_to_iter,
							   library_get_all_songs (l->priv->library)));
	node = g_value_get_pointer (&val);
	for (li = node_get_children (node); li != NULL; li = g_list_next (li))
	{
		iters = g_list_append (iters, g_hash_table_lookup (l->priv->albums_node_to_iter, li->data));
	}
	rb_tree_model_filter_show_iters (RB_TREE_MODEL_FILTER (l->priv->albums_filtermodel), iters);
	g_list_free (iters);

	/* select "All" album node */
	select_albums_all (l);

	/* update All node */
	check_album_all_count (l);

	/* update statusbar */
	update_statusbar (l);
}

/**
 * albums_selection_changed_cb: selection in the albums treeviwe changed
 */
static void
albums_selection_changed_cb (GtkTreeSelection *selection, LibraryUI *l)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GValue val = {0, };
	Node *grandparent = NULL;

	/* get the selected node */
	if (gtk_tree_selection_get_selected (selection, &model, &iter) == FALSE) return;
	gtk_tree_model_get_value (model, &iter, ALBUM_COL_NODE, &val);

	/* update the LibraryUI struct to reflect the change in selection */
	l->priv->selected_album = g_value_get_pointer (&val);

	/* reset search */
	if (l->priv->ui != NULL && l->priv->querying == FALSE)
		ui_clear_search_entry (UI (l->priv->ui));

	/* create a filter */
	if (l->priv->selected_artist != library_get_all_albums (l->priv->library))
	{
		grandparent = l->priv->selected_artist;
	}

	/* run the filter */
	songs_view_filter (l->priv->songs, g_value_get_pointer (&val), grandparent);

	/* update statusbar */
	update_statusbar (l);
}

/**
 * songs_row_activated_cb: row in the songs list doubleclicked (activated)
 */
static void
songs_row_activated_cb (SongsView *v, Node *node, LibraryUI *l)
{
	/* set the node to be playing */
	state_manager_set_playing_node (l->priv->sm, node);
}

/**
 * queue_request_next_cb: the song stopped playing, we need to feed the queue
 * a new one
 */
static void
queue_request_next_cb (Queue *q, LibraryUI *l)
{
	Node *new;

	/* check that we are indeed the active source */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* get the next node */
	new = songs_view_get_next (l->priv->songs);
	
	/* and play it */
	state_manager_set_playing_node (l->priv->sm, new);
}

/**
 * select_albums_all: select "All" albums node
 */
static void
select_albums_all (LibraryUI *l)
{
	GtkTreeIter tmp;
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->albums_node_to_iter,
						 library_get_all_songs (l->priv->library));

	/* unselect to make sure the signal is emitted */
	gtk_tree_selection_unselect_all
		(gtk_tree_view_get_selection (GTK_TREE_VIEW (l->priv->albums)));

	/* convert the path from the ref into a sortmodel path */
	store_iter_to_sort_iter (l->priv->albums_filtermodel,
				 l->priv->albums_sortmodel,
				 &tmp, iter);

	/* and select the resulting path */
	gtk_tree_selection_select_iter (gtk_tree_view_get_selection
					(GTK_TREE_VIEW (l->priv->albums)),
					&tmp);
}

/**
 * select_artists_all: select "All" artists node
 */
static void
select_artists_all (LibraryUI *l)
{
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->artists_node_to_iter,
						 library_get_all_albums (l->priv->library));

	/* unselect to make sure the signal is emitted */
	gtk_tree_selection_unselect_all
		(gtk_tree_view_get_selection (GTK_TREE_VIEW (l->priv->artists)));

	/* and select the resulting path */
	gtk_tree_selection_select_iter (gtk_tree_view_get_selection
					(GTK_TREE_VIEW (l->priv->artists)), iter);
}

/**
 * ui_source_selected_cb: a source selected, check whether it's us
 */
static void
ui_source_selected_cb (UI *ui, gint id, LibraryUI *l)
{
	gboolean active = (id == l->priv->id);

	l->priv->active = active;

	if (active == TRUE)
	{
		/* clear out search entry */
		ui_clear_search_entry (l->priv->ui);
		
		/* update statusbar */
		update_statusbar (l);

		/* we cannot cut from or paste to the library */
		ui_set_can_paste (ui, FALSE);

		/* check songsview state */
		songs_view_check_state (l->priv->songs);

		/* we need to update the action button */
		ui_merge_widgets (l->priv->ui, l->priv->action_widgets);
	}
}

/**
 * ui_cut_cb: copy event caught, cutt all selected nodes to the clipboard
 */
static void
ui_cut_cb (UI *ui, LibraryUI *l)
{
	GList *sel;

	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;
	
	/* get the selection */
	sel = songs_view_get_checked_nodes (l->priv->songs);

	/* and remove it */
	remove_nodes (l, sel);
}

/**
 * ui_copy_cb: copy event caught, copy selected nodes to the clipboard
 */
static void
ui_copy_cb (UI *ui, LibraryUI *l)
{
	GList *sel;

	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;

	/* get the selection */
	sel = songs_view_get_checked_nodes (l->priv->songs);

	/* set to clipboard */
	source_manager_set_clipboard (l->priv->manager, sel);
}

/**
 * ui_rewind_cb: rewind signal caught, play previous song
 */
static void
ui_rewind_cb (UI *ui, LibraryUI *l)
{
	Node *new;

	/* check that we are indeed the active source */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* get prev song */
	new = songs_view_get_prev (l->priv->songs);
	
	/* and play it */
	state_manager_set_playing_node (l->priv->sm, new);
}

/**
 * ui_forward_cb: rewind signal caught, play next song
 */
static void
ui_forward_cb (UI *ui, LibraryUI *l)
{
	Node *new;

	/* check that we are indeed the active source */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* get next song */
	new = songs_view_get_next (l->priv->songs);
	
	/* and play it */
	state_manager_set_playing_node (l->priv->sm, new);
}

/**
 * songs_view_rewind_available_cb: rewind isnt possible
 */
static void
songs_view_rewind_available_cb (SongsView *v, gboolean avail, LibraryUI *l)
{
	/* if we're not playing, ignore */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* update ui */
	ui_set_can_rewind (l->priv->ui, avail);
}

/**
 * songs_view_forward_available_cb: forward isnt possible
 */
static void
songs_view_forward_available_cb (SongsView *v, gboolean avail, LibraryUI *l)
{
	/* if we're not playing, ignore */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* update ui */
	ui_set_can_forward (l->priv->ui, avail);
}

/**
 * songs_view_selection_available_cb: selection availability changed
 */
static void
songs_view_selection_available_cb (SongsView *v, gboolean avail, LibraryUI *l)
{
	if (l->priv->active == FALSE) return;
	
	ui_set_can_cut (l->priv->ui, avail);
	ui_set_can_copy (l->priv->ui, avail);
}

/**
 * browser_cb: action signal received, change the layout
 */
static void
browser_cb (GtkWidget *widget, LibraryUI *l)
{
	library_ui_show_browser (l, !eel_gconf_get_boolean (CONF_SHOW_BROWSER));
}

/**
 * ui_play_cb: play signal received, set queue to be playing
 */
static void
ui_play_cb (UI *ui, LibraryUI *l)
{
	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;

	/* start playing */
	if (state_manager_get_paused (l->priv->sm) == TRUE)
		state_manager_set_paused (l->priv->sm, FALSE);
	else
		state_manager_set_playing_node (l->priv->sm,
						songs_view_get_first (l->priv->songs));
}

/**
 * ui_stop_cb: stop signal received, stop the playing source
 */
static void
ui_stop_cb (UI *ui, LibraryUI *l)
{
	/* if we're not playing, ignore */	
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* stop playing */	
	state_manager_set_playing_node (l->priv->sm, NULL);
	state_manager_set_active (l->priv->sm, FALSE);
}

/**
 * ui_query_cb: search entry activated, filter the list
 */
static void 
ui_query_cb (UI *ui, gchar *text, LibraryUI *l)
{
	Node *search;

	if (l->priv->active == FALSE) return;
	
	/* start querying */
	l->priv->querying = TRUE;
	
	/* create a filter */
	search = library_search (l->priv->library, text);

	/* select "All" nodes */
	if (l->priv->selected_artist != library_get_all_albums (l->priv->library))
		select_artists_all (l);

	/* run the filter */
	songs_view_filter (l->priv->songs, search, NULL);

	/* cleanup */
	g_object_unref (G_OBJECT (search));

	/* done querynig */
	l->priv->querying = FALSE;

	/* update statusbar */
	update_statusbar (l);
}

/**
 * ui_pause_cb: pause signal received, set queue to be paused
 */
static void
ui_pause_cb (UI *ui, LibraryUI *l)
{
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	state_manager_set_paused (l->priv->sm, TRUE);
}

/**
 * songs_row_remove_request_cb: delete pressed in songsview, remove
 * selected nodes
 */
static void
songs_row_remove_request_cb (SongsView *v, GList *nodes, LibraryUI *l)
{
	remove_nodes (l, nodes);

	g_list_free (nodes);
}

/**
 * check_artist_all_count: update All item node in the artist list
 */
static void
check_artist_all_count (LibraryUI *l)
{
	gchar *new;
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->artists_node_to_iter,
						 library_get_all_albums (l->priv->library));

	/* - 1 since we do not count the All row (it's not an album) */
	new = g_strdup_printf (_("All (%d albums)"), l->priv->albums_liststore->length - 1);

	gtk_list_store_set (l->priv->artists_liststore, iter,
			    ARTIST_COL_NAME, new, -1);

	g_free (new);

}

/**
 * check_album_all_count: update All item node in the album list
 */
static void
check_album_all_count (LibraryUI *l)
{
	gchar *new;
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->albums_node_to_iter,
						 library_get_all_songs (l->priv->library));

	new = g_strdup_printf (_("All (%d songs)"), songs_view_num_rows (l->priv->songs));

	gtk_list_store_set (l->priv->albums_liststore, iter,
			    ALBUM_COL_NAME, new, -1);

	g_free (new);
}

/**
 * ui_set_shuffle_cb: shuffle reqeuest caught, pass on to songsview
 */
static void
ui_set_shuffle_cb (UI *ui, gboolean shuffle, LibraryUI *l)
{
	songs_view_set_shuffle (l->priv->songs, shuffle);
}

/**
 * ui_set_repeat_cb: repeat reqeuest caught, pass on to songsview
 */
static void
ui_set_repeat_cb (UI *ui, gboolean repeat, LibraryUI *l)
{
	songs_view_set_repeat (l->priv->songs, repeat);
}

/**
 * update_statusbar: set the statusbar message
 */
static void
update_statusbar (LibraryUI *l)
{
	gchar *msg;

	if (l->priv->ui == NULL) return;
	
	msg = songs_view_build_status_message (l->priv->songs);
	ui_set_status (l->priv->ui, msg);
	g_free (msg);
}

/**
 * library_ui_get_selection: get the checkbox selection
 */
GList *
library_ui_get_selection (LibraryUI *l)
{
	return songs_view_get_checked_nodes (l->priv->songs);
}

/**
 * ui_clear_cb: clear everything
 */
static void
ui_clear_cb (UI *ui, LibraryUI *l)
{
	GList *sel;

	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;
	
	/* get the visible nodes */
	sel = songs_view_get_shown_nodes (l->priv->songs);

	/* remove em all */
	remove_nodes (l, sel);

	g_list_free (sel);
}

/**
 * ui_select_all_cb: select everything
 */
static void
ui_select_all_cb (UI *ui, LibraryUI *l)
{
	GList *vis;
	
	if (l->priv->active == FALSE) return;
	
	vis = songs_view_get_shown_nodes (l->priv->songs);
	songs_view_select_nodes (l->priv->songs, vis);
	g_list_free (vis);
}

/**
 * ui_select_none_cb: unselect everything
 */
static void
ui_select_none_cb (UI *ui, LibraryUI *l)
{
	if (l->priv->active == FALSE) return;
	
	songs_view_select_nodes (l->priv->songs, NULL);
}

/**
 * remove_nodes: remove a list of nodes from the playlist
 */
static void
remove_nodes (LibraryUI *l, GList *sel)
{
	GList *li;

	/* iterate over the selection removing the items */
	for (li = sel; li != NULL; li = g_list_next (li))
	{
		Node *node = li->data;
		if (node == state_manager_get_playing_node (l->priv->sm))
		{
			/* if this node was playing, stop that */
			state_manager_set_playing_node (l->priv->sm, NULL);
		}
		library_remove_uri (l->priv->library,
				    node_get_string_property (node, "uri"));
	}

	update_statusbar (l);
}

/**
 * source_manager_stop_playing_cb: stop playing
 */
static void
source_manager_stop_playing_cb (SourceManager *m, gpointer exception, LibraryUI *l)
{
	if (exception == l) return;

	state_manager_set_active (l->priv->sm, FALSE);
}

/**
 * library_ui_show_browser: sets the album/artist browser visibility
 */
static void
library_ui_show_browser (LibraryUI *l, gboolean show)
{
	if (show == TRUE)
	{
		gtk_widget_show (l->priv->hbox);
	}
	else
	{
		gtk_widget_hide (l->priv->hbox);
	}

	eel_gconf_set_boolean (CONF_SHOW_BROWSER, show);
}

/**
 * state_manager_start_playing_cb: forward signal
 */
static void
state_manager_start_playing_cb (StateManager *m, LibraryUI *l)
{
	g_signal_emit (G_OBJECT (l), library_ui_signals[START_PLAYING], 0);
}

/**
 * state_manager_stop_playing_cb: forward signal
 */
static void
state_manager_stop_playing_cb (StateManager *m, LibraryUI *l)
{
	g_signal_emit (G_OBJECT (l), library_ui_signals[STOP_PLAYING], 0);
}

/**
 * library_ui_set_property: set a LibraryUI property
 */
static void
library_ui_set_property (GObject *object,
			 guint prop_id,
			 const GValue *value,
			 GParamSpec *pspec)
{
	LibraryUI *l = LIBRARY_UI (object);

	switch (prop_id)
	{
		case PROP_SOURCE_MANAGER:
			library_ui_set_manager (l, g_value_get_object (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/**
 * library_ui_get_property: get a LibraryUI property
 */
static void
library_ui_get_property (GObject *object,
			 guint prop_id,
			 GValue *value,
			 GParamSpec *pspec)
{
	LibraryUI *l = LIBRARY_UI (object);

	switch (prop_id)
	{
		case PROP_SOURCE_MANAGER:
			g_value_set_object (value, l->priv->manager);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}
