/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: node.c,v 1.22 2002/04/02 07:01:07 jbaayen Exp $
 */

#include <glib.h>
#include <string.h>
#include <stdio.h>

#include "node.h"

struct _NodeDetails
{
	GList *parents;
	Node *refnode;
	GList *children;
	Node *grandparent;

	GHashTable *properties;
	GHashTable *int_properties;

	char *saved_xml_node_name;
};

/* object function prototypes */
static void node_class_init (NodeClass *klass);
static void node_init (Node *t);
static void node_finalize (GObject *object);

/* local function prototypes */
static gboolean free_props (gpointer key, gpointer val, gpointer data);

/* globals */
GObjectClass *parent_class = NULL;

GType 
node_get_type (void)
{
        static GType node_type = 0;

        if (node_type == 0)
        {
                static const GTypeInfo our_info =
                {
                        sizeof (NodeClass),
                        NULL, /* base_init */
                        NULL, /* base_finalize */
                        (GClassInitFunc) node_class_init,
                        NULL,
                        NULL, /* class_data */
                        sizeof (Node),
                        0,    /* n_preallocs */
                        (GInstanceInitFunc) node_init
                };

                node_type = g_type_register_static (G_TYPE_OBJECT,
						    "Node",
						    &our_info, 0);
        }

        return node_type;

}

static void
node_class_init (NodeClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);

        parent_class = g_type_class_peek_parent (klass);

        object_class->finalize = node_finalize;
}

static void
node_init (Node *t)
{
        t->details = g_new0 (NodeDetails, 1);

	t->details->properties     = g_hash_table_new (g_str_hash, g_str_equal);
	t->details->int_properties = g_hash_table_new (g_str_hash, g_str_equal);
	t->details->children       = NULL;
	t->details->parents        = NULL;
	t->details->grandparent    = NULL;
	t->details->refnode        = NULL;
	t->details->saved_xml_node_name = NULL;
}

static void
node_finalize (GObject *object)
{
        Node *t;
	GList *l;

        g_return_if_fail (object != NULL);
        g_return_if_fail (IS_NODE (object));
        
        t = NODE (object);

        g_return_if_fail (t->details != NULL);

	for (l = t->details->children; l != NULL; l = g_list_next (l))
	{
		Node *tmp = l->data;
		g_list_remove (tmp->details->parents, t);
	}
	g_hash_table_foreach_remove (t->details->properties,
				     (GHRFunc) free_props, NULL);
	g_hash_table_destroy (t->details->properties);
	g_hash_table_destroy (t->details->int_properties);

	g_free (t->details->saved_xml_node_name);

        g_free (t->details);

        G_OBJECT_CLASS (parent_class)->finalize (object);
}

/* xml_name is optional, and may be set later, but is convenient
   for saving the node to XML */
Node *
node_new (const char *xml_name)
{
	Node *n;

	n = NODE (g_object_new (TYPE_NODE, NULL));

	if (xml_name != NULL) {
		node_set_xml_name (n, xml_name);
	}

	g_return_val_if_fail (n->details != NULL, NULL);

	return n;
}

GList *
node_get_parents (Node *node)
{
	g_return_val_if_fail (IS_NODE (node), NULL);

        return node->details->parents;
}

Node *
node_get_grandparent (Node *node)
{
	g_return_val_if_fail (IS_NODE (node), NULL);

        return node->details->grandparent;
}

void
node_set_grandparent (Node *node, Node *granny)
{
	g_return_if_fail (IS_NODE (node));

	node->details->grandparent = granny;
}

GList *
node_get_children (Node *node)
{
	g_return_val_if_fail (IS_NODE (node), NULL);

        return node->details->children;
}

const gchar *
node_get_string_property (Node *node, const gchar *property)
{
	Node *n;

	g_return_val_if_fail (IS_NODE (node), NULL);

	n = node->details->refnode ? node->details->refnode : node;
	g_return_val_if_fail (IS_NODE (n), NULL);
	
	return (gchar*) g_hash_table_lookup (n->details->properties,
					     property);
}

gint
node_get_int_property (Node *node, const gchar *property)
{
	Node *n;

	g_return_val_if_fail (IS_NODE (node), -1);

	n = node->details->refnode ? node->details->refnode : node;
	g_return_val_if_fail (IS_NODE (n), -1);

	return GPOINTER_TO_INT (g_hash_table_lookup (n->details->int_properties,
			                             property));
}

void 
node_append (Node *parent, Node *node)
{
	Node *p;

	g_return_if_fail (IS_NODE (node));

	if (node_has_child (parent, node) == TRUE) return;

	if (parent != NULL)
	{
		p = parent->details->refnode ? parent->details->refnode : parent;
		g_return_if_fail (IS_NODE (p));

		node->details->parents = g_list_append (node->details->parents, p);

		p->details->children = 
			g_list_append (p->details->children,
				       node);
	}

	/* The node with parent == NULL is the toplevel */
}

void 
node_remove (Node *node,
	     gint h_level,
	     gint l_level,
	     ActionNotifier func,
	     gpointer data)
{
	GList *c, *parents, *p;
	Node *grandparent = node_get_grandparent (node);

	g_return_if_fail (IS_NODE (node));

	if (h_level < l_level) return;

	parents = node->details->parents;

	for (p = parents; p != NULL; p = g_list_next (p))
	{
		Node *parent = p->data;
		gboolean found = FALSE;
		gboolean last = FALSE;

		if (parent == NULL) continue;

		parent->details->children = 
			g_list_remove (parent->details->children, node);

		/* last node of this grandparent ? */
		if (grandparent != NULL && g_list_length (parent->details->children) > 0)
		{
			GList *s;

			for (s = parent->details->children; s != NULL; s = g_list_next (s))
			{
				if (grandparent ==
				    node_get_grandparent (s->data))
				{
					found = TRUE;
					continue;
				}
			}

			/* no more items of the in this grandparent */
			if (found == FALSE)
			{
				grandparent->details->children =
					g_list_remove (grandparent->details->children, parent);
				if (func)
					func (parent, l_level, data);
				if (g_list_length (grandparent->details->children) == 0)
				{
					if (func)
						func (grandparent, l_level + 1, data);
					node_remove (grandparent, h_level, l_level + 2,
						     func, data);
				}
				last = TRUE;
			}
		}
		
		/* last node ? */
		if (last == FALSE && g_list_length (parent->details->children) == 0)
		{
			if (func)
				func (parent, l_level, data);
			node_remove (parent, h_level, l_level + 1, 
				     func, data);
		}
	}

	/* remove children */
	for (c = node->details->children; 
	     c != NULL;
	     c = g_list_next (c))
	{
		if (func)
			func (NODE (c->data), l_level, data);
		node_remove (NODE (c->data), h_level, l_level - 1, func, data);
	}

	g_object_unref (node);
}

void
node_set_string_property (Node *node, 
			  const gchar *property,
			  gchar *value)
{
	Node *n;
	char *old_value;

	g_return_if_fail (IS_NODE (node));

	n = node->details->refnode ? node->details->refnode : node;
	g_return_if_fail (IS_NODE (n));

	old_value = g_hash_table_lookup (n->details->properties, property);

	if (old_value != NULL)
	{
	        g_free (old_value);
	}

	g_hash_table_insert (n->details->properties,
			     g_strdup (property), value);
}

void
node_set_int_property (Node *node, 
		       const gchar *property,
		       gint value)
{
	Node *n;

	g_return_if_fail (IS_NODE (node));

	n = node->details->refnode ? node->details->refnode : node;
	g_return_if_fail (IS_NODE (n));

	g_hash_table_insert (n->details->int_properties,
			     g_strdup (property), GINT_TO_POINTER (value));
}

static gboolean
free_props (gpointer key, gpointer val, gpointer data)
{
	g_free (key);
	g_free (val);
	return TRUE;
}

gboolean
node_has_child (Node *node, Node *child)
{
	return (g_list_find (node_get_children (node), child) != NULL);
}

/* Set the name of the XML node to save this node to if/when we save */
void 
node_set_xml_name (Node *node, const char *xml_name)
{
	g_free (node->details->saved_xml_node_name);

	node->details->saved_xml_node_name = g_strdup (xml_name);
}

static void
save_string_property (const char *key, const char *value, xmlNodePtr xml_node)
{
	xmlSetProp (xml_node, key, value);	
}

static void
save_int_property (const char *key, gpointer pvalue, xmlNodePtr xml_node)
{
	char *string_value;
	int value = GPOINTER_TO_INT (pvalue);

	string_value = g_strdup_printf ("%d", value);
	xmlSetProp (xml_node, key, string_value);

	g_free (string_value);
}

/* Save this node, but not its children, return the newly created XML node */
xmlNodePtr
node_save_to_xml (Node *node, Node *grandparent, xmlNodePtr parent, gboolean save_children)
{
	xmlNodePtr xml_node;

	if (grandparent != NULL && 
	    grandparent != node &&
	    g_list_find (grandparent->details->children, node) == NULL)
	{
		if (node_get_grandparent (node) != grandparent) return NULL;
	}

	if (node->details->saved_xml_node_name == NULL)
	{
		g_warning ("Saving unnamed node to XML, using generic name \"Node\"");
		node->details->saved_xml_node_name = g_strdup ("Node");
	}

	xml_node = xmlNewChild (parent, NULL, node->details->saved_xml_node_name, NULL);

	g_hash_table_foreach (node->details->properties,
			      (GHFunc) save_string_property, xml_node);
	g_hash_table_foreach (node->details->int_properties,
			      (GHFunc) save_int_property, xml_node);

	if (save_children == TRUE)
	{
		GList *l;

		for (l = node->details->children; l != NULL; l = g_list_next (l))
		{
			node_save_to_xml (l->data, grandparent, xml_node, TRUE);
		}
	}

	return xml_node;
}
