/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: playlist-ui.c,v 1.61 2002/04/11 15:31:40 jbaayen Exp $
 */

#include <libgnomevfs/gnome-vfs-ops.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-init.h>
#include <libxml/tree.h>
#include <string.h>

#include "playlist-ui.h"
#include "queue.h"
#include "node-utils.h"
#include "songs-view.h"
#include "icons.h"
#include "state-manager.h"
#include "eel-gconf-extensions.h"

struct _PlaylistUIPrivate
{
	SourceManager *manager;   /* the associated SourceManager */

	Library *library;         /* the associated Library */

	Queue *queue;             /* the associated Queue */

	StateManager *sm;         /* the associated StateManager */

	UI *ui;                   /* the associated UI */

	char *name;               /* the playlist name */
	char *file;               /* the playlist filename */

	SongsView *songs;         /* the list of songs */

	int id;                   /* our source ID */

	gboolean active;          /* are we the selected source ? */

	GList *action_widgets;    /* list of action widgets */
};

/* all the different signals */
enum
{
	START_PLAYING,
	STOP_PLAYING,
	REMOVED,
	LAST_SIGNAL
};

/* GObject properties */
enum
{
	PROP_0,
	PROP_SOURCE_MANAGER,
	PROP_FILE_NAME,
	PROP_NAME,
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint playlist_ui_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void playlist_ui_class_init (PlaylistUIClass *klass);
static void playlist_ui_init (PlaylistUI *l);
static void playlist_ui_finalize (GObject *object);
static void playlist_ui_set_property (GObject *object,
			              guint prop_id,
			              const GValue *value,
			              GParamSpec *pspec);
static void playlist_ui_get_property (GObject *object,
			              guint prop_id,
			              GValue *value,
			              GParamSpec *pspec);

/* signal prototypes */
static void node_changed_cb (Library *l, Node *node, PlaylistUI *lui);
static void node_deleted_cb (Library *l, Node *node, PlaylistUI *lui);
static void queue_request_next_cb (Queue *q, PlaylistUI *l);
static void ui_source_selected_cb (UI *ui, int id, PlaylistUI *l);
static void ui_source_deleted_cb (UI *ui, int id, PlaylistUI *l);
static void ui_source_edited_cb (UI *ui, int id, char *name, PlaylistUI *l);
static void ui_cut_cb (UI *ui, PlaylistUI *l);
static void ui_copy_cb (UI *ui, PlaylistUI *l);
static void ui_paste_cb (UI *ui, PlaylistUI *l);
static void songs_row_remove_request_cb (SongsView *v, GList *nodes,
					 PlaylistUI *l);
static void songs_row_activated_cb (SongsView *v, Node *node, PlaylistUI *l);
static void ui_rewind_cb (UI *ui, PlaylistUI *l);
static void ui_forward_cb (UI *ui, PlaylistUI *l);
static void songs_view_rewind_available_cb (SongsView *v, gboolean avail,
						     PlaylistUI *l);
static void songs_view_forward_available_cb (SongsView *v, gboolean avail,
						     PlaylistUI *l);
static void ui_play_cb (UI *ui, PlaylistUI *l);
static void ui_stop_cb (UI *ui, PlaylistUI *l);
static void ui_query_cb (UI *ui, char *text, PlaylistUI *l);
static void ui_pause_cb (UI *ui, PlaylistUI *l);
static void ui_set_shuffle_cb (UI *ui, gboolean shuffle, PlaylistUI *l);
static void ui_set_repeat_cb (UI *ui, gboolean repeat, PlaylistUI *l);
static void ui_clear_cb (UI *ui, PlaylistUI *l);
static void ui_select_all_cb (UI *ui, PlaylistUI *l);
static void ui_select_none_cb (UI *ui, PlaylistUI *l);
static void source_manager_clipboard_available_cb (SourceManager *m, gboolean available, PlaylistUI *l);
static void songs_view_selection_available_cb (SongsView *v, gboolean avail, PlaylistUI *l);
static void source_manager_stop_playing_cb (SourceManager *m, gpointer exception, PlaylistUI *l);
static void state_manager_start_playing_cb (StateManager *m, PlaylistUI *l);
static void state_manager_stop_playing_cb (StateManager *m, PlaylistUI *l);

/* local functions */
static void playlist_ui_save_to_xml (PlaylistUI *l, const char *fn);
static void playlist_ui_save_song_node (Node *song, xmlNodePtr parent_node);
static void playlist_ui_load_from_xml (PlaylistUI *l, const char *fn);
static void playlist_ui_load_song_from_xml (PlaylistUI *l, xmlNodePtr node);
static char *filename_from_name (const char *name);
static void update_statusbar (PlaylistUI *l);
static void remove_nodes (PlaylistUI *l, GList *sel);
static void playlist_ui_set_manager (PlaylistUI *lui, SourceManager *m);

/**
 * playlist_ui_get_type: get the GObject type of the PlaylistUI
 */
GType
playlist_ui_get_type (void)
{
	static GType playlist_ui_type = 0;

  	if (playlist_ui_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (PlaylistUIClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) playlist_ui_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (PlaylistUI),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) playlist_ui_init
      		};

      		playlist_ui_type = g_type_register_static (G_TYPE_OBJECT,
                				           "PlaylistUI",
                                           	           &our_info, 0);
    	}

	return playlist_ui_type;
}

/**
 * playlist_ui_class_init: initialize the PlaylistUI class
 */
static void
playlist_ui_class_init (PlaylistUIClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = playlist_ui_finalize;

	object_class->set_property = playlist_ui_set_property;
	object_class->get_property = playlist_ui_get_property;

	/* init properties */
	g_object_class_install_property (object_class,
					 PROP_SOURCE_MANAGER,
					 g_param_spec_object ("source-manager",
						 	      "SourceManager",
							      "Associated SourceManager object",
							      source_manager_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_FILE_NAME,
					 g_param_spec_string ("file-name",
						 	      "filename",
							      "Filename of this playlist",
							      NULL,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_NAME,
					 g_param_spec_string ("name",
						 	      "name",
							      "Name of this playlist",
							      _("Untitled"),
							      G_PARAM_READWRITE));

	/* init signals */
	playlist_ui_signals[START_PLAYING] =
   		g_signal_new ("start_playing",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PlaylistUIClass, start_playing),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	playlist_ui_signals[STOP_PLAYING] =
   		g_signal_new ("stop_playing",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PlaylistUIClass, stop_playing),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	playlist_ui_signals[REMOVED] =
   		g_signal_new ("removed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PlaylistUIClass, removed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__INT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_INT);
}

/**
 * playlist_ui_init: intialize the PlaylistUI object
 */
static void
playlist_ui_init (PlaylistUI *l)
{
	l->priv = g_new0 (PlaylistUIPrivate, 1);

	/* create a songs view */
	l->priv->songs = SONGS_VIEW (songs_view_new ());

	/* connect songsview signals */
	g_signal_connect (G_OBJECT (l->priv->songs), "row_remove_request",
			  G_CALLBACK (songs_row_remove_request_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "row_activated",
			  G_CALLBACK (songs_row_activated_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "rewind_available",
			  G_CALLBACK (songs_view_rewind_available_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "forward_available",
			  G_CALLBACK (songs_view_forward_available_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "selection_available",
			  G_CALLBACK (songs_view_selection_available_cb), l);

	/* nullify */
	l->priv->active = FALSE;
	l->priv->ui = NULL;
}

/**
 * playlist_ui_finalize: finalize the PlaylistUI object
 */
static void
playlist_ui_finalize (GObject *object)
{
	PlaylistUI *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_PLAYLIST_UI (object));
	
   	s = PLAYLIST_UI (object);

	g_return_if_fail (s->priv != NULL);

	/* disconnect signals */
	g_signal_handlers_disconnect_matched (G_OBJECT (s->priv->ui),
					      G_SIGNAL_MATCH_DATA,
					      0,
					      0,
					      NULL,
					      NULL,
					      object);
	g_signal_handlers_disconnect_matched (G_OBJECT (s->priv->library),
					      G_SIGNAL_MATCH_DATA,
					      0,
					      0,
					      NULL,
					      NULL,
					      object);
	g_signal_handlers_disconnect_matched (G_OBJECT (s->priv->queue),
					      G_SIGNAL_MATCH_DATA,
					      0,
					      0,
					      NULL,
					      NULL,
					      object);
	g_signal_handlers_disconnect_matched (G_OBJECT (s->priv->manager),
					      G_SIGNAL_MATCH_DATA,
					      0,
					      0,
					      NULL,
					      NULL,
					      object);

	/* save the playlist to a file */
	playlist_ui_save_to_xml (s, s->priv->file);

	/* get rid of songsview */
	gtk_widget_destroy (GTK_WIDGET (s->priv->songs));

	/* get rid of statemanager */
	g_object_unref (G_OBJECT (s->priv->sm));

	/* get rid of actionbuttons */
	g_list_foreach (s->priv->action_widgets, (GFunc) gtk_widget_destroy, NULL);
	g_list_free (s->priv->action_widgets);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * playlist_ui_new: create a new PlaylistUI object
 */
PlaylistUI *
playlist_ui_new (SourceManager *manager)
{
 	PlaylistUI *l;

	l = PLAYLIST_UI (g_object_new (TYPE_PLAYLIST_UI,
				       "source-manager", manager,
				       NULL));

	g_return_val_if_fail (l->priv != NULL, NULL);

	return l;
}

/**
 * playlist_ui_set_manager: set the SourceManager for this PlaylistUI
 */
static void
playlist_ui_set_manager (PlaylistUI *lui, SourceManager *m)
{
	GdkPixbuf *playlist_pixbuf;
	GtkWidget *button;
	GdkPixbuf *icon;

	lui->priv->manager = m;
	lui->priv->library = source_manager_get_library (m);
	lui->priv->ui      = source_manager_get_ui (m);
	lui->priv->queue   = source_manager_get_queue (m);

	/* hook up to library */
	g_signal_connect (G_OBJECT (lui->priv->library), "node_changed",
			  G_CALLBACK (node_changed_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->library), "node_deleted",
			  G_CALLBACK (node_deleted_cb), lui);

	/* connect manager signals */
	g_signal_connect (G_OBJECT (lui->priv->manager),
			  "clipboard_available", G_CALLBACK (source_manager_clipboard_available_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->manager),
			  "stop_playing", G_CALLBACK (source_manager_stop_playing_cb), lui);

	/* connect signal to feed new songs to the queue when it wants them */
	g_signal_connect (G_OBJECT (lui->priv->queue),
			  "request_next", G_CALLBACK (queue_request_next_cb),
			  lui);
	
	/* add to the list of sources */
	lui->priv->id = ui_append_source (lui->priv->ui, GTK_WIDGET (lui->priv->songs));

	/* load the pixbuf used as the source icon */
	playlist_pixbuf = gtk_widget_render_icon (lui->priv->ui->notebook,
						  RHYTHMBOX_STOCK_PLAYLIST,
						  GTK_ICON_SIZE_MENU, NULL);

	/* set source properties */
	ui_source_set_name (lui->priv->ui, lui->priv->id, lui->priv->name);
	ui_source_set_icon (lui->priv->ui, lui->priv->id, playlist_pixbuf);
	ui_source_set_editable (lui->priv->ui, lui->priv->id, TRUE);	

	/* connect ui signals */
	g_signal_connect (G_OBJECT (lui->priv->ui), "source_selected",
			  G_CALLBACK (ui_source_selected_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "source_deleted",
			  G_CALLBACK (ui_source_deleted_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "source_edited",
			  G_CALLBACK (ui_source_edited_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "cut",
			  G_CALLBACK (ui_cut_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "copy",
			  G_CALLBACK (ui_copy_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "paste",
			  G_CALLBACK (ui_paste_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "clear",
			  G_CALLBACK (ui_clear_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "select_none",
			  G_CALLBACK (ui_select_none_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "select_all",
			  G_CALLBACK (ui_select_all_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "rewind",
			  G_CALLBACK (ui_rewind_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "forward",
			  G_CALLBACK (ui_forward_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "play",
			  G_CALLBACK (ui_play_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "stop",
			  G_CALLBACK (ui_stop_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "query",
			  G_CALLBACK (ui_query_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "pause",
			  G_CALLBACK (ui_pause_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "set_shuffle",
			  G_CALLBACK (ui_set_shuffle_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->ui), "set_repeat",
			  G_CALLBACK (ui_set_repeat_cb), lui);

	/* add state manager */
	lui->priv->sm = state_manager_new (lui->priv->queue, lui->priv->ui, lui->priv->songs, lui->priv->id);

	g_signal_connect (G_OBJECT (lui->priv->sm), "start_playing",
			  G_CALLBACK (state_manager_start_playing_cb), lui);
	g_signal_connect (G_OBJECT (lui->priv->sm), "stop_playing",
			  G_CALLBACK (state_manager_stop_playing_cb), lui);

	/* load state */
	songs_view_set_shuffle (lui->priv->songs, eel_gconf_get_boolean (CONF_SHUFFLE));
	songs_view_set_repeat (lui->priv->songs, eel_gconf_get_boolean (CONF_REPEAT));

	/* build action widgets */
	icon = gtk_widget_render_icon (GTK_WIDGET (lui->priv->songs),
				       RHYTHMBOX_STOCK_ACTION_PLAYLIST,
				       RHYTHMBOX_ICON_SIZE_ACTION, NULL);
	button = ui_new_action_button (lui->priv->ui, _("Burn CD"),
				       _("Burn a CD"),
				       icon, FALSE);
	g_object_unref (G_OBJECT (icon));

	lui->priv->action_widgets = NULL;
	lui->priv->action_widgets = g_list_append (lui->priv->action_widgets, button);

	/* show */
	gtk_widget_show (GTK_WIDGET (lui->priv->songs));
}

/**
 * playlist_ui_get_source_id: return the source id of the library
 */
int
playlist_ui_get_source_id (PlaylistUI *l)
{
	return l->priv->id;
}

/**
 * node_changed_cb: a node in the library was changed, update the ui
 */
static void
node_changed_cb (Library *l, Node *node, PlaylistUI *lui)
{
	songs_view_update_node (lui->priv->songs, node);
}

/**
 * node_deleted_cb: a node was deleted from the library, update the ui
 */
static void
node_deleted_cb (Library *l, Node *node, PlaylistUI *lui)
{
	LibraryNodeType type = node_get_int_property (node, "type");	
	switch (type)
	{
	case LIBRARY_NODE_SONG:
		/* this function checks whether this node is valid for this
		 * list internally */
		songs_view_remove_node (lui->priv->songs, node);
		if (node == state_manager_get_playing_node (lui->priv->sm))
		{
			/* was playing, stop that */
			state_manager_set_playing_node (lui->priv->sm, NULL);
		}
		break;
	default:
		break;
	}
}

/**
 * songs_row_activated_cb: row in the songs list doubleclicked (activated)
 */
static void
songs_row_activated_cb (SongsView *v, Node *node, PlaylistUI *l)
{
	/* set the node to be playing */
	state_manager_set_playing_node (l->priv->sm, node);
}

/**
 * queue_request_next_cb: the song stopped playing, we need to feed the queue
 * a new one
 */
static void
queue_request_next_cb (Queue *q, PlaylistUI *l)
{
	Node *new;

	/* check that we are indeed the active source */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* get the next node */
	new = songs_view_get_next (l->priv->songs);
	
	/* and play it ! */
	state_manager_set_playing_node (l->priv->sm, new);
}

/**
 * playlist_ui_set_name: set the name of the current playlist
 */
void
playlist_ui_set_name (PlaylistUI *l, const char *name)
{
	char *old = g_strdup (l->priv->file);
	
	l->priv->name = g_strdup (name);
	
	g_assert (l->priv->ui != NULL);

	/* update in the ui */
	ui_source_set_name (l->priv->ui, l->priv->id, l->priv->name);

	/* rename the playlist file */
	l->priv->file = filename_from_name (name);
	if (old && strcmp (old, l->priv->file) != 0)
	{
		gnome_vfs_move (old, l->priv->file, FALSE);
	}

	g_free (old);

	g_object_notify (G_OBJECT (l), "name");
}

/**
 * playlist_ui_load_from_file: load playlist data from a file
 */
void
playlist_ui_load_from_file (PlaylistUI *l, const char *fn)
{
	if (fn == NULL)
	{
		GnomeVFSHandle *handle;

		/* no filename yet, make a unique filename */
		l->priv->file = filename_from_name (l->priv->name);

		/* claim the file */
		gnome_vfs_create (&handle, l->priv->file,
				  GNOME_VFS_OPEN_WRITE,
				  FALSE,
				  0644);
		gnome_vfs_close (handle);
	}
	else
	{
		/* load the file */
		l->priv->file = g_strdup (fn);
		playlist_ui_load_from_xml (l, l->priv->file);
	}

	g_object_notify (G_OBJECT (l), "file-name");
}

/**
 * playlist_ui_get_file: return filename this pls is associated with
 */
const char *
playlist_ui_get_file (PlaylistUI *l)
{
	return (const char *) l->priv->file;
}

/**
 * ui_source_selected_cb: a source selected, check whether it's us
 */
static void
ui_source_selected_cb (UI *ui, int id, PlaylistUI *l)
{
	gboolean active = (id == l->priv->id);

	l->priv->active = active;

	if (active == TRUE)
	{
		/* reset search */
		ui_clear_search_entry (l->priv->ui);

		/* update statusbar */
		update_statusbar (l);

		/* we can do all fun C&P stuff */
		ui_set_can_paste (ui, (g_list_length (source_manager_get_clipboard (l->priv->manager)) > 0));
		songs_view_check_state (l->priv->songs);

		/* we need to update the action button */
		ui_merge_widgets (l->priv->ui, l->priv->action_widgets);
	}
}

/**
 * ui_source_deleted_cb: a source delete request was caught,
 * see if it's us - if so remove ourself
 */
static void
ui_source_deleted_cb (UI *ui, int id, PlaylistUI *l)
{
	if (id != l->priv->id) return;

	/* if we were playing, stop that */
	if (state_manager_get_active (l->priv->sm) == TRUE)
	{
		state_manager_set_playing_node (l->priv->sm, NULL);
		state_manager_set_active (l->priv->sm, FALSE);
	}
	
	/* remove from sourcelist */
	ui_remove_source (l->priv->ui, l->priv->id);

	/* let the manager know we are gone */
	g_signal_emit (G_OBJECT (l), playlist_ui_signals[REMOVED], 0, id);

	/* remove the file */
	gnome_vfs_unlink (l->priv->file);

	g_object_unref (l);
}

/**
 * ui_source_edited_cb: a source editing request was caught,
 * check whether it's us - if so adjust our name
 */
static void
ui_source_edited_cb (UI *ui, int id, char *name, PlaylistUI *l)
{
	if (id != l->priv->id) return;

	/* change the name */
	playlist_ui_set_name (l, name);
}

/**
 * ui_cut_cb: copy event caught, cutt all selected nodes to the clipboard
 */
static void
ui_cut_cb (UI *ui, PlaylistUI *l)
{
	GList *sel;

	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;
	
	/* get the selection */
	sel = songs_view_get_checked_nodes (l->priv->songs);

	/* and set it to the clipboard */
	source_manager_set_clipboard (l->priv->manager, sel);

	/* get rid of the nodes */
	remove_nodes (l, sel);
}

/**
 * ui_copy_cb: copy event caught, copy all selected nodes to the clipboard
 */
static void
ui_copy_cb (UI *ui, PlaylistUI *l)
{
	GList *sel;

	/* ignore if we're not active */
	if (l->priv->active == FALSE) return;
	
	/* get the selection */
	sel = songs_view_get_checked_nodes (l->priv->songs);

	/* and set it to the clipboard */
	source_manager_set_clipboard (l->priv->manager, sel);
}

/**
 * ui_paste_cb: caught a paste event, paste all nodes from the clipboard
 */
static void
ui_paste_cb (UI *ui, PlaylistUI *l)
{
	GList *clipboard;

	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;

	/* get the clipboard */
	clipboard = source_manager_get_clipboard (l->priv->manager);

	playlist_ui_add_nodes (l, clipboard);
}

/**
 * playlist_ui_add_nodes: add @nodes to @l
 */
void
playlist_ui_add_nodes (PlaylistUI *l, GList *nodes)
{
	GList *li;

	for (li = nodes; li != NULL; li = g_list_next (li))
	{
		Node *node = li->data;
		songs_view_add_node (l->priv->songs, node);
	}

	update_statusbar (l);
}

/**
 * songs_key_press_event_cb: remove selected node if keyval == delete
 */
static void
songs_row_remove_request_cb (SongsView *v, GList *nodes, PlaylistUI *l)
{
	remove_nodes (l, nodes);

	g_list_free (nodes);

	return;
}

/**
 * ui_rewind_cb: rewind signal caught, play previous song
 */
static void
ui_rewind_cb (UI *ui, PlaylistUI *l)
{
	Node *new;

	/* check that we are indeed the active source */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* get the prev row */
	new = songs_view_get_prev (l->priv->songs);
	
	/* and play it */
	state_manager_set_playing_node (l->priv->sm, new);
}

/**
 * ui_forward_cb: rewind signal caught, play next song
 */
static void
ui_forward_cb (UI *ui, PlaylistUI *l)
{
	Node *new;

	/* check that we are indeed the active source */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* get the next row */
	new = songs_view_get_next (l->priv->songs);
	
	/* and play it */
	state_manager_set_playing_node (l->priv->sm, new);
}

/**
 * songs_view_rewind_available_cb: rewind isnt possible
 */
static void
songs_view_rewind_available_cb (SongsView *v, gboolean avail, PlaylistUI *l)
{
	/* if we're not playing, ignore */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;
	
	/* update the ui */
	ui_set_can_rewind (l->priv->ui, avail);
}

/**
 * songs_view_forward_available_cb: forward isnt possible
 */
static void
songs_view_forward_available_cb (SongsView *v, gboolean avail, PlaylistUI *l)
{
	/* if we're not playing, ignore */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* update the ui */
	ui_set_can_forward (l->priv->ui, avail);
}

/**
 * songs_view_selection_available_cb: selection availability changed
 */
static void
songs_view_selection_available_cb (SongsView *v, gboolean avail, PlaylistUI *l)
{
	if (l->priv->active == FALSE) return;
	
	ui_set_can_cut (l->priv->ui, avail);
	ui_set_can_copy (l->priv->ui, avail);
}

/**
 * playlist_ui_save_to_xml: dump the playlist to a file
 */
static void
playlist_ui_save_to_xml (PlaylistUI *l, const char *fn)
{
	xmlDocPtr doc;
	GList *list;

	xmlIndentTreeOutput = TRUE;

	doc = xmlNewDoc ("1.0");
	doc->children = xmlNewDocNode (doc, NULL, "RhythmBoxPlaylist", NULL);
	xmlSetProp (doc->children, "name", l->priv->name);

	list = songs_view_get_all_nodes (l->priv->songs);
	g_list_foreach (list, (GFunc) playlist_ui_save_song_node, doc->children);

	xmlSaveFormatFile (fn, doc, 1);
}

/**
 * playlist_ui_save_song_node: save a song node to the xml file
 */
static void
playlist_ui_save_song_node (Node *song, xmlNodePtr parent_node)
{
	xmlNodePtr song_node;
	char *tmp = g_strdup_printf ("%d", node_get_int_property (song, "id"));

	song_node = xmlNewChild (parent_node, NULL, "Song", NULL);

	xmlSetProp (song_node, "id", tmp);

	g_free (tmp);
}

/**
 * playlist_ui_load_from_xml: load a playlist into @l from @fn
 */
static void
playlist_ui_load_from_xml (PlaylistUI *l, const char *fn)
{
	xmlDocPtr doc;
	xmlNodePtr child;

	doc = xmlParseFile (fn);

	if (doc == NULL) return;

	l->priv->name = xmlGetProp (doc->children, "name");
	ui_source_set_name (l->priv->ui, l->priv->id, l->priv->name);

	for (child = doc->children->children; child != NULL; child = child->next)
		playlist_ui_load_song_from_xml (l, child);

	xmlFreeDoc (doc);
}

/**
 * playlist_ui_load_song_from_xml: loads a song from the xml file
 */
static void
playlist_ui_load_song_from_xml (PlaylistUI *l, xmlNodePtr node)
{
	char *id;
	Node *song;

	id = xmlGetProp (node, "id");

	if (id == NULL) return;

	/* get the song for this ID */
	song = library_node_from_id (l->priv->library, atoi (id));

	g_free (id);

	if (song == NULL) return;

	songs_view_add_node (l->priv->songs, song);
}

/**
 * filename_from_name: makes a unique filename for @name
 */
static char *
filename_from_name (const char *name)
{
	char *tmp, *ret, *asciiname;
	int i = 0;

	if (name == NULL)
		return NULL;

	asciiname = g_filename_from_utf8 (name,
			g_utf8_strlen (name, -1),
			NULL, NULL, NULL);
	ret = NULL;

	tmp = g_strconcat (asciiname, ".rbp", NULL);

	/* continue to we got a unique name */
	while (ret == NULL)
	{
		char *tmp2 = g_build_filename (g_get_home_dir (), GNOME_DOT_GNOME,
					       "rhythmbox", tmp, NULL);
		g_free (tmp);
		if (!g_file_test (tmp2, G_FILE_TEST_EXISTS))
		{
			ret = tmp2;
		}
		else
		{
			tmp = g_strdup_printf ("%s%d.rbp", asciiname, i);
			g_free (tmp2);
		}
		i++;
	}

	g_free (asciiname);

	return ret;
}

/**
 * ui_play_cb: play signal received, set queue to be playing
 */
static void
ui_play_cb (UI *ui, PlaylistUI *l)
{
	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;

	/* start playing */
	if (state_manager_get_paused (l->priv->sm) == TRUE)
		state_manager_set_paused (l->priv->sm, FALSE);
	else
		state_manager_set_playing_node (l->priv->sm,
						songs_view_get_first (l->priv->songs));
}

/**
 * ui_stop_cb: stop signal received, stop the playing source
 */
static void
ui_stop_cb (UI *ui, PlaylistUI *l)
{
	/* if we're not playing, ignore */
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	/* stop playing */
	state_manager_set_playing_node (l->priv->sm, NULL);
	state_manager_set_active (l->priv->sm, FALSE);
}

/**
 * ui_query_cb: search entry activated, filter the list
 */
static void 
ui_query_cb (UI *ui, char *text, PlaylistUI *l)
{
	Node *search;

	if (l->priv->active == FALSE) return;
	
	/* create a filter */
	search = library_search (l->priv->library, text);

	/* run the filter */
	songs_view_filter (l->priv->songs, search, NULL);

	/* cleanup */
	g_object_unref (G_OBJECT (search));

	/* update statusbar */
	update_statusbar (l);
}

/**
 * ui_pause_cb: pause signal received, set queue to be paused
 */
static void
ui_pause_cb (UI *ui, PlaylistUI *l)
{
	if (state_manager_get_active (l->priv->sm) == FALSE) return;

	state_manager_set_paused (l->priv->sm, TRUE);
}

/**
 * ui_set_shuffle_cb: shuffle reqeuest caught, pass on to songsview
 */
static void
ui_set_shuffle_cb (UI *ui, gboolean shuffle, PlaylistUI *l)
{
	songs_view_set_shuffle (l->priv->songs, shuffle);
}

/**
 * ui_set_repeat_cb: repeat reqeuest caught, pass on to songsview
 */
static void
ui_set_repeat_cb (UI *ui, gboolean repeat, PlaylistUI *l)
{
	songs_view_set_repeat (l->priv->songs, repeat);
}

/**
 * update_statusbar: set the statusbar message
 */
static void
update_statusbar (PlaylistUI *l)
{
	char *msg;

	if (l->priv->ui == NULL) return;
	
	msg = songs_view_build_status_message (l->priv->songs);
	ui_set_status (l->priv->ui, msg);
	g_free (msg);
}

/**
 * playlist_ui_get_selection: get the checkbox selection
 */
GList *
playlist_ui_get_selection (PlaylistUI *l)
{
	return songs_view_get_checked_nodes (l->priv->songs);
}

/**
 * ui_clear_cb: clear everything
 */
static void
ui_clear_cb (UI *ui, PlaylistUI *l)
{
	GList *sel;

	/* if we're not active, ignore */
	if (l->priv->active == FALSE) return;
	
	/* get the selection */
	sel = songs_view_get_shown_nodes (l->priv->songs);

	remove_nodes (l, sel);

	g_list_free (sel);
}

/**
 * ui_select_all_cb: select everything
 */
static void
ui_select_all_cb (UI *ui, PlaylistUI *l)
{
	GList *vis;
	
	if (l->priv->active == FALSE) return;
	
	vis = songs_view_get_shown_nodes (l->priv->songs);
	songs_view_select_nodes (l->priv->songs, vis);
	g_list_free (vis);
}

/**
 * ui_select_none_cb: unselect everything
 */
static void
ui_select_none_cb (UI *ui, PlaylistUI *l)
{
	if (l->priv->active == FALSE) return;
	
	songs_view_select_nodes (l->priv->songs, NULL);
}

/**
 * remove_nodes: remove a list of nodes from the playlist
 */
static void
remove_nodes (PlaylistUI *l, GList *sel)
{
	GList *li;

	/* iterate over the selection removing the items */
	for (li = sel; li != NULL; li = g_list_next (li))
	{
		Node *node = li->data;
		if (node == state_manager_get_playing_node (l->priv->sm))
		{
			/* if this node was playing, stop that */
			state_manager_set_playing_node (l->priv->sm, NULL);
		}
		songs_view_remove_node (l->priv->songs, node);
	}

	update_statusbar (l);
}

/**
 * source_manager_clipboard_available_cb: clipboard availability changed, update paste
 * menuitem sensitivity
 */
static void
source_manager_clipboard_available_cb (SourceManager *m, gboolean available, PlaylistUI *l)
{
	if (l->priv->active == FALSE) return;
	ui_set_can_paste (l->priv->ui, available);
}

/**
 * source_manager_stop_playing_cb: stop playing
 */
static void
source_manager_stop_playing_cb (SourceManager *m, gpointer exception, PlaylistUI *l)
{
	if (exception == l) return;

	state_manager_set_active (l->priv->sm, FALSE);
}

/**
 * state_manager_start_playing_cb: forward signal
 */
static void
state_manager_start_playing_cb (StateManager *m, PlaylistUI *l)
{
	g_signal_emit (G_OBJECT (l), playlist_ui_signals[START_PLAYING], 0);
}

/**
 * state_manager_stop_playing_cb: forward signal
 */
static void
state_manager_stop_playing_cb (StateManager *m, PlaylistUI *l)
{
	g_signal_emit (G_OBJECT (l), playlist_ui_signals[STOP_PLAYING], 0);
}

/**
 * playlist_ui_set_property: set a PlaylistUI property
 */
static void
playlist_ui_set_property (GObject *object,
			  guint prop_id,
			  const GValue *value,
			  GParamSpec *pspec)
{
	PlaylistUI *l = PLAYLIST_UI (object);

	switch (prop_id)
	{
		case PROP_SOURCE_MANAGER:
			playlist_ui_set_manager (l, g_value_get_object (value));
			break;
		case PROP_FILE_NAME:
			playlist_ui_load_from_file (l, g_value_get_string (value));
			break;
		case PROP_NAME:
			playlist_ui_set_name (l, g_value_get_string (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/**
 * playlist_ui_get_property: get a PlaylistUI property
 */
static void
playlist_ui_get_property (GObject *object,
			  guint prop_id,
			  GValue *value,
			  GParamSpec *pspec)
{
	PlaylistUI *l = PLAYLIST_UI (object);

	switch (prop_id)
	{
		case PROP_SOURCE_MANAGER:
			g_value_set_object (value, l->priv->manager);
			break;
		case PROP_FILE_NAME:
			g_value_set_string (value, l->priv->file);
			break;
		case PROP_NAME:
			g_value_set_string (value, l->priv->name);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/**
 * playlist_ui_start_editing: start editing the name of this source in the sources
 * treeview
 */
void
playlist_ui_start_editing (PlaylistUI *l)
{
	ui_source_start_editing (l->priv->ui, l->priv->id);
}
