/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: songs-view.c,v 1.61 2002/04/13 21:54:29 kristian Exp $
 */

#include <libgnome/gnome-i18n.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <gtk/gtktreemodel.h>
#include <gdk/gdkkeysyms.h>
#include <stdlib.h>
#include <string.h>

#include "library-ui.h"
#include "rbtreemodelfilter.h"
#include "queue.h"
#include "node-utils.h"
#include "treeview-utils.h"
#include "songs-view.h"
#include "state.h"
#include "icons.h"
#include "eel-gconf-extensions.h"

/* columns of the list */
const char *column_strings[] = {"Song", "Time", "Artist", "Album", "Track #", "Genre", NULL};

typedef enum
{
	SONGS_COL_PIXBUF    = 0,
	SONGS_COL_SONG      = 1,
	SONGS_COL_TIME      = 2,
	SONGS_COL_ARTIST    = 3,
	SONGS_COL_ALBUM     = 4,
	SONGS_COL_TRACK     = 5,
	SONGS_COL_GENRE     = 6,
	SONGS_COL_NODE      = 7,
	SONGS_COL_PLAYING   = 8,
	SONGS_COL_VISIBLE   = 9,
	SONGS_COL_SELECTED  = 10,
	SONGS_COL_TIME_INT  = 11,
	SONGS_COL_TRACK_INT = 12,
	SONGS_COL_COUNT     = 13
} SongRow;

typedef struct
{
	gint n_songs;
	gint n_seconds;
	GnomeVFSFileSize n_bytes;
} StatusbarInfo;

struct _SongsViewPrivate
{
	GtkWidget *songs;                 /* the treeview */
	GtkListStore *songs_liststore;    /* the store */
	GHashTable *songs_node_to_iter;    /* eats nodes, shits rowiters */
	GtkTreeModel *songs_filtermodel;  /* the filtermodel */
	GtkTreeModel *songs_sortmodel;    /* the sortmodel */

	GtkWidget *scrolledwindow;

	GdkPixbuf *volume_pixbuf;         /* the pixbuf used to indicate that
					     a row is playing */
	GdkPixbuf *volume_off_pixbuf;     /* the pixbuf used to indicate that
					     a row is paused */

	GtkTreeIter *active_row;          /* the current row in the list
					     (ie the row that prev/next etc work
					      from) */
	GtkTreeIter *playing_row;         /* the row in the list that is currently
					     playing */
	
	GList *columns;                   /* list of columns in the insertion
					     order, which is also the same
					     order as the preference */

	gint n_rows;                      /* numer of items in the view */

	gboolean repeat;                  /* repeat mode ? */
	gboolean shuffle;                 /* shuffle mode ? */

	/* sort order data */
	GList *album_sort_order;
	GList *artist_sort_order;
	GList *track_sort_order;
	GList *genre_sort_order;

	GHashTable *checked_nodes;        /* hash of currently checked nodes */
};

/* all the different signals */
enum
{
	ROW_ACTIVATED,
	ROW_REMOVE_REQUEST,
	REWIND_AVAILABLE,
	FORWARD_AVAILABLE,
	SELECTION_AVAILABLE,
	LAST_SIGNAL
};

typedef enum
{
	ITER_VALIDATION_VISIBLE = 1 << 0,
	ITER_VALIDATION_CHECKED = 1 << 1
} IterValidationFlags;

/* globals */
static GObjectClass *parent_class = NULL;

static guint songs_view_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void songs_view_class_init (SongsViewClass *klass);
static void songs_view_init (SongsView *l);
static void songs_view_finalize (GObject *object);

/* signal prototypes */
static void songs_row_activated_cb (GtkTreeView *tree_view, GtkTreePath *path,
			            GtkTreeViewColumn *column, SongsView *l);
static gboolean songs_key_press_event_cb (GtkWidget *widget, GdkEventKey *event,
					  SongsView *l);
static void row_toggled_cb (GtkCellRendererToggle *toggle, const gchar *path_str,
			    SongsView *v);
static void pref_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry,
	                  SongsView *v);
static void row_changed_cb (GtkTreeModel *model, GtkTreePath *start_path, GtkTreeIter *start_iter, SongsView *v);

/* local function prototypes */
static gboolean get_selection (GtkTreeModel *model, GtkTreePath *path,
			       GtkTreeIter *iter, void **data);
static gboolean get_all (GtkTreeModel *model, GtkTreePath *path,
			 GtkTreeIter *iter, void **data);
static gboolean get_all_visible (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data);
static void songs_view_update_node_internal (SongsView *l, GtkTreeIter *iter, Node *node);
static void check_pref (SongsView *v);
static gint compare_func (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			  gpointer user_data);
static gint compare_func_internal (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
				   gint column);
static Node *songs_view_get_random (SongsView *l);
static gboolean gather_stats (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
			      StatusbarInfo *info);
static gboolean set_checked (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, GList *nodes);
static void get_real_selection (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data);
static gboolean songs_view_get_real_next (SongsView *l, GtkTreeIter *start);
static gboolean songs_view_get_real_prev (SongsView *l, GtkTreeIter *start);
static gboolean songs_view_get_real_first (SongsView *l, GtkTreeIter *iter);
static gboolean songs_view_get_real_selected (SongsView *l, GtkTreeIter *selected);
static gboolean songs_view_iter_is_valid (SongsView *l, GtkTreeIter *iter, IterValidationFlags flags);
static void songs_view_store_iter_to_sort_iter (SongsView *l, GtkTreeIter *converted, GtkTreeIter *in);
static Node *songs_view_iter_to_node (SongsView *l, GtkTreeIter *iter);
static gboolean songs_view_iter_in_sort_model (SongsView *l,
					       GtkTreeIter *iter,
					       GtkTreeIter *converted);

/**
 * songs_view_get_type: get the GObject type of the SongsView
 */
GType
songs_view_get_type (void)
{
	static GType songs_view_type = 0;

  	if (songs_view_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (SongsViewClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) songs_view_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (SongsView),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) songs_view_init
      		};

      		songs_view_type = g_type_register_static (GTK_TYPE_HBOX,
                				          "SongsView",
                                           	          &our_info, 0);
    	}

	return songs_view_type;
}

/**
 * songs_view_class_init: initialize the SongsView class
 */
static void
songs_view_class_init (SongsViewClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = songs_view_finalize;

	/* init signals */
	songs_view_signals[ROW_ACTIVATED] =
   		g_signal_new ("row_activated",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SongsViewClass, row_activated),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__OBJECT,
			      G_TYPE_NONE,
			      1,
			      node_get_type ());
	songs_view_signals[ROW_REMOVE_REQUEST] =
   		g_signal_new ("row_remove_request",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SongsViewClass, row_remove_request),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__POINTER,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_POINTER);
	songs_view_signals[REWIND_AVAILABLE] =
   		g_signal_new ("rewind_available",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SongsViewClass, rewind_available),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	songs_view_signals[FORWARD_AVAILABLE] =
   		g_signal_new ("forward_available",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SongsViewClass, forward_available),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	songs_view_signals[SELECTION_AVAILABLE] =
   		g_signal_new ("selection_available",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SongsViewClass, selection_available),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
}

/**
 * songs_view_init: intialize the SongsView object
 */
static void
songs_view_init (SongsView *l)
{
	GtkTreeView *current;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;

	l->priv = g_new (SongsViewPrivate, 1);

	/* nullify */
	l->priv->active_row = NULL;
	l->priv->playing_row = NULL;
	l->priv->columns = NULL;
	l->priv->n_rows = 0;
	l->priv->repeat = FALSE;
	l->priv->shuffle = FALSE;
	l->priv->checked_nodes = g_hash_table_new (NULL, NULL);

	/* setup the scrolledwindow */
	l->priv->scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (l->priv->scrolledwindow),
					     GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (l->priv->scrolledwindow),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

	/* load pixbufs */
	l->priv->volume_pixbuf = gtk_widget_render_icon (GTK_WIDGET (l),
							 RHYTHMBOX_STOCK_VOLUME,
							 GTK_ICON_SIZE_MENU, NULL);
	l->priv->volume_off_pixbuf = gtk_widget_render_icon (GTK_WIDGET (l),
							     RHYTHMBOX_STOCK_VOLUME_OFF,
							     GTK_ICON_SIZE_MENU, NULL);

	/* create the treeview and pack it into the scrollewindow */
	l->priv->songs = gtk_tree_view_new ();
	gtk_container_add (GTK_CONTAINER (l->priv->scrolledwindow), l->priv->songs);

	/* la la la */
	current = GTK_TREE_VIEW (l->priv->songs);

	/* create the store */
	l->priv->songs_liststore = gtk_list_store_new (SONGS_COL_COUNT,
				    		       GDK_TYPE_PIXBUF,
						       G_TYPE_STRING,
						       G_TYPE_STRING,
						       G_TYPE_STRING,
						       G_TYPE_STRING,
						       G_TYPE_STRING,
						       G_TYPE_STRING,
						       G_TYPE_POINTER,
						       G_TYPE_BOOLEAN,
						       G_TYPE_BOOLEAN,
						       G_TYPE_BOOLEAN,
						       G_TYPE_INT,
						       G_TYPE_INT);

	/* connect row_changed handler */
	g_signal_connect (G_OBJECT (l->priv->songs_liststore), "row_changed",
                          G_CALLBACK (row_changed_cb), l);

	/* create the node-eating-ref-shitting hash */
	l->priv->songs_node_to_iter = g_hash_table_new (g_direct_hash, g_direct_equal);

	/* create the filter & sortmodels */
	l->priv->songs_filtermodel = rb_tree_model_filter_new_with_model 
		(GTK_TREE_MODEL (l->priv->songs_liststore), SONGS_COL_VISIBLE, NULL);
	l->priv->songs_sortmodel = gtk_tree_model_sort_new_with_model
		(l->priv->songs_filtermodel);

	/* we do Smart Sorting[tm], for example if albums are the same, look
	 * for track order, etc */
	l->priv->album_sort_order = NULL;
	l->priv->album_sort_order =
		g_list_append (l->priv->album_sort_order, GINT_TO_POINTER (SONGS_COL_ALBUM));
	l->priv->album_sort_order =
		g_list_append (l->priv->album_sort_order, GINT_TO_POINTER (SONGS_COL_TRACK_INT));
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (l->priv->songs_sortmodel),
					 SONGS_COL_ALBUM, compare_func,
					 l->priv->album_sort_order, NULL);
	
	l->priv->artist_sort_order = NULL;
	l->priv->artist_sort_order =
		g_list_append (l->priv->artist_sort_order, GINT_TO_POINTER (SONGS_COL_ARTIST));
	l->priv->artist_sort_order =
		g_list_append (l->priv->artist_sort_order, GINT_TO_POINTER (SONGS_COL_ALBUM));
	l->priv->artist_sort_order =
		g_list_append (l->priv->artist_sort_order, GINT_TO_POINTER (SONGS_COL_TRACK_INT));
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (l->priv->songs_sortmodel),
					 SONGS_COL_ARTIST, compare_func,
					 l->priv->artist_sort_order, NULL);
	
	l->priv->track_sort_order = NULL;
	l->priv->track_sort_order =
		g_list_append (l->priv->track_sort_order, GINT_TO_POINTER (SONGS_COL_TRACK_INT));
	l->priv->track_sort_order =
		g_list_append (l->priv->track_sort_order, GINT_TO_POINTER (SONGS_COL_ALBUM));
	l->priv->track_sort_order =
		g_list_append (l->priv->track_sort_order, GINT_TO_POINTER (SONGS_COL_ARTIST));
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (l->priv->songs_sortmodel),
					 SONGS_COL_TRACK_INT, compare_func,
					 l->priv->track_sort_order, NULL);
	
	l->priv->genre_sort_order = NULL;
	l->priv->genre_sort_order =
		g_list_append (l->priv->genre_sort_order, GINT_TO_POINTER (SONGS_COL_GENRE));
	l->priv->genre_sort_order =
		g_list_append (l->priv->genre_sort_order, GINT_TO_POINTER (SONGS_COL_ARTIST));
	l->priv->genre_sort_order =
		g_list_append (l->priv->genre_sort_order, GINT_TO_POINTER (SONGS_COL_ALBUM));
	l->priv->genre_sort_order =
		g_list_append (l->priv->genre_sort_order, GINT_TO_POINTER (SONGS_COL_TRACK_INT));
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (l->priv->songs_sortmodel),
					 SONGS_COL_GENRE, compare_func,
					 l->priv->genre_sort_order, NULL);

	/* load model into view */
	gtk_tree_view_set_model (current, l->priv->songs_sortmodel);

	/* set view properties */
	gtk_tree_view_set_rules_hint (current, TRUE);
	gtk_tree_view_set_headers_visible (current, TRUE);

	/* playing state column */
	renderer = gtk_cell_renderer_pixbuf_new ();
	gtk_tree_view_insert_column_with_attributes (current,
						     SONGS_COL_PIXBUF, "",
						     renderer,
						     "pixbuf", SONGS_COL_PIXBUF,
						     NULL);
	column = gtk_tree_view_get_column (current, SONGS_COL_PIXBUF);
	gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_FIXED);
	gtk_tree_view_column_set_fixed_width (column,
				      gdk_pixbuf_get_width (l->priv->volume_pixbuf) + 5);
	gtk_tree_view_column_set_reorderable (column, TRUE);

	/* selection + song title column */
	column = gtk_tree_view_column_new ();
	l->priv->columns = g_list_append (l->priv->columns, column);
	gtk_tree_view_column_set_title (column, _(column_strings[0]));
	renderer = gtk_cell_renderer_toggle_new ();
	gtk_tree_view_column_pack_start (column, renderer, FALSE);
	gtk_tree_view_column_add_attribute (column, renderer,
					    "active", SONGS_COL_SELECTED);
	g_signal_connect (G_OBJECT (renderer), "toggled",
			  G_CALLBACK (row_toggled_cb), l);
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_column_pack_start (column, renderer, FALSE);
	gtk_tree_view_column_add_attribute (column, renderer,
					    "text", SONGS_COL_SONG);
	gtk_tree_view_column_set_resizable (column, TRUE);
	gtk_tree_view_column_set_reorderable (column, TRUE);
	gtk_tree_view_column_set_sort_column_id (column, SONGS_COL_SONG);
	gtk_tree_view_append_column (current, column);

	/* playing time column */
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (current,
						     SONGS_COL_TIME, _(column_strings[1]),
						     renderer,
						     "text", SONGS_COL_TIME,
						     NULL);
	column = gtk_tree_view_get_column (current, SONGS_COL_TIME);
	l->priv->columns = g_list_append (l->priv->columns, column);
	gtk_tree_view_column_set_resizable (column, TRUE);
	gtk_tree_view_column_set_reorderable (column, TRUE);
	gtk_tree_view_column_set_sort_column_id (column, SONGS_COL_TIME_INT);

	/* artist column */
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (current,
						     SONGS_COL_ARTIST, _(column_strings[2]),
						     renderer,
						     "text", SONGS_COL_ARTIST,
						     NULL);
	column = gtk_tree_view_get_column (current, SONGS_COL_ARTIST);
	l->priv->columns = g_list_append (l->priv->columns, column);
	gtk_tree_view_column_set_resizable (column, TRUE);
	gtk_tree_view_column_set_reorderable (column, TRUE);
	gtk_tree_view_column_set_sort_column_id (column, SONGS_COL_ARTIST);

	/* album column */
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (current,
						     SONGS_COL_ALBUM, _(column_strings[3]),
						     renderer,
						     "text", SONGS_COL_ALBUM,
						     NULL);
	column = gtk_tree_view_get_column (current, SONGS_COL_ALBUM);
	l->priv->columns = g_list_append (l->priv->columns, column);
	gtk_tree_view_column_set_resizable (column, TRUE);
	gtk_tree_view_column_set_reorderable (column, TRUE);
	gtk_tree_view_column_set_sort_column_id (column, SONGS_COL_ALBUM);

	/* track column */
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (current,
						     SONGS_COL_TRACK, _(column_strings[4]),
						     renderer,
						     "text", SONGS_COL_TRACK,
						     NULL);
	column = gtk_tree_view_get_column (current, SONGS_COL_TRACK);
	l->priv->columns = g_list_append (l->priv->columns, column);
	gtk_tree_view_column_set_resizable (column, TRUE);
	gtk_tree_view_column_set_reorderable (column, TRUE);
	gtk_tree_view_column_set_sort_column_id (column, SONGS_COL_TRACK_INT);

	/* genre column */
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (current,
						     SONGS_COL_GENRE, _(column_strings[5]),
						     renderer,
						     "text", SONGS_COL_GENRE,
						     NULL);
	column = gtk_tree_view_get_column (current, SONGS_COL_GENRE);
	l->priv->columns = g_list_append (l->priv->columns, column);
	gtk_tree_view_column_set_resizable (column, TRUE);
	gtk_tree_view_column_set_reorderable (column, TRUE);
	gtk_tree_view_column_set_sort_column_id (column, SONGS_COL_GENRE);

	/* set selection mode */
	selection = gtk_tree_view_get_selection (current);
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);

	/* connect signals */
	g_signal_connect (G_OBJECT (l->priv->songs), "key_press_event",
			  G_CALLBACK (songs_key_press_event_cb), l);
	g_signal_connect (G_OBJECT (l->priv->songs), "row_activated",
			  G_CALLBACK (songs_row_activated_cb), l);

	/* install notifiers */
	eel_gconf_notification_add (CONF_SONGS_VIEW_COLUMNS,
				    (GConfClientNotifyFunc) pref_changed, l);

	/* load states */
	check_pref (l);
	state_load_treeview_layout (l->priv->songs, "SongsView");

	/* and we add the scrolledwindow to oursefl */
	gtk_widget_show_all (l->priv->scrolledwindow);
	gtk_container_add (GTK_CONTAINER (l), l->priv->scrolledwindow);
	gtk_container_set_border_width (GTK_CONTAINER (l), 5);
}

/**
 * songs_view_finalize: finalize the SongsView object
 */
static void
songs_view_finalize (GObject *object)
{
	SongsView *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_SONGS_VIEW (object));
	
   	s = SONGS_VIEW (object);

	g_return_if_fail (s->priv != NULL);

	/* save state */
	state_save_treeview_layout (s->priv->songs, "SongsView");

	/* free hashtable */
	free_node_to_iter_hash (s->priv->songs_node_to_iter);

	/* unref the pixbuf */
	g_object_unref (G_OBJECT (s->priv->volume_pixbuf));

	/* free columns list */
	g_list_free (s->priv->columns);

	/* free sorting data */
	g_list_free (s->priv->album_sort_order);
	g_list_free (s->priv->artist_sort_order);
	g_list_free (s->priv->track_sort_order);
	g_list_free (s->priv->genre_sort_order);

	/* hash of checked nodes */
	g_hash_table_destroy (s->priv->checked_nodes);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * songs_view_new: create a new SongsView object
 */
GtkWidget *
songs_view_new (void)
{
 	SongsView *l;

	l = SONGS_VIEW (g_object_new (TYPE_SONGS_VIEW, NULL));

	g_return_val_if_fail (l->priv != NULL, NULL);

	return GTK_WIDGET (l);
}

/**
 * songs_view_add_node: Add a song to the list
 */
void
songs_view_add_node (SongsView *l, Node *node)
{
	GtkTreeIter *iter;

	/* no duplicates please */
	if (g_hash_table_lookup (l->priv->songs_node_to_iter, node) != NULL)
		return; 

	l->priv->n_rows++;

	/* set properties into the row */
	iter = g_new0 (GtkTreeIter, 1);
	gtk_list_store_append (l->priv->songs_liststore, iter);
	gtk_list_store_set (l->priv->songs_liststore, iter,
			    SONGS_COL_PIXBUF,   NULL,
			    SONGS_COL_NODE,     node,
			    SONGS_COL_PLAYING,  FALSE,
			    SONGS_COL_VISIBLE,  TRUE,
			    SONGS_COL_SELECTED, TRUE,
			    -1);

	/* make this the active row if there's none set */
	if (l->priv->active_row == NULL)
		l->priv->active_row = iter;

	/* insert into the hashtable */
	g_hash_table_insert (l->priv->songs_node_to_iter, node, iter);

	/* set the song info into the row */
	songs_view_update_node_internal (l, iter, node);

	/* check state */
	songs_view_check_state (l);

	return;
}

/**
 * songs_view_remove_node: Remov a song from the SongsView
 */
void
songs_view_remove_node (SongsView *l, Node *node)
{
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->songs_node_to_iter, node);
	GtkTreePath *path;

	if (iter == NULL) return;
	
	l->priv->n_rows--;

	/* if we are removing the active row, we want to set a new active row */
	if (iter == l->priv->active_row)
	{
		Node *newnode = songs_view_get_next (l);
		if (newnode == NULL)
		{
			newnode = songs_view_get_prev (l);
		}
		if (newnode != NULL)
		{
			GtkTreeIter *iter2 =
				g_hash_table_lookup (l->priv->songs_node_to_iter, newnode);
			l->priv->active_row = iter2;
		}
		else
			l->priv->active_row = NULL;
		l->priv->playing_row = NULL;
	}

	/* remove from the hashtable */
	g_hash_table_remove (l->priv->songs_node_to_iter, node);

	/* store path */
	path = gtk_tree_model_get_path (GTK_TREE_MODEL (l->priv->songs_liststore), iter);

	/* update checked_dounter */
	g_hash_table_remove (l->priv->checked_nodes, node);
	
	/* remove from the store */
	gtk_list_store_remove (l->priv->songs_liststore, iter);

	/* free iter */
	gtk_tree_iter_free (iter);

	/* check rewind/forward sensitivity */
	songs_view_check_state (l);
}

/**
 * songs_view_filter: filter model according to filter
 */
void
songs_view_filter (SongsView *l, Node *node, Node *grandparent)
{
	GList *li;
	GList *iters = NULL;

	/* get to be filtered gtktreerowreferences */
	for (li = node_get_children (node); li != NULL && li->data != NULL; li = g_list_next (li))
	{
		Node *node = li->data;
		GtkTreeIter *iter;

		/* check the grandparent */
		if (grandparent != NULL)
			if (node_get_grandparent (node) != grandparent)
				continue;

		/* get the iter for this node */
		iter = g_hash_table_lookup (l->priv->songs_node_to_iter, node);
		if (iter == NULL)
			continue;

		iters = g_list_append (iters, iter);
	}

	/* do the actual filtering */
	rb_tree_model_filter_show_iters (RB_TREE_MODEL_FILTER (l->priv->songs_filtermodel), iters);

	g_list_free (iters);

	/* check rewind/forward sensitivity with the new list setup */
	songs_view_check_state (l);
}

/**
 * songs_row_activated_cb: row in the songs list doubleclicked (activated)
 */
static void
songs_row_activated_cb (GtkTreeView *tree_view, GtkTreePath *path,
			GtkTreeViewColumn *column, SongsView *l)
{
	GtkTreeIter iter;
	GtkTreeModel *model = gtk_tree_view_get_model (tree_view);
	GValue val = {0, };
	Node *node;

	/* get the node associated with this row */
	gtk_tree_model_get_iter (model, &iter, path);
	gtk_tree_model_get_value (model, &iter, SONGS_COL_NODE, &val);
	node = g_value_get_pointer (&val);

	g_signal_emit (G_OBJECT (l), songs_view_signals[ROW_ACTIVATED], 0, node);
}

/**
 * songs_key_press_event_cb: remove selected node if keyval == delete
 */
static gboolean
songs_key_press_event_cb (GtkWidget *widget, GdkEventKey *event, SongsView *l)
{
	GList *sel;

	if (event->keyval != GDK_Delete) return FALSE;

	sel = songs_view_get_selected_nodes (l);

	g_signal_emit (G_OBJECT (l), songs_view_signals[ROW_REMOVE_REQUEST], 0, sel);

	return TRUE;
}

/**
 * songs_view_get_next: return the node after the currently active node
 */
Node *
songs_view_get_next (SongsView *l)
{
	GtkTreeIter iter;

	/* if shuffle, get a random node */
	if (l->priv->shuffle == TRUE)
	{
		return songs_view_get_random (l);
	}

	g_assert (l->priv->active_row != NULL);

	/* don't look for a next if the current row isn't visible */
	if (!songs_view_iter_in_sort_model (l, l->priv->active_row, &iter))
		return NULL;

	if (songs_view_iter_is_valid (l, &iter,
				      ITER_VALIDATION_VISIBLE) == FALSE)
		return NULL;

	/* get next */
	if (!songs_view_get_real_next (l, &iter))
		if (l->priv->repeat && !songs_view_get_real_first (l, &iter))
			return NULL;

	do
	{
		if (songs_view_iter_is_valid (l, &iter, ITER_VALIDATION_VISIBLE | ITER_VALIDATION_CHECKED))
			return songs_view_iter_to_node (l, &iter);
	}
	while (songs_view_get_real_next (l, &iter));

	return NULL;
}

/**
 * songs_view_get_prev: return the node before the currently active node
 */
Node *
songs_view_get_prev (SongsView *l)
{
	GtkTreeIter iter;

	/* if shuffle, get a random node */
	if (l->priv->shuffle == TRUE)
	{
		return songs_view_get_random (l);
	}

	g_assert (l->priv->active_row != NULL);

	/* don't look for a prev if the current row isn't visible */
	if (!songs_view_iter_in_sort_model (l, l->priv->active_row, &iter))
		return NULL;
	if (songs_view_iter_is_valid (l, &iter,
				      ITER_VALIDATION_VISIBLE) == FALSE)
		return NULL;

	/* get prev */
	if (songs_view_get_real_prev (l, &iter))
		return songs_view_iter_to_node (l, &iter);

	return NULL;
}

/**
 * songs_view_set_node_state: set the playing state of @node
 */
void
songs_view_set_node_state (SongsView *l, Node *node, SongsViewNodeState state)
{
	GdkPixbuf *pixbuf = NULL;

	/* if node is NULL, do not update any row */
	if (node == NULL) return;

	/* decide what image to use */
	switch (state)
	{
	case SONGS_VIEW_NODE_STATE_PLAYING:
		pixbuf = l->priv->volume_pixbuf;
		l->priv->active_row = l->priv->playing_row =
			g_hash_table_lookup (l->priv->songs_node_to_iter, node);
		break;
	case SONGS_VIEW_NODE_STATE_PAUSED:
		pixbuf = l->priv->volume_off_pixbuf;
		l->priv->active_row = l->priv->playing_row =
			g_hash_table_lookup (l->priv->songs_node_to_iter, node);
		break;
	case SONGS_VIEW_NODE_STATE_STOPPED:
		pixbuf = NULL;
		l->priv->playing_row = NULL;
		break;
	default:
		g_warning ("Should never get here!\n");
		break;
	}

	g_assert (l->priv->active_row != NULL);

	/* put image */
	gtk_list_store_set (l->priv->songs_liststore, l->priv->active_row,
			    SONGS_COL_PIXBUF, pixbuf, -1);

	/* deselect if we're stopping the node and we're not in repeat mode */
	if (state == SONGS_VIEW_NODE_STATE_STOPPED && l->priv->repeat == FALSE)
	{
		gtk_list_store_set (l->priv->songs_liststore, l->priv->active_row,
				    SONGS_COL_SELECTED, FALSE,
				    -1);
	}

	songs_view_check_state (l);
}

/**
 * songs_view_update_node: set row properties according to @node
 */
void
songs_view_update_node (SongsView *l, Node *node)
{
	GtkTreeIter *iter = g_hash_table_lookup (l->priv->songs_node_to_iter, node);

	songs_view_update_node_internal (l, iter, node);
}

/**
 * songs_view_update_node_internal: update @iter to contain info from @node
 */
static void
songs_view_update_node_internal (SongsView *l, GtkTreeIter *iter, Node *node)
{
	int totalseconds = node_get_int_property (node, SONG_PROPERTY_LENGTH);
	int minutes = totalseconds / 60;
	int seconds = totalseconds % 60;
	char *time = g_strdup_printf ("%d:%.2d", minutes, seconds);
	const char *track = node_get_string_property (node, SONG_PROPERTY_TRACK_NUMBER);
	const char *genre = node_get_string_property (node, SONG_PROPERTY_GENRE);
	char *song = g_strdup (node_get_string_property (node, NODE_PROPERTY_NAME));

	int track_int;
	char *tmp;
	char *pretty_track = NULL;
	char **parts;
	int i;
	
	Node *artist_node = song_get_artist (node);
	Node *album_node = song_get_album (node);

	GnomeVFSURI *uri;

	const char *artist = NULL;
	const char *album = NULL;

	/* get the track sorting int */
	if (track != NULL && strstr (track, "/") != NULL)
		tmp = g_strndup (track, strlen (track) - strlen (strstr (track, "/")));
	else
		tmp = g_strdup (track);
	track_int = tmp != NULL ? atoi (tmp) : 0;
	g_free (tmp);

	/* make a pretty track string */
	parts = track != NULL ? g_strsplit (track, "/", 0) : NULL; 

	for (i = 0; parts != NULL && parts[i] != NULL && i < 2; i++)
	{
		if (pretty_track == NULL)
			pretty_track = g_strdup_printf ("%.2d", atoi (parts[i]));
		else
			pretty_track = g_strdup_printf ("%s of %.2d", pretty_track,
							atoi (parts[i]));
	}

	g_strfreev (parts);

	/* get parent properties */
	if (artist_node != NULL) 
		artist = node_get_string_property (artist_node, NODE_PROPERTY_NAME);
	else 
		artist = "";

	if (album_node != NULL)
		album = node_get_string_property (album_node, NODE_PROPERTY_NAME);
	else
		album = "";

	if ((song == NULL) || (song[0] == '\0'))
	{
		/* there's not a useful title for this song, use the filename */
		uri = gnome_vfs_uri_new (node_get_string_property (node, SONG_PROPERTY_URI));
		song = gnome_vfs_uri_extract_short_name (uri);
		gnome_vfs_uri_unref (uri);
	}

	/* set properties into the row */
	gtk_list_store_set (l->priv->songs_liststore, iter,
			    SONGS_COL_SONG,      song,
			    SONGS_COL_TIME,      time,
			    SONGS_COL_ARTIST,    artist,
			    SONGS_COL_ALBUM,     album,
			    SONGS_COL_TRACK,     pretty_track,
			    SONGS_COL_TRACK_INT, track_int,
			    SONGS_COL_GENRE,     genre,
			    SONGS_COL_TIME_INT,  totalseconds,
			    -1);

	/* free the temporarily created time string */
	g_free (time);
	/* free the title string, since it might have been an escaped URI */
	g_free (song);
	g_free (pretty_track);
}

/**
 * songs_view_get_checked_nodes: return a GList containing the list's selection
 */
GList *
songs_view_get_checked_nodes (SongsView *l)
{
	GList *list = NULL;
	gtk_tree_model_foreach (GTK_TREE_MODEL (l->priv->songs_sortmodel),
			        (GtkTreeModelForeachFunc) get_selection,
				(void **) &list);
	return list;
}

/**
 * songs_view_get_all_nodes: get all nodes in from this view
 */
GList *
songs_view_get_all_nodes (SongsView *l)
{
	GList *list = NULL;
	gtk_tree_model_foreach (GTK_TREE_MODEL (l->priv->songs_liststore),
			        (GtkTreeModelForeachFunc) get_all,
				(void **) &list);
	return list;
}

/**
 * songs_view_get_shown_nodes: get all visible nodes in from this view
 */
GList *
songs_view_get_shown_nodes (SongsView *l)
{
	GList *list = NULL;
	gtk_tree_model_foreach (GTK_TREE_MODEL (l->priv->songs_liststore),
			        (GtkTreeModelForeachFunc) get_all_visible,
				(void **) &list);
	return list;
}

/**
 * songs_view_check_state: check whether there are any rows before or after
 * the current row, if any change in availability occured emit signals
 */
void
songs_view_check_state (SongsView *l)
{
	gboolean prev, next, empty;
	
	empty = !(g_hash_table_size (l->priv->checked_nodes) > 0);

	/* we trick get_prev/next into thinking we are in normal mode
	 * to get a normal guess */
	if (l->priv->playing_row != NULL)
	{
		if (empty == TRUE)
			next = prev = FALSE;
		else
		{
			next = songs_view_get_next (l) != NULL;
			prev = songs_view_get_prev (l) != NULL;
		}
	}
	else
		prev = next = FALSE;
		
	g_signal_emit (G_OBJECT (l), songs_view_signals[REWIND_AVAILABLE], 0, prev);
	g_signal_emit (G_OBJECT (l), songs_view_signals[FORWARD_AVAILABLE], 0, next);

	g_signal_emit (G_OBJECT (l), songs_view_signals[SELECTION_AVAILABLE],
		       0, !empty);
}

/**
 * row_toggled_cb: row toggled, update liststore
 */
static void
row_toggled_cb (GtkCellRendererToggle *toggle, const gchar *path_str, SongsView *v)
{
	GtkTreeIter iter;
	GtkTreePath *path = gtk_tree_path_new_from_string (path_str);

	/* get selected iter */
	gtk_tree_model_get_iter (GTK_TREE_MODEL (v->priv->songs_sortmodel), &iter, path);
	gtk_tree_model_sort_convert_iter_to_child_iter
		(GTK_TREE_MODEL_SORT (v->priv->songs_sortmodel), &iter, &iter);
	gtk_tree_model_filter_convert_iter_to_child_iter
		(GTK_TREE_MODEL_FILTER (v->priv->songs_filtermodel), &iter, &iter);
	
	/* toggle the toggle */
	gtk_list_store_set (v->priv->songs_liststore, &iter,
			    SONGS_COL_SELECTED, !toggle->active,
			    -1);

	gtk_tree_path_free (path);

	/* check rewind/forward sensitivity */
	songs_view_check_state (v);
}

/**
 * get_selection: helper func for getting all visible & selected nodes
 */
static gboolean
get_selection (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data)
{
	GList **ret = (GList **) data;
	GValue val = {0, };

	/* check whether this thing is selected */
	gtk_tree_model_get_value (model, iter, SONGS_COL_SELECTED, &val);
	if (!g_value_get_boolean (&val)) return FALSE;

	g_value_unset (&val);

	/* get the node */
	gtk_tree_model_get_value (model, iter, SONGS_COL_NODE, &val);
	
	/* append to the list */
	*ret = g_list_append (*ret, g_value_get_pointer (&val));

	/* go on */
	return FALSE;
}

/**
 * get_all: helper func for getting all nodes
 */
static gboolean
get_all (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data)
{
	GList **ret = (GList **) data;
	GValue val = {0, };

	/* get the node */
	gtk_tree_model_get_value (model, iter, SONGS_COL_NODE, &val);
	
	/* append it to the list */
	*ret = g_list_append (*ret, g_value_get_pointer (&val));

	/* go on */
	return FALSE;
}

/**
 * get_all_visible: helper func for getting all visible nodes
 */
static gboolean
get_all_visible (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data)
{
	GList **ret = (GList **) data;
	GValue val = {0, };

	/* check visibility */
	gtk_tree_model_get_value (model, iter, SONGS_COL_VISIBLE, &val);
	if (!g_value_get_boolean (&val)) return FALSE;

	g_value_unset (&val);

	/* get the node */
	gtk_tree_model_get_value (model, iter, SONGS_COL_NODE, &val);
	
	/* append it to the list */
	*ret = g_list_append (*ret, g_value_get_pointer (&val));

	/* go on */
	return FALSE;
}

/**
 * songs_view_get_first: return the first visible & selected node
 */
Node *
songs_view_get_first (SongsView *l)
{
	GtkTreeIter iter;

	/* if shuffle, get a random node */
	if (l->priv->shuffle == TRUE)
	{
		return songs_view_get_random (l);
	}

	/* get first */
	if (!songs_view_get_real_selected (l, &iter))
		if (!songs_view_get_real_first (l, &iter))
			return NULL;

	do
	{
		if (songs_view_iter_is_valid (l, &iter, ITER_VALIDATION_VISIBLE | ITER_VALIDATION_CHECKED))
			return songs_view_iter_to_node (l, &iter);
	}
	while (songs_view_get_real_next (l, &iter));

	return NULL;
}

/**
 * pref_changed: gconf pref was changed, update ui
 */
static void
pref_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry,
	      SongsView *v)
{
	check_pref (v);
}

/**
 * check_pref: apply pref to view
 */
static void
check_pref (SongsView *v)
{
	GSList *list = eel_gconf_get_integer_list (CONF_SONGS_VIEW_COLUMNS), *l;
	gint i;

	for (l = list, i = 0; l != NULL; l = g_slist_next (l), i++)
	{
		GtkTreeViewColumn *column = (g_list_nth (v->priv->columns, i))->data;

		/* show or hide column */
		gtk_tree_view_column_set_visible (column, GPOINTER_TO_INT (l->data));
	}
}

/**
 * songs_view_num_rows: get number of rows in the songsview
 */
gint
songs_view_num_rows (SongsView *v)
{
	if (v->priv->n_rows == 0) return 0;

	return gtk_tree_model_iter_n_children (v->priv->songs_filtermodel,
					       NULL);
}

/**
 * compare_func: songsview treemodel sorting func, it tries to be smart by
 * applying "priority chains" for sorting.
 */
static gint
compare_func (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer user_data)
{
	GList *order = (GList *) user_data;
	GList *l;
	gint retval = 0;

	for (l = order; l != NULL && retval == 0; l = g_list_next (l))
	{
		gint column = GPOINTER_TO_INT (l->data);

		retval = compare_func_internal (model, a, b, column);
	}

	return retval;
}

/**
 * compare_func_internal: compare_func helper
 */
static gint
compare_func_internal (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gint column)
{
	GType type = gtk_tree_model_get_column_type (model, column);
	GValue a_value = {0, };
	GValue b_value = {0, };
	gint retval;
	const gchar *stra, *strb;

	gtk_tree_model_get_value (model, a, column, &a_value);
	gtk_tree_model_get_value (model, b, column, &b_value);

	switch (G_TYPE_FUNDAMENTAL (type))
	{
	case G_TYPE_BOOLEAN:
		if (g_value_get_int (&a_value) < g_value_get_int (&b_value))
			retval = -1;
      		else if (g_value_get_int (&a_value) == g_value_get_int (&b_value))
			retval = 0;
      		else
			retval = 1;
      		break;
    	case G_TYPE_INT:
		if (g_value_get_int (&a_value) < g_value_get_int (&b_value))
			retval = -1;
  		else if (g_value_get_int (&a_value) == g_value_get_int (&b_value))
			retval = 0;
		else
			retval = 1;
      		break;
	case G_TYPE_STRING:
		stra = g_value_get_string (&a_value);
		strb = g_value_get_string (&b_value);
		if (stra == NULL) stra = "";
		if (strb == NULL) strb = "";
		retval = g_utf8_collate (stra, strb);
		break;
 	case G_TYPE_POINTER:
	case G_TYPE_BOXED:
	case G_TYPE_OBJECT:
	default:
		g_warning ("Attempting to sort on invalid type %s\n", g_type_name (type));
		retval = FALSE;
		break;
	}

	g_value_unset (&a_value);
	g_value_unset (&b_value);

	return retval;
}

/**
 * songs_view_set_repeat: set repeat mode on or off
 */
void
songs_view_set_repeat (SongsView *l, gboolean repeat)
{
	l->priv->repeat = repeat;

	songs_view_check_state (l);
}

/**
 * songs_view_set_shuffle: set shuffle mode on or off
 */
void
songs_view_set_shuffle (SongsView *l, gboolean shuffle)
{
	l->priv->shuffle = shuffle;

	songs_view_check_state (l);
}

/**
 * songs_view_get_random: get a random visible node
 */
static Node *
songs_view_get_random (SongsView *l)
{
	gint index;
	gchar *path_str;
	GtkTreePath *path;
	GtkTreeIter iter;
	Node *node = NULL;
	GHashTable *previous_paths;

	if (songs_view_num_rows (l) == 0) return NULL;
	
	/* hashtable of already-checked rows */
	previous_paths = g_hash_table_new (NULL, NULL);

	while (node == NULL)
	{
		/* we had all items ? */
		if (g_hash_table_size (previous_paths) == songs_view_num_rows (l))
		{
			g_hash_table_destroy (previous_paths);
			return NULL;
		}
		
		/* start getting random nodes until we find one that is selected */
		index = random () % songs_view_num_rows (l);

		/* check we havent already had this one */
		if (g_hash_table_lookup (previous_paths, GINT_TO_POINTER (index)) != NULL)
			continue;
		
		/* insert in had paths node */
		g_hash_table_insert (previous_paths, GINT_TO_POINTER (index),
				     GINT_TO_POINTER (TRUE));

		/* get threepath */
		path_str = g_strdup_printf ("%d", index);
		path = gtk_tree_path_new_from_string (path_str);
		g_free (path_str);
	
		/* get iter */
		gtk_tree_model_get_iter (GTK_TREE_MODEL (l->priv->songs_sortmodel), &iter,
					 path);
		gtk_tree_path_free (path);

		/* get state */
		if (songs_view_iter_is_valid (l, &iter, ITER_VALIDATION_CHECKED) == FALSE) continue;

		/* get node */
		node = songs_view_iter_to_node (l, &iter);

		if (l->priv->playing_row ==
		    g_hash_table_lookup (l->priv->songs_node_to_iter, node))
		{
			/* nope we're already playing this one */
			continue;
		}
	}

	g_hash_table_destroy (previous_paths);

	return node;
}

/**
 * songs_view_build_status_message: generate a new status bar message
 */
gchar *
songs_view_build_status_message (SongsView *v)
{
	StatusbarInfo *info = g_new0 (StatusbarInfo, 1);
	gint hours;
	gint minutes;
	gint seconds;
	gchar *ret, *filesize;

	info->n_bytes = 0;
	info->n_seconds = 0;
	info->n_songs = 0;
	
	gtk_tree_model_foreach (GTK_TREE_MODEL (v->priv->songs_liststore),
			        (GtkTreeModelForeachFunc) gather_stats,
				info);

	hours   = info->n_seconds / (60 * 60);
	minutes = info->n_seconds / 60 - hours * 60;
	seconds = info->n_seconds % 60;

	filesize = gnome_vfs_format_file_size_for_display (info->n_bytes);

	ret = g_strdup_printf (_("%d songs, %d:%.2d:%.2d total time, %s"),
				info->n_songs, hours, minutes, seconds, filesize);

	g_free (info);
	g_free (filesize);

	return ret;
}

/**
 * gather_stats: helper func for gathering statusbar stats
 */
static gboolean
gather_stats (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, StatusbarInfo *info)
{
	GValue val = {0, };
	Node *node;

	/* check whether this thing is visible */
	gtk_tree_model_get_value (model, iter, SONGS_COL_VISIBLE, &val);
	if (!g_value_get_boolean (&val)) return FALSE;

	g_value_unset (&val);

	/* get the node */
	gtk_tree_model_get_value (model, iter, SONGS_COL_NODE, &val);
	node = g_value_get_pointer (&val);

	/* update stats */
	info->n_songs++;
	info->n_seconds = info->n_seconds +
		node_get_int_property (node, "length");
	info->n_bytes = info->n_bytes +
		node_get_int_property (node, "filesize");

	/* go on */
	return FALSE;
}

/**
 * songs_view_select_nodes: deselect all vis nodes
 */
void
songs_view_select_nodes (SongsView *v, GList *nodes)
{
	gtk_tree_model_foreach (GTK_TREE_MODEL (v->priv->songs_liststore),
			        (GtkTreeModelForeachFunc) set_checked,
				nodes);

	/* check rewind/forward sensitivity with the new list setup */
	songs_view_check_state (v);
}

/**
 * set_checked: change the selection of all visible nodes
 */
static gboolean
set_checked (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, GList *nodes)
{
	GValue val = {0, };
	Node *node;

	/* check whether this thing is visible */
	gtk_tree_model_get_value (model, iter, SONGS_COL_VISIBLE, &val);
	if (!g_value_get_boolean (&val)) return FALSE;

	/* get node */
	g_value_unset (&val);
	gtk_tree_model_get_value (model, iter, SONGS_COL_NODE, &val);
	node = g_value_get_pointer (&val);

	/* change selection */
	gtk_list_store_set (GTK_LIST_STORE (model), iter,
			    SONGS_COL_SELECTED, (g_list_find (nodes, node) != NULL), -1);

	/* go on */
	return FALSE;
}

/**
 * songs_view_get_selected_nodes: return all selected nodes
 */
GList *
songs_view_get_selected_nodes (SongsView *l)
{
	GtkTreeSelection *selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (l->priv->songs));
	GList *list = NULL;

	gtk_tree_selection_selected_foreach (selection,
					     (GtkTreeSelectionForeachFunc) get_real_selection,
					     (void **) &list);

	return list;
}

/**
 * make_selection_list: utl func to retreive the selection of a treeview
 * supporting multiple selection
 */
static void
get_real_selection (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
		    void **data)
{
	GList **list = (GList **) data;
	GValue val = {0, };

	gtk_tree_model_get_value (model, iter, SONGS_COL_NODE, &val);
	*list = g_list_append (*list, g_value_get_pointer (&val));
}

/**
 * row_changed_cb: row props changed, update checked_counter
 */
static void
row_changed_cb (GtkTreeModel *model, GtkTreePath *start_path, GtkTreeIter *start_iter, SongsView *v)
{
	GValue val = {0, };
	gboolean visible, checked;
	Node *node;
	
	gtk_tree_model_get_value (model, start_iter, SONGS_COL_VISIBLE, &val);
	visible = g_value_get_boolean (&val);
	g_value_unset (&val);
	
	gtk_tree_model_get_value (model, start_iter, SONGS_COL_SELECTED, &val);
	checked = g_value_get_boolean (&val);
	g_value_unset (&val);

	gtk_tree_model_get_value (model, start_iter, SONGS_COL_NODE, &val);
	node = g_value_get_pointer (&val);
	g_value_unset (&val);

	if (visible == TRUE && checked == TRUE)
		g_hash_table_replace (v->priv->checked_nodes, node, node);
	else
		g_hash_table_remove (v->priv->checked_nodes, node);
}

/**
 * songs_view_get_real_next: get the next node starting from @start. returns NULL if not found.
 */
static gboolean
songs_view_get_real_next (SongsView *l, GtkTreeIter *start)
{
	/* get next visible node */
	while (TRUE)
	{
		if (!gtk_tree_model_iter_next (l->priv->songs_sortmodel,
					       start))
		{
			start->stamp = 0;
			start->user_data = NULL;
			return FALSE;
		}

		if (songs_view_iter_is_valid (l, start, ITER_VALIDATION_VISIBLE | ITER_VALIDATION_CHECKED))
			return TRUE;
	}

	start->stamp = 0;
	start->user_data = NULL;

	return FALSE;
}

/**
 * songs_view_get_real_prev: get the previous node starting from @start. returns NULL if not found.
 */
static gboolean
songs_view_get_real_prev (SongsView *l, GtkTreeIter *start)
{
	GtkTreePath *path = gtk_tree_model_get_path (l->priv->songs_sortmodel, start);

	/* get prev visible node */
	while (TRUE) {
		if (gtk_tree_path_prev (path) == FALSE)
		{
			start->stamp = 0;
			start->user_data = NULL;
			return FALSE;
		}

		gtk_tree_model_get_iter (l->priv->songs_sortmodel, start, path);

		if (songs_view_iter_is_valid (l, start, ITER_VALIDATION_VISIBLE | ITER_VALIDATION_CHECKED))
			return TRUE;
	}

	start->stamp = 0;
	start->user_data = NULL;

	return FALSE;
}

/**
 * songs_view_get_real_first: get the first iter
 */
static gboolean
songs_view_get_real_first (SongsView *l, GtkTreeIter *iter)
{
	return gtk_tree_model_get_iter_first (l->priv->songs_sortmodel, iter);
}

/**
 * songs_view_get_real_selected: get the selected iter
 */
static gboolean
songs_view_get_real_selected (SongsView *l, GtkTreeIter *selected)
{
	GtkTreeIter *iter;
	GList *sel;
	Node *node;

	/* get selection */
	sel = songs_view_get_selected_nodes (l);
	if (g_list_length (sel) == 0)
		return FALSE;

	/* get the iter */
	node = (g_list_first (sel))->data;
	iter = g_hash_table_lookup (l->priv->songs_node_to_iter, node);

	store_iter_to_sort_iter (l->priv->songs_filtermodel,
				 l->priv->songs_sortmodel,
				 selected,
				 iter);

	return TRUE;
}

/**
 * songs_view_iter_is_valid: check whether this iter is sutiable
 */
static gboolean
songs_view_iter_is_valid (SongsView *l, GtkTreeIter *iter, IterValidationFlags flags)
{
	gboolean ok = FALSE;
	GValue val = {0, };

	if (iter == NULL) return FALSE;

	if (flags & ITER_VALIDATION_VISIBLE)
	{
		/* check whether this row is visible */
		gtk_tree_model_get_value (l->priv->songs_sortmodel,
					  iter, SONGS_COL_VISIBLE, &val);
		ok = g_value_get_boolean (&val);
		g_value_unset (&val);
	}

	if (flags & ITER_VALIDATION_CHECKED)
	{
		/* check whether this row is selected */
		gtk_tree_model_get_value (l->priv->songs_sortmodel,
					  iter, SONGS_COL_SELECTED, &val);
		ok = g_value_get_boolean (&val);
		g_value_unset (&val);
	}

	return ok;
}

/**
 * songs_view_store_iter_to_sort_iter: convert a iter from the liststore to
 * a sortmodel iter
 */
static void
songs_view_store_iter_to_sort_iter (SongsView *l,
				    GtkTreeIter *converted,
				    GtkTreeIter *in)
{
	/* convert the active_row path to a gtktreemodelsort path */
	return store_iter_to_sort_iter (l->priv->songs_filtermodel,
					l->priv->songs_sortmodel,
					converted,
					in);
}

/**
 * songs_view_iter_to_node: convert a sortmodel gtktreeiter to a Node
 */
static Node *
songs_view_iter_to_node (SongsView *l, GtkTreeIter *iter)
{
	GValue val = {0, };

	if (iter == NULL) return NULL;

	gtk_tree_model_get_value (l->priv->songs_sortmodel, iter,
				  SONGS_COL_NODE, &val);

	return g_value_get_pointer (&val);
}

static gboolean
songs_view_iter_in_sort_model (SongsView *l,
			       GtkTreeIter *iter,
			       GtkTreeIter *converted)
{
	GValue val = {0,};
	Node *node;
	gint id;
	GtkTreeModel *liststore = GTK_TREE_MODEL (l->priv->songs_liststore);
	GtkTreeIter tmp;

	if (!iter)
	{
		converted->stamp = 0;
		converted->user_data = NULL;
		return FALSE;
	}

	/* simple optimization */
	if (gtk_tree_model_iter_n_children (liststore, NULL) ==
	    gtk_tree_model_iter_n_children (l->priv->songs_sortmodel, NULL))
	{
		songs_view_store_iter_to_sort_iter (l, converted, iter);

		return TRUE;
	}

	/* check for real */
	gtk_tree_model_get_value (liststore, iter, SONGS_COL_NODE, &val);
	node = g_value_get_pointer (&val);
	g_value_unset (&val);
	id = node_get_int_property (node, "id");

	/* search for id in the sortmodel -- slow, but oh well */
	gtk_tree_model_get_iter_first (l->priv->songs_sortmodel, &tmp);

	do
	{
		gtk_tree_model_get_value (l->priv->songs_sortmodel, &tmp,
					  SONGS_COL_NODE, &val);
		node = g_value_get_pointer (&val);
		g_value_unset (&val);
		if (id == node_get_int_property (node, "id"))
		{
			*converted = tmp;
			return TRUE;
		}
	}
	while (gtk_tree_model_iter_next (l->priv->songs_sortmodel, &tmp));

	return FALSE;
}
