/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: source-manager.c,v 1.10 2002/04/10 20:50:39 jbaayen Exp $
 */

#include <libgnome/gnome-init.h>

#include "misc.h"
#include "queue.h"
#include "source-manager.h"
#include "ui.h"
#include "file-watcher.h"
#include "library.h"
#include "library-ui.h"
#include "playlist-ui.h"
#include "eel-gconf-extensions.h"

#define CONF_PLAYLISTS "/apps/rhythmbox/SourceManager/playlists"

struct _SourceManagerPrivate
{
	Library *library;
	UI *ui;
	Queue *queue;

	GList *clipboard;

	GHashTable *sources;

	GObject *playing_source;
	GObject *selected_source;
	
	guint timeout;
};

/* all the different signals */
enum
{
	CLIPBOARD_AVAILABLE,
	STOP_PLAYING, /* if this is emitted the source should stop playing since another source started */
	LAST_SIGNAL
};

/* GObject properties */
enum
{
	PROP_0,
	PROP_QUEUE,
	PROP_UI,
	PROP_LIBRARY,
	PROP_CLIPBOARD
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint source_manager_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void source_manager_class_init (SourceManagerClass *klass);
static void source_manager_init (SourceManager *s);
static void source_manager_finalize (GObject *object);
static void source_manager_set_property (GObject *object,
			                 guint prop_id,
			                 const GValue *value,
			                 GParamSpec *pspec);
static void source_manager_get_property (GObject *object,
			                 guint prop_id,
			                 GValue *value,
			                 GParamSpec *pspec);

/* local function prototypes */
static void list_and_unref_source (gpointer key, GObject *val, void **data);
static void add_playlist (SourceManager *m, const gchar *name, const gchar *fn,
			  gboolean from_sel, gboolean new);
static void sync_ui (SourceManager *m);
static void source_manager_release_brakes (SourceManager *m);

/* callback prototypes */
static void ui_new_playlist_cb (UI *ui, gchar *name, gboolean from_sel, SourceManager *m);
static void start_playing_cb (GObject *object, SourceManager *m);
static void stop_playing_cb (GObject *object, SourceManager *m);
static void removed_cb (GObject *object, gint id, SourceManager *m);
static void node_deleted_cb (Library *l, Node *node, SourceManager *m);
static void queue_state_changed_cb (Queue *q, QueueState state, SourceManager *m);
static void ui_volume_changed_cb (UI *ui, gfloat volume, SourceManager *m);
static void ui_source_selected_cb (UI *ui, gint id, SourceManager *m);
static void ui_new_song_cb (UI *ui, gchar *fn, SourceManager *m);
static gboolean idle_func (SourceManager *m);
static void ui_progress_changed_cb (UI *ui, float progress, SourceManager *m);
static void ui_set_shuffle_cb (UI *ui, gboolean shuffle, SourceManager *m);
static void ui_set_repeat_cb (UI *ui, gboolean repeat, SourceManager *m);
static GList *active_source_get_selection (SourceManager *m);

/**
 * source_manager_get_type: get the GObject type of the statusbar
 */
GType
source_manager_get_type (void)
{
	static GType source_manager_type = 0;

  	if (source_manager_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (SourceManagerClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) source_manager_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (SourceManager),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) source_manager_init
      		};

      		source_manager_type = g_type_register_static (G_TYPE_OBJECT,
                				              "SourceManager",
                                           	              &our_info, 0);
    	}

	return source_manager_type;
}

/**
 * source_manager_class_init: initialize the SourceManager class
 */
static void
source_manager_class_init (SourceManagerClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = source_manager_finalize;

	object_class->set_property = source_manager_set_property;
	object_class->get_property = source_manager_get_property;

	/* init properties */
	g_object_class_install_property (object_class,
					 PROP_QUEUE,
					 g_param_spec_object ("queue",
						 	      "UI",
							      "Associated UI object",
							      queue_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_UI,
					 g_param_spec_object ("ui",
						 	      "UI",
							      "Associated UI object",
							      ui_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_LIBRARY,
					 g_param_spec_object ("library",
						 	      "Library",
							      "Associated Library object",
							      library_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_CLIPBOARD,
					 g_param_spec_pointer ("clipboard",
						 	       "Clipboard",
							       "List of Nodes in the clipboard",
							       G_PARAM_READWRITE));

	/* init signals */
	source_manager_signals[CLIPBOARD_AVAILABLE] =
   		g_signal_new ("clipboard_available",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SourceManagerClass, clipboard_available),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	source_manager_signals[STOP_PLAYING] =
   		g_signal_new ("stop_playing",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SourceManagerClass, stop_playing),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__POINTER,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_POINTER);
}

/**
 * source_manager_init: intialize the statusbar object
 */
static void
source_manager_init (SourceManager *s)
{
	s->priv = g_new0 (SourceManagerPrivate, 1);
	s->priv->sources = g_hash_table_new (NULL, NULL);
	s->priv->playing_source = NULL;
}

/**
 * source_manager_finalize: finalize the queue object
 */
static void
source_manager_finalize (GObject *object)
{
	SourceManager *s;
	GSList *list = NULL;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_SOURCE_MANAGER (object));
	
   	s = SOURCE_MANAGER (object);

	g_return_if_fail (s->priv != NULL);

	/* remove idle func */
	gtk_timeout_remove (s->priv->timeout);

	/* remove all sources, and list playlist filenames */
	g_hash_table_foreach (s->priv->sources,
			      (GHFunc) list_and_unref_source, (void **) &list);
	g_hash_table_destroy (s->priv->sources);

	/* save list of playlist filenames */
	eel_gconf_set_string_list (CONF_PLAYLISTS, list);

	/* cleanup */
	g_slist_foreach (list, (GFunc) g_free, NULL);
	g_slist_free (list);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * source_manager_new: create a new SourceManager object
 */
SourceManager *
source_manager_new (Queue *queue, UI *ui, Library *library)
{
 	SourceManager *s;

	s = SOURCE_MANAGER (g_object_new (TYPE_SOURCE_MANAGER,
					  "queue", queue,
					  "ui", ui,
					  "library", library,
					  NULL));

	g_return_val_if_fail (s->priv != NULL, NULL);

	source_manager_release_brakes (s);

	return s;
}

/**
 * source_manager_set_ui: attach the UI to the manager
 */
void
source_manager_set_ui (SourceManager *m, UI *ui)
{
	m->priv->ui = ui;

	/* sync ui to queue wrt volume */
	ui_set_volume (ui, queue_get_volume (m->priv->queue));

	/* restore order settings */
	ui_set_shuffle (ui, eel_gconf_get_boolean (CONF_SHUFFLE));
	ui_set_repeat  (ui, eel_gconf_get_boolean (CONF_REPEAT));

	/* connect playlist-related signals */
	g_signal_connect (G_OBJECT (ui), "new_playlist",
			  G_CALLBACK (ui_new_playlist_cb), m);
	g_signal_connect (G_OBJECT (ui), "volume_changed",
			  G_CALLBACK (ui_volume_changed_cb), m);
	g_signal_connect (G_OBJECT (ui), "source_selected",
			  G_CALLBACK (ui_source_selected_cb), m);
	g_signal_connect (G_OBJECT (ui), "new_song",
			  G_CALLBACK (ui_new_song_cb), m);
	g_signal_connect (G_OBJECT (ui), "progress_changed",
			  G_CALLBACK (ui_progress_changed_cb), m);
	g_signal_connect (G_OBJECT (ui), "set_shuffle",
			  G_CALLBACK (ui_set_shuffle_cb), m);
	g_signal_connect (G_OBJECT (ui), "set_repeat",
			  G_CALLBACK (ui_set_repeat_cb), m);

	g_object_notify (G_OBJECT (m), "ui");
}

/**
 * source_manager_get_ui: get the attached UI
 */
UI *
source_manager_get_ui (SourceManager *m)
{
	return m->priv->ui;
}

/**
 * source_manager_set_queue: attach the manager to a queue
 */
void
source_manager_set_queue (SourceManager *m, Queue *queue)
{
	m->priv->queue = queue;

	g_signal_connect (G_OBJECT (queue), "state_changed",
			  G_CALLBACK (queue_state_changed_cb), m);

	g_object_notify (G_OBJECT (m), "queue");
}

/**
 * source_manager_get_queue: return the attached queue
 */
Queue *
source_manager_get_queue (SourceManager *m)
{
	return m->priv->queue;
}

/**
 * source_manager_release_brakes: call this when all data has 
 * been attached. This starts the manager.
 */
static void
source_manager_release_brakes (SourceManager *m)
{
	GSList *playlists, *l;
	gint id;
	LibraryUI *lui;

	/* reset ui */
	ui_set_song (m->priv->ui, NULL, NULL, NULL);
	ui_set_can_forward (m->priv->ui, FALSE);
	ui_set_can_rewind (m->priv->ui, FALSE);

	/* load the library ui */
	lui = library_ui_new (m);
	id = library_ui_get_source_id (lui);

	g_signal_connect (G_OBJECT (lui), "start_playing",
			  G_CALLBACK (start_playing_cb), m);
	g_signal_connect (G_OBJECT (lui), "stop_playing",
			  G_CALLBACK (stop_playing_cb), m);

	g_hash_table_insert (m->priv->sources, GINT_TO_POINTER (id), lui);

	/* select the library by default */
	ui_select_source (m->priv->ui, id);

	/* everything attached, the library can start loading/monitoring now */
	library_release_brakes (m->priv->library);

	/* load playlists */
	playlists = eel_gconf_get_string_list (CONF_PLAYLISTS);
	for (l = playlists; l != NULL; l = g_slist_next (l))
	{
		gchar *fn = l->data;

		add_playlist (m, NULL, fn, FALSE, FALSE);

		g_free (fn);
	}
	g_slist_free (playlists);

	/* add idle func that is responsible for updating the playing time */
	m->priv->timeout = gtk_timeout_add (10, (GtkFunction) idle_func, m);
}

/**
 * ui_new_playlist_cb: new playlist request caught
 */
static void
ui_new_playlist_cb (UI *ui, gchar *name, gboolean from_sel, SourceManager *m)
{
	add_playlist (m, name, NULL, from_sel, TRUE);
}

/**
 * source_manager_set_clipboard: set the nodes clipboard
 */
void            
source_manager_set_clipboard (SourceManager *m, GList *nodes)
{
	g_list_free (m->priv->clipboard);
	m->priv->clipboard = nodes;
	g_signal_emit (G_OBJECT (m), source_manager_signals[CLIPBOARD_AVAILABLE], 0,
		       (g_list_length (nodes) > 0));
	g_object_notify (G_OBJECT (m), "clipboard");
}

/**
 * source_manager_get_clipboard: get the nodes clipboard
 */
GList *
source_manager_get_clipboard (SourceManager *m)
{
	return m->priv->clipboard;
}

/**
 * list_and_unref_source: hashtable foreach helper func that unrefs each value, and
 * if it's a playlist adds it's filename to the list
 */
static void
list_and_unref_source (gpointer key, GObject *val, void **data)
{
	GSList **ret = (GSList **) data;
	if (G_OBJECT_TYPE (G_OBJECT (val)) == TYPE_PLAYLIST_UI)
	{
		*ret = g_slist_append (*ret,
				       g_strdup (playlist_ui_get_file (PLAYLIST_UI (val))));
	}
	g_object_unref (val);
}

/**
 * start_playing_cb: one source starts playing, unset playing states
 * on all other sources
 */
static void
start_playing_cb (GObject *object, SourceManager *m)
{
	m->priv->playing_source = object;

	/* signal all other sources to stop playing */
	g_signal_emit (G_OBJECT (m), source_manager_signals[STOP_PLAYING], 0, m->priv->playing_source);
}

/**
 * stop_playing_cb: a source stopped playing
 */
static void
stop_playing_cb (GObject *object, SourceManager *m)
{
	if (m->priv->playing_source == object)
	{
		m->priv->playing_source = NULL;
	}
}

/**
 * removed_cb: source removed, remove from sources hash
 */
static void
removed_cb (GObject *object, gint id, SourceManager *m)
{
	if (object == m->priv->playing_source)
	{
		/* this source was playing, reset playing state */
		m->priv->playing_source = NULL;
	}
	g_hash_table_remove (m->priv->sources, GINT_TO_POINTER (id));
}

/**
 * add_playlist: add a new playlist
 */
static void
add_playlist (SourceManager *m, const gchar *name, const gchar *fn, gboolean from_sel, gboolean new)
{
	PlaylistUI *pui = playlist_ui_new (m);
	gint id;

	if (name != NULL)
		playlist_ui_set_name (pui, name);
	playlist_ui_load_from_file (pui, fn);
	id = playlist_ui_get_source_id (pui);

	if (from_sel == TRUE)
	{
		GList *selection = active_source_get_selection (m);
		playlist_ui_add_nodes (pui, selection);
		g_list_free (selection);
	}

	if (new == TRUE)
	{
		playlist_ui_start_editing (pui);
	}

	g_signal_connect (G_OBJECT (pui), "start_playing",
			  G_CALLBACK (start_playing_cb), m);
	g_signal_connect (G_OBJECT (pui), "stop_playing",
			  G_CALLBACK (stop_playing_cb), m);
	g_signal_connect (G_OBJECT (pui), "removed",
			  G_CALLBACK (removed_cb), m);

	g_hash_table_insert (m->priv->sources, GINT_TO_POINTER (id), pui);
}

/**
 * node_deleted_cb: node deleted, remove from the clipboard if it was there
 */
static void
node_deleted_cb (Library *l, Node *node, SourceManager *m)
{
	if (g_list_find (m->priv->clipboard, node) != NULL)
	{
		g_list_remove (m->priv->clipboard, node);
		if (g_list_length (m->priv->clipboard) == 0)
		{
			g_signal_emit (G_OBJECT (m),
				       source_manager_signals[CLIPBOARD_AVAILABLE], 0,
				       FALSE);
		}
	}
}

/**
 * queue_state_changed_cb: queue state changed, update the ui
 */
static void
queue_state_changed_cb (Queue *q, QueueState state, SourceManager *m)
{
	sync_ui (m);
}

/**
 * sync_ui: sync UI with manager
 */
static void
sync_ui (SourceManager *m)
{
	QueueState state = queue_get_state (m->priv->queue);
	UIState newstate = UI_STATE_STOPPED;

	switch (state)
	{
	case QUEUE_STATE_STOPPED:
		newstate = UI_STATE_STOPPED;
		break;
	case QUEUE_STATE_PLAYING:
		if (m->priv->selected_source == m->priv->playing_source)
			newstate = UI_STATE_PLAYING;
		else
			newstate = UI_STATE_PLAYING_WITH_STOP;
		break;
	case QUEUE_STATE_PAUSED:
		if (m->priv->selected_source == m->priv->playing_source)
			newstate = UI_STATE_PAUSED;
		else
			newstate = UI_STATE_PLAYING_WITH_STOP;
		break;
	default:
		g_warning ("Shouldn't get here!\n");
		break;
	}
	
	ui_set_state (m->priv->ui, newstate);
}

/**
 * ui_volume_changed_cb: volume in the ui changed, update the queue
 */
static void
ui_volume_changed_cb (UI *ui, gfloat volume, SourceManager *m)
{
	queue_set_volume (m->priv->queue, volume);
}

/**
 * ui_source_selected_cb: source selected, check whether we need to
 * update the playing button
 */
static void
ui_source_selected_cb (UI *ui, gint id, SourceManager *m)
{
	GObject *source = g_hash_table_lookup (m->priv->sources,
					       GINT_TO_POINTER (id));
	if (source == NULL) return;
	
	m->priv->selected_source = source;

	sync_ui (m);
}

/**
 * ui_new_song_cb: add a new song to the library
 */
static void
ui_new_song_cb (UI *ui, gchar *fn, SourceManager *m)
{
	library_add_uri (m->priv->library, fn);
}

/**
 * idle_func: timeout function, used to update playing time and such
 */
static gboolean
idle_func (SourceManager *m)
{
	const char *current;
	
	current = queue_get_current_uri (m->priv->queue);

	if (current != NULL)
	{
		long elapsed = queue_get_elapsed_time (m->priv->queue);
		float progress = queue_get_progress (m->priv->queue);
		ui_set_elapsed_time (m->priv->ui, elapsed, progress);
	}

	return TRUE;
}

/**
 * ui_progress_changed_cb: seek request caught, do the seek
 */
static void
ui_progress_changed_cb (UI *ui, float progress, SourceManager *m)
{
	queue_set_progress (m->priv->queue, progress);
}

/**
 * ui_set_shuffle_cb: ui shuffle state changed, store new setting
 */
static void
ui_set_shuffle_cb (UI *ui, gboolean shuffle, SourceManager *m)
{
	eel_gconf_set_boolean (CONF_SHUFFLE, shuffle);
}

/**
 * ui_set_repeat_cb: ui repeat state changed, store new setting
 */
static void
ui_set_repeat_cb (UI *ui, gboolean repeat, SourceManager *m)
{
	eel_gconf_set_boolean (CONF_REPEAT, repeat);
}

/**
 * active_source_get_selection: return the list of checked nodes for the active source
 */
static GList *
active_source_get_selection (SourceManager *m)
{
	GType type = G_OBJECT_TYPE (G_OBJECT (m->priv->selected_source));
	if (type == TYPE_LIBRARY_UI)
		return library_ui_get_selection (LIBRARY_UI (m->priv->selected_source));
	else if (type == TYPE_PLAYLIST_UI) 
		return playlist_ui_get_selection (PLAYLIST_UI (m->priv->selected_source));
	else
		return NULL;
}

/**
 * source_manager_set_library: associate a library with this manager
 */
void
source_manager_set_library (SourceManager *m, Library *l)
{
	m->priv->library = l;
	g_signal_connect (G_OBJECT (m->priv->library),
			  "node_deleted", G_CALLBACK (node_deleted_cb), m);
	g_object_notify (G_OBJECT (m), "library");
}

/**
 * source_manager_get_library: return the library object associated with this manager
 */
Library *
source_manager_get_library (SourceManager *m)
{
	return m->priv->library;
}

/**
 * source_manager_set_property: set a SourceManager property
 */
static void
source_manager_set_property (GObject *object,
			     guint prop_id,
			     const GValue *value,
			     GParamSpec *pspec)
{
	SourceManager *m = SOURCE_MANAGER (object);

	switch (prop_id)
	{
		case PROP_QUEUE:
			source_manager_set_queue (m, g_value_get_object (value));
			break;
		case PROP_UI:
			source_manager_set_ui (m, g_value_get_object (value));
			break;
		case PROP_LIBRARY:
			source_manager_set_library (m, g_value_get_object (value));
			break;
		case PROP_CLIPBOARD:
			source_manager_set_clipboard (m, g_value_get_pointer (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/**
 * source_manager_get_property: get a StateManager property
 */
static void
source_manager_get_property (GObject *object,
			     guint prop_id,
			     GValue *value,
			     GParamSpec *pspec)
{
	SourceManager *m = SOURCE_MANAGER (object);

	switch (prop_id)
	{
		case PROP_QUEUE:
			g_value_set_object (value, m->priv->queue);
			break;
		case PROP_UI:
			g_value_set_object (value, m->priv->ui);
			break;
		case PROP_LIBRARY:
			g_value_set_object (value, m->priv->library);
			break;
		case PROP_CLIPBOARD:
			g_value_set_pointer (value, m->priv->clipboard);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}
