/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: state-manager.c,v 1.7 2002/04/11 15:31:40 jbaayen Exp $
 */

#include <string.h>

#include "misc.h"
#include "state-manager.h"
#include "node-utils.h"

struct _StateManagerPrivate
{
	Queue *queue;    /* associated Queue */
	
	SongsView *view; /* associated SongsView */

	UI *ui;          /* associated UI */

	int id;          /* source ID */

	Node *node;      /* curerntly playing node */
	
	gboolean paused; /* are we currently paused ? */
	gboolean active; /* are we currently playing (active) ? */
};

/* all the different signals */
enum
{
	START_PLAYING,
	STOP_PLAYING,
	LAST_SIGNAL
};

/* GObject properties */
enum
{
	PROP_0,
	PROP_QUEUE,
	PROP_UI,
	PROP_SONGS_VIEW,
	PROP_SOURCE_ID,
	PROP_ACTIVE,
	PROP_PAUSED,
	PROP_PLAYING_NODE
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint state_manager_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void state_manager_class_init (StateManagerClass *klass);
static void state_manager_init (StateManager *s);
static void state_manager_finalize (GObject *object);
static void state_manager_set_property (GObject *object,
			                guint prop_id,
			                const GValue *value,
			                GParamSpec *pspec);
static void state_manager_get_property (GObject *object,
			                guint prop_id,
			                GValue *value,
			                GParamSpec *pspec);

/**
 * state_manager_get_type: get the GObject type of the StateManager 
 */
GType
state_manager_get_type (void)
{
	static GType state_manager_type = 0;

  	if (state_manager_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (StateManagerClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) state_manager_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (StateManager),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) state_manager_init
      		};

      		state_manager_type = g_type_register_static (G_TYPE_OBJECT,
                				             "StateManager",
                                           	             &our_info, 0);
    	}

	return state_manager_type;
}

/**
 * state_manager_class_init: initialize the StateManager class
 */
static void
state_manager_class_init (StateManagerClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = state_manager_finalize;

	object_class->set_property = state_manager_set_property;
	object_class->get_property = state_manager_get_property;

	/* init properties */
	g_object_class_install_property (object_class,
					 PROP_QUEUE,
					 g_param_spec_object ("queue",
						 	      "UI",
							      "Associated UI object",
							      queue_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_UI,
					 g_param_spec_object ("ui",
						 	      "UI",
							      "Associated UI object",
							      ui_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_SONGS_VIEW,
					 g_param_spec_object ("songs-view",
						 	      "SongsView",
							      "Associated SongsView object",
							      songs_view_get_type (),
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_SOURCE_ID,
					 g_param_spec_int ("source-id",
						 	   "Source ID",
							   "Source ID",
							   0, G_MAXINT, 0,
							   G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_ACTIVE,
					 g_param_spec_boolean ("active",
						 	       "Active",
							       "Whether we're active",
							       FALSE,
							       G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_PAUSED,
					 g_param_spec_boolean ("paused",
						 	       "Paused",
							       "Whether we're paused",
							       FALSE,
							       G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_PLAYING_NODE,
					 g_param_spec_object ("playing-node",
						 	      "Playing Node object",
							      "Currently playing Node object",
							      node_get_type (),
							      G_PARAM_READWRITE));

	/* init signals */
	state_manager_signals[START_PLAYING] =
   		g_signal_new ("start_playing",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (StateManagerClass, start_playing),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	state_manager_signals[STOP_PLAYING] =
   		g_signal_new ("stop_playing",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (StateManagerClass, stop_playing),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
}

/**
 * state_manager_init: intialize the StateManager object
 */
static void
state_manager_init (StateManager *w)
{
	w->priv = g_new0 (StateManagerPrivate, 1);

	w->priv->active = FALSE;
	w->priv->paused = FALSE;
}

/**
 * state_manager_finalize: finalize the StateManager object
 */
static void
state_manager_finalize (GObject *object)
{
	StateManager *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_STATE_MANAGER (object));
	
   	s = STATE_MANAGER (object);

	g_return_if_fail (s->priv != NULL);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * state_manager_new: create a new StateManager object
 */
StateManager *
state_manager_new (Queue *queue, UI *ui, SongsView *view, int id)
{
 	StateManager *s;

	s = STATE_MANAGER (g_object_new (TYPE_STATE_MANAGER,
					 "queue", queue,
					 "ui", ui,
					 "songs-view", view,
					 "source-id", id,
					 NULL));

	g_return_val_if_fail (s->priv != NULL, NULL);

	return s;
}

/**
 * state_manager_get_active: return whether the manager is currently active or not
 */
gboolean
state_manager_get_active (StateManager *m)
{
	return m->priv->active;
}

/**
 * state_manager_set_active: unset this source's activity
 */
void
state_manager_set_active (StateManager *m, gboolean active)
{
	if (m->priv->active != active)
	{
		m->priv->active = active;
		m->priv->paused = FALSE;

		if (active == TRUE)
		{
			g_signal_emit (G_OBJECT (m), state_manager_signals[START_PLAYING], 0);
		}
		else
		{
			g_signal_emit (G_OBJECT (m), state_manager_signals[STOP_PLAYING], 0);
			songs_view_set_node_state (m->priv->view, m->priv->node, SONGS_VIEW_NODE_STATE_STOPPED);
		}
		
		ui_source_set_playing (m->priv->ui, m->priv->id, active);
	}

	g_object_notify (G_OBJECT (m), "active");
}

/**
 * state_manager_set_playing_node: set node currently playing
 */
void
state_manager_set_playing_node (StateManager *m, Node *node)
{
	const char *artist = NULL;
	const char *name = NULL;
	const char *uri = NULL;
	Node *artist_node;

	/* set ourselves to be active if necessary */
	if (m->priv->active == FALSE && node != NULL)
		state_manager_set_active (m, TRUE);

	/* update view */
	songs_view_set_node_state (m->priv->view, m->priv->node, SONGS_VIEW_NODE_STATE_STOPPED);
	m->priv->node = node;
	songs_view_set_node_state (m->priv->view, m->priv->node, SONGS_VIEW_NODE_STATE_PLAYING);

	if (node != NULL)
	{
		/* load new node into queue */
		queue_set_current_uri (m->priv->queue,
				       node_get_string_property (node, "uri"),
				       (long) node_get_int_property (node, "length"));
			
		/* get node properties, if any */
		artist_node = song_get_artist (node);
		if (artist_node != NULL)
		{
			artist = node_get_string_property (artist_node, NODE_PROPERTY_NAME);
		}
		else
		{
			artist = "";
		}
		name = node_get_string_property (node, NODE_PROPERTY_NAME);
		uri = node_get_string_property (node, SONG_PROPERTY_URI);
		
		/* start playing */
		queue_set_state (m->priv->queue, QUEUE_STATE_PLAYING);
	}
	else
	{
		/* no node, so stop palying */
		queue_set_state (m->priv->queue, QUEUE_STATE_STOPPED);
	}

	/* update ui */
	ui_set_song (m->priv->ui, name, artist, uri);

	/* notify object */
	g_object_notify (G_OBJECT (m), "playing-node");
}

/**
 * state_manager_get_playing_node: get node currently playing
 */
Node *
state_manager_get_playing_node (StateManager *m)
{
	return m->priv->node;
}

/**
 * state_manager_set_paused: set the manager to be paused
 */
void
state_manager_set_paused (StateManager *m, gboolean paused)
{
	m->priv->paused = paused;

	if (paused == TRUE)
	{
		queue_set_state (m->priv->queue, QUEUE_STATE_PAUSED);
		songs_view_set_node_state (m->priv->view, m->priv->node, SONGS_VIEW_NODE_STATE_PAUSED);
	}
	else
	{
		queue_set_state (m->priv->queue, QUEUE_STATE_PLAYING);
		songs_view_set_node_state (m->priv->view, m->priv->node, SONGS_VIEW_NODE_STATE_PLAYING);
	}

	g_object_notify (G_OBJECT (m), "paused");
}

/**
 * state_manager_get_paused: get paused state
 */
gboolean 
state_manager_get_paused (StateManager *m)
{
	return m->priv->paused;
}

/**
 * state_manager_set_property: set a StateManager property
 */
static void
state_manager_set_property (GObject *object,
			    guint prop_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
	StateManager *m = STATE_MANAGER (object);

	switch (prop_id)
	{
		case PROP_QUEUE:
			m->priv->queue = g_value_get_object (value);
			break;
		case PROP_UI:
			m->priv->ui = g_value_get_object (value);
			break;
		case PROP_SONGS_VIEW:
			m->priv->view = g_value_get_object (value);
			break;
		case PROP_SOURCE_ID:
			m->priv->id = g_value_get_int (value);
			break;
		case PROP_ACTIVE:
			state_manager_set_active (m, g_value_get_boolean (value));
			break;
		case PROP_PAUSED:
			state_manager_set_paused (m, g_value_get_boolean (value));
			break;
		case PROP_PLAYING_NODE:
			state_manager_set_playing_node (m, g_value_get_object (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/**
 * state_manager_get_property: get a StateManager property
 */
static void
state_manager_get_property (GObject *object,
			    guint prop_id,
			    GValue *value,
			    GParamSpec *pspec)
{
	StateManager *m = STATE_MANAGER (object);

	switch (prop_id)
	{
		case PROP_QUEUE:
			g_value_set_object (value, m->priv->queue);
			break;
		case PROP_UI:
			g_value_set_object (value, m->priv->ui);
			break;
		case PROP_SONGS_VIEW:
			g_value_set_object (value, m->priv->view);
			break;
		case PROP_SOURCE_ID:
			g_value_set_int (value, m->priv->id);
			break;
		case PROP_ACTIVE:
			g_value_set_boolean (value, m->priv->active);
			break;
		case PROP_PAUSED:
			g_value_set_boolean (value, m->priv->paused);
			break;
		case PROP_PLAYING_NODE:
			g_value_set_object (value, m->priv->node);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}
