/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: queue.c,v 1.21 2002/04/10 20:50:40 jbaayen Exp $
 */

#include <libgnome/gnome-i18n.h>
#include <monkey-media.h>

#include "misc.h"
#include "queue.h"
#include "eel-gconf-extensions.h"

#define CONF_VOLUME "/apps/rhythmbox/Queue/volume"

struct _QueuePrivate
{
	/* pointer to the monkey-sound object */
	MonkeyMedia *player;

	/* current queue state (playing, stopped, paused...) */
	QueueState state;

	/* currently loaded uri */
	const char *current_uri;

	/* elapsed time calculation stuff */
	long total_seconds;
	GTimer *timer;
};

/* all the different signals */
enum
{
	REQUEST_NEXT,
	STATE_CHANGED,
	LAST_SIGNAL
};

/* GObject properties */
enum
{
	PROP_0,
	PROP_STATE,
	PROP_VOLUME,
	PROP_URI,
	PROP_PROGRESS,
	PROP_ELAPSED_TIME
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint queue_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void queue_class_init (QueueClass *klass);
static void queue_init (Queue *s);
static void queue_finalize (GObject *object);
static void queue_set_property (GObject *object,
			        guint prop_id,
			        const GValue *value,
			        GParamSpec *pspec);
static void queue_get_property (GObject *object,
		                guint prop_id,
		                GValue *value,
		                GParamSpec *pspec);

/* local function prototypes */
static void song_played_cb (MonkeyMedia *player, Queue *q);

/**
 * queue_get_type: get the GObject type of the statusbar
 */
GType
queue_get_type (void)
{
	static GType queue_type = 0;

  	if (queue_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (QueueClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) queue_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (Queue),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) queue_init
      		};

      		queue_type = g_type_register_static (G_TYPE_OBJECT,
                				     "Queue",
                                           	     &our_info, 0);
    	}

	return queue_type;
}

/**
 * queue_class_init: initialize the Queue class
 */
static void
queue_class_init (QueueClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = queue_finalize;

	object_class->set_property = queue_set_property;
	object_class->get_property = queue_get_property;

	/* init properties */
	g_object_class_install_property (object_class,
					 PROP_STATE,
					 g_param_spec_enum ("state",
						 	    "Queue state",
							    "Queue playing state",
							    queue_state_get_type (),
							    QUEUE_STATE_STOPPED,
							    G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_VOLUME,
					 g_param_spec_float ("volume",
						 	     "Volume",
							     "Playing volume",
							     0.0, 1.0, 1.0,
							     G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_URI,
					 g_param_spec_string ("uri",
						 	      "Loaded URI",
							      "Currently loaded URI",
							      NULL,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_PROGRESS,
					 g_param_spec_float ("progress",
						 	     "Progress",
							     "Playing progress",
							     0.0, 1.0, 0.0,
							     G_PARAM_READABLE));
	g_object_class_install_property (object_class,
					 PROP_ELAPSED_TIME,
					 g_param_spec_long ("elapsed-time",
						 	    "Elapsed time",
							    "Elapsed playing time in seconds",
							    0, G_MAXLONG, 0,
							    G_PARAM_READABLE));

	/* init signals */
	queue_signals[REQUEST_NEXT] =
   		g_signal_new ("request_next",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (QueueClass, request_next),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	queue_signals[STATE_CHANGED] =
   		g_signal_new ("state_changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (QueueClass, state_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__ENUM,
			      G_TYPE_NONE,
			      1,
			      queue_state_get_type ());
}

/**
 * queue_init: intialize the statusbar object
 */
static void
queue_init (Queue *s)
{
	s->priv = g_new0 (QueuePrivate, 1);

	s->priv->player = monkey_media_new ();

	g_assert (s->priv->player != NULL);

	/* setup timer */
	s->priv->timer = g_timer_new ();

	/* restore volume */
	queue_set_volume (s, (float) eel_gconf_get_integer (CONF_VOLUME) / 100);

	s->priv->current_uri = NULL;

	g_signal_connect (G_OBJECT (s->priv->player),
			  "eos",
			  G_CALLBACK (song_played_cb),
			  s);
}

/**
 * queue_finalize: finalize the queue object
 */
static void
queue_finalize (GObject *object)
{
	Queue *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_QUEUE (object));
	
   	s = QUEUE (object);

	g_return_if_fail (s->priv != NULL);

	/* kill timer */
	g_timer_destroy (s->priv->timer);

	/* save volume */
	eel_gconf_set_integer (CONF_VOLUME, queue_get_volume (s) * 100);

	g_object_unref (s->priv->player);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * queue_new: create a new queue object
 */
Queue *
queue_new (void)
{
 	Queue *s;

	s = QUEUE (g_object_new (TYPE_QUEUE, NULL));

	g_return_val_if_fail (s->priv != NULL, NULL);

	return s;
}

/**
 * queue_set_state: set the Queue's state
 */
void
queue_set_state (Queue *q, QueueState state)
{
	q->priv->state = state;

	switch (state)
	{
	case QUEUE_STATE_STOPPED:
	        monkey_media_stream_stop (q->priv->player);
		q->priv->current_uri = NULL;
		g_timer_stop (q->priv->timer);
		g_timer_reset (q->priv->timer);
		break;
	case QUEUE_STATE_PLAYING:
		monkey_media_stream_play (q->priv->player);
		g_timer_start (q->priv->timer);
		break;
	case QUEUE_STATE_PAUSED:
		monkey_media_stream_pause (q->priv->player);
		g_timer_stop (q->priv->timer);
		break;
	default:
		g_warning ("Invalid state!\n");
		break;
	}
	
	g_signal_emit (G_OBJECT (q), queue_signals[STATE_CHANGED], 0, q->priv->state);

	g_object_notify (G_OBJECT (q), "state");
}

/**
 * queue_get_state: return the queueueu's state
 */
QueueState
queue_get_state (Queue *q)
{
	return q->priv->state;
}

/**
 * queue_set_current_uri: set the uri currently playing
 * (pass -1 to total_seconds if the total playing time is unknown,
 *  for icecast streams for example. it is required for proper
 *  elapsed playing time calculation)
 */
void
queue_set_current_uri (Queue *q, const char *uri, long total_seconds)
{
	/* update struct */
	q->priv->current_uri = uri;
	q->priv->total_seconds = total_seconds;
	
	/* reset timer */
	g_timer_stop (q->priv->timer);
	g_timer_reset (q->priv->timer);

	/* load into monkey-sound */
	monkey_media_stream_new (q->priv->player, q->priv->current_uri);

	g_object_notify (G_OBJECT (q), "uri");
}

/**
 * queue_get_current_uri: get the uri currently playing
 */
const char *
queue_get_current_uri (Queue *q)
{
	return q->priv->current_uri;
}

/**
 * song_played_cb: song finished playing, request a new song
 */
static void
song_played_cb (MonkeyMedia *player, Queue *q)
{
	queue_set_state (q, QUEUE_STATE_STOPPED);

	g_signal_emit (G_OBJECT (q), queue_signals[REQUEST_NEXT], 0);
}

/**
 * queue_set_volume: set the volume
 */
void
queue_set_volume (Queue *q, gfloat volume)
{
	monkey_media_audio_set_volume (q->priv->player, volume);

	g_object_notify (G_OBJECT (q), "volume");
}

/**
 * queue_get_volume: get the volume
 */
float
queue_get_volume (Queue *q)
{
	return monkey_media_audio_get_volume (q->priv->player);
}

/**
 * queue_get_progress: return progress
 */
float
queue_get_progress (Queue *q)
{
	return monkey_media_stream_get_progress (q->priv->player);
}

/**
 * queue_set_progress: change playing progress (ie seek)
 */
void
queue_set_progress (Queue *q, float progress)
{
	monkey_media_stream_set_progress (q->priv->player, progress);
}

/**
 * queue_get_elapsed_time: get time spent playing the current uri
 */
long
queue_get_elapsed_time (Queue *q)
{
	/* FIXME when gstreamer is fixed we should use the monkey-sound call,
	 * for now this hack
	return monkey_media_stream_get_elapsed_time (q->priv->player);
	*/
	if (q->priv->total_seconds >= 0)
	{
		/* we know the total number of seconds, so we can calculate
		 * the elapsed time */
		return (long) (q->priv->total_seconds * queue_get_progress (q));
	}
	else
	{
		/* we dont know the number of seconds, which implies we
		 * cannot seek, so the timer will be OK */
		return (long) g_timer_elapsed (q->priv->timer, NULL);
	}
}

/**
 * queue_state_get_type: get the type of the QueueState enum
 */
GType
queue_state_get_type (void)
{
	static GType etype = 0;

	if (etype == 0)
	{
		static const GEnumValue values[] = {
			{ QUEUE_STATE_STOPPED, "QUEUE_STATE_STOPPED", "stopped" },
			{ QUEUE_STATE_PLAYING, "QUEUE_STATE_PLAYING", "playing" },
			{ QUEUE_STATE_PAUSED,  "QUEUE_STATE_PAUSED",  "paused"  },
			{ 0, NULL, NULL }
		};

		etype = g_enum_register_static ("QueueState", values);
	}

	return etype;
}

/**
 * queue_set_property: set a Queue property
 */
static void
queue_set_property (GObject *object,
		    guint prop_id,
		    const GValue *value,
		    GParamSpec *pspec)
{
	Queue *q = QUEUE (object);

	switch (prop_id)
	{
		case PROP_STATE:
			queue_set_state (q, g_value_get_enum (value));
			break;
		case PROP_VOLUME:
			queue_set_volume (q, g_value_get_float (value));
			break;
		case PROP_URI:
			queue_set_current_uri (q, g_value_get_string (value), -1);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/**
 * queue_get_property: get a Queue property
 */
static void
queue_get_property (GObject *object,
		    guint prop_id,
		    GValue *value,
		    GParamSpec *pspec)
{
	Queue *q = QUEUE (object);

	switch (prop_id)
	{
		case PROP_STATE:
			g_value_set_enum (value, queue_get_state (q));
			break;
		case PROP_VOLUME:
			g_value_set_float (value, queue_get_volume (q));
			break;
		case PROP_URI:
			g_value_set_string (value, queue_get_current_uri (q));
			break;
		case PROP_PROGRESS:
			g_value_set_float (value, queue_get_progress (q));
			break;
		case PROP_ELAPSED_TIME:
			g_value_set_long (value, queue_get_elapsed_time (q));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}
