/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: main.c,v 1.30 2002/04/01 14:57:44 jbaayen Exp $
 */

#include <libgnomevfs/gnome-vfs-directory.h>
#include <libgnomevfs/gnome-vfs-init.h>
#include <libgnomeui/gnome-uidefs.h>
#include <libgnomeui/gnome-ui-init.h>
#include <libgnomeui/gnome-window-icon.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-program.h>
#include <libgnome/gnome-init.h>
#include <gtk/gtkmain.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "ui.h"
#include "source-manager.h"
#include "library.h"
#include "queue.h"
#include "main.h"
#include "misc.h"
#include "icons.h"
#include "e-splash.h"
#include "rhythmbox-automation.h"
#include "eel-gconf-extensions.h"

#include "debug.h"

/* function prototypes */
static void rb_init (void);
static void rb_shutdown (void);
static void handle_command_line (void);

/* globals */
RhythmBox *rhythmbox;

/* command line argument parsing structure */
static gboolean play_option    = FALSE;
static gboolean pause_option   = FALSE;
static gboolean rewind_option  = FALSE;
static gboolean forward_option = FALSE;
static gboolean show_option    = FALSE;
static gboolean hide_option    = FALSE;
static gchar *add_filename     = NULL;
static gboolean applet_option  = FALSE;
static gboolean quit_option    = FALSE;

static struct poptOption popt_options[] =
{
	{ "play", 'p', POPT_ARG_NONE, &play_option, 0,
	  N_("Start playing"),
	  NULL },
	{ "pause", 'u', POPT_ARG_NONE, &pause_option, 0,
	  N_("Pause current song"),
	  NULL },
	{ "rewind", 'r', POPT_ARG_NONE, &rewind_option, 0,
	  N_("Skip backwards in playlist"),
	  NULL },
	{ "forward", 'f', POPT_ARG_NONE, &forward_option, 0,
	  N_("Skip forward in playlist"),
	  NULL },
	{ "show", 's', POPT_ARG_NONE, &show_option, 0,
	  N_("Show the RhythmBox window"),
	  NULL },
	{ "hide", 'h', POPT_ARG_NONE, &hide_option, 0,
          N_("Hide the RhythmBox window"),
	  NULL },
	{ "add-file", 'n', POPT_ARG_STRING, &add_filename, 0,
	  N_("Add a file to the library"),
	  N_("FILE") },
	{ "new-applet", 'a', POPT_ARG_NONE, &applet_option, 0,
	  N_("Open a new RhythmBox applet"),
	  NULL },
	{ "quit", 'q', POPT_ARG_NONE, &quit_option, 0,
	  N_("Quit a running RhythmBox"),
	  NULL },

	/* terminator, must be last */
	{ NULL, 0, 0, NULL, 0, NULL, NULL }
};

/**
 * main: main program function
 */
gint
main (gint argc, gchar **argv)
{
	/* init gnome */
	gnome_program_init ("rhythmbox", VERSION,
			    LIBGNOMEUI_MODULE, argc, argv, 
			    GNOME_PARAM_POPT_TABLE, popt_options,
			    GNOME_PARAM_HUMAN_READABLE_NAME, _("RhythmBox"),
			    NULL);

	/* init */
	rb_init ();

	/* start main loop */
	gtk_main ();

	/* shutdown rhythmbox */
	rb_shutdown ();

	return 0;
}

/**
 * init: start initializing rb
 */
static void 
rb_init (void)
{
	GdkPixbuf *ui_icon, *queue_icon, *library_icon;
	gchar *dirname, *fn;
	GtkWidget *splash;
	CORBA_Object factory;

	/* check wehther we are running already */
	factory = bonobo_activation_activate_from_id 
		("OAFIID:GNOME_RhythmBox_Automation_Factory",
		 Bonobo_ACTIVATION_FLAG_EXISTING_ONLY, 
		 NULL, NULL);

	if (factory == NULL)
	{
		if (quit_option == TRUE)
			exit (0);

		rhythmbox_automation_new ();
	}
	else
	{
		/* handle command line stuff */
		handle_command_line ();

		/* and we're done */
		exit (0);
	}

	/* thread support */
	if (!g_thread_supported ()) g_thread_init (NULL);
	
	/* debugging */
	make_warnings_and_criticals_stop_in_debugger ();

	/* init the splash screen */
	fn = user_file ("splash-ui.png", FALSE);
 	ui_icon = gdk_pixbuf_new_from_file (fn, NULL);
	g_free (fn);
	fn = user_file ("splash-queue.png", FALSE);
 	queue_icon = gdk_pixbuf_new_from_file (fn, NULL);
	g_free (fn);
	fn = user_file ("splash-library.png", FALSE);
 	library_icon = gdk_pixbuf_new_from_file (fn, NULL);
	g_free (fn);
 
 	splash = e_splash_new ();
 	g_signal_connect (G_OBJECT (splash), "delete_event",
 			  G_CALLBACK (gtk_widget_hide_on_delete), 0);
 	gtk_widget_show (splash);
 
 	e_splash_add_icon (E_SPLASH (splash), ui_icon);
 	e_splash_add_icon (E_SPLASH (splash), queue_icon);
 	e_splash_add_icon (E_SPLASH (splash), library_icon);
 
 	g_object_unref (G_OBJECT (ui_icon));
 	g_object_unref (G_OBJECT (queue_icon));
 	g_object_unref (G_OBJECT (library_icon));
 
	/* give the splash a chance to show itself */
 	while (gtk_events_pending ()) gtk_main_iteration ();

	/* init gnome-vfs */
	gnome_vfs_init ();

	/* set the default icon */
	gnome_window_icon_set_default_from_file (PIXMAP_DIR "/rhythmbox.png");

	/* init random generator */
	srandom (time (NULL));

	/* init gconf */
	eel_gconf_monitor_add ("/apps/rhythmbox");
	eel_gconf_monitor_add ("/desktop/gnome/interface");

	/* ensure ~/.gnome/rhythmbox/ exists */
	dirname = g_build_filename (g_get_home_dir(), GNOME_DOT_GNOME,
				    "rhythmbox", NULL);
	
	if (g_file_test (dirname, G_FILE_TEST_IS_DIR) == FALSE)
	{
		if (g_file_test (dirname, G_FILE_TEST_EXISTS) == TRUE)
		{
			gchar *msg;

			msg = g_strdup_printf (_("%s exists, please move this file out of the way, and restart Rhythmbox"), dirname);
			error_dialog (msg);
		}

		if (mkdir (dirname, 488) != 0)
                {
			gchar *msg;

                        msg = g_strdup_printf (_("Couldn't create Rhythmbox preferences directory (%s)\nReason invoked: %s"), dirname, strerror(errno));
			error_dialog_with_details
				(_("Couldn't create Rhythmbox preferences directory."),
				 msg);
                }
	}

	/* load icons */
	init_icon_factory ();

	/* create rhtyhbmox struct */
	rhythmbox = g_new0 (RhythmBox, 1);

	/* create a new player window object */
	rhythmbox->ui = ui_new ();

	/* done, update splash */
	e_splash_set_icon_highlight (E_SPLASH (splash), 0, TRUE);
	while (gtk_events_pending ()) gtk_main_iteration ();

	/* create a queue */
	rhythmbox->queue = queue_new ();

	/* done, update splash */
	e_splash_set_icon_highlight (E_SPLASH (splash), 1, TRUE);
	while (gtk_events_pending ()) gtk_main_iteration ();

	/* init the library */
	rhythmbox->library = library_new ();
	
	/* init the source manager */
	rhythmbox->manager = source_manager_new (rhythmbox->queue,
						 rhythmbox->ui,
						 rhythmbox->library);

	/* done, update splash */
	e_splash_set_icon_highlight (E_SPLASH (splash), 2, TRUE);
	while (gtk_events_pending ()) gtk_main_iteration ();

	/* done itializing, show ui & get rid of splashscreen */
	ui_set_visible (rhythmbox->ui, TRUE);

	/* hide splash */
	gtk_widget_hide (splash);

	/* handle command line stuff */
	handle_command_line ();
}

/**
 * rhythmbox_exit: exit the gtk main loop, which will make the program
 * terminate
 */
void
rhythmbox_exit (RhythmBox *rhythmbox)
{
	/* terminate the gtk main loop */
	gtk_main_quit ();
}

/**
 * shutdown: shutdown rhythmbox
 */
static void
rb_shutdown (void)
{
	/* hide the ui already, makes shutdown look faster :) */
	ui_set_visible (rhythmbox->ui, FALSE);
	
	/* quitting, unref the playlist and the window */
	g_object_unref (rhythmbox->manager);
	g_object_unref (rhythmbox->library);
	g_object_unref (rhythmbox->queue);
	g_object_unref (rhythmbox->ui);
	g_free (rhythmbox);

	/* shutdown gconf */
	eel_gconf_monitor_remove ("/apps/rhythmbox");
	eel_gconf_monitor_remove ("/desktop/gnome/interface");
}

/**
 * handle_command_line: handle command line arguments
 */
static void
handle_command_line (void)
{
	CORBA_Environment corba_env;
	GNOME_RhythmBoxAutomation server;

	CORBA_exception_init (&corba_env);

	server = bonobo_activation_activate_from_id ("OAFIID:GNOME_RhythmBox_Automation",
						     0, NULL, &corba_env);
	if (server == NULL)
		return;

	if (quit_option == TRUE)
	{
		GNOME_RhythmBoxAutomation_quit (server, &corba_env);
	}
	else
	{
		if (add_filename != NULL && g_file_test (add_filename, G_FILE_TEST_EXISTS) == TRUE)
			GNOME_RhythmBoxAutomation_addFile (server, add_filename, &corba_env);
		if (show_option == TRUE)
			GNOME_RhythmBoxAutomation_setVisible (server, TRUE, &corba_env);
		if (hide_option == TRUE)
			GNOME_RhythmBoxAutomation_setVisible (server, FALSE, &corba_env);
		if (applet_option == TRUE)
			GNOME_RhythmBoxAutomation_newApplet (server, &corba_env);
		if (play_option == TRUE)
			GNOME_RhythmBoxAutomation_play (server, &corba_env);
		if (pause_option == TRUE)
			GNOME_RhythmBoxAutomation_pause (server, &corba_env);
		if (rewind_option == TRUE)
			GNOME_RhythmBoxAutomation_rewind (server, &corba_env);
		if (forward_option == TRUE)
			GNOME_RhythmBoxAutomation_forward (server, &corba_env);
	}

	bonobo_object_release_unref (server, &corba_env);

	CORBA_exception_free (&corba_env);
}
