/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgnome/gnome-util.h>
#include <libgnome/gnome-i18n.h>
#include <bonobo/bonobo-dock-item.h>
#include <libgnomecanvas/gnome-canvas.h>
#include <libgnomecanvas/gnome-canvas-pixbuf.h>
#include <libgnomecanvas/gnome-canvas-text.h>
#include <libgnomecanvas/gnome-canvas-widget.h>
#include <libgnomevfs/gnome-vfs-uri.h>

#include "misc.h"
#include "control-bar.h"
#include "ui.h"
#include "icons.h"
#include "lcdhscale.h"
#include "rb-image-button.h"

struct _ControlBarPrivate
{
	/* lcd stuff */
	GtkWidget *lcd;
	GnomeCanvasItem *lcd_text;
	GnomeCanvasItem *lcd_elapsed;
	GnomeCanvasItem *lcd_state;
	GnomeCanvasItem *lcd_welcome;
	GtkWidget *lcd_seek;
	GtkObject *seek_adj;
	
	/* search entry */
	GtkWidget *entry;

	/* volume slider */
	GtkWidget *slider;
	GtkObject *vol_adj;

	/* play buttons etc */
	GtkWidget *rewind;
	GtkWidget *play;
	GtkWidget *forward;

	/* main widgets */
	GtkWidget *dock;
	GtkWidget *control_bar;

	/* merged buttons */
	GtkWidget *merged_buttons;
	GtkTooltips *tooltips;
};

/* ALL THe different signals */
enum
{
	PLAY_CLICKED,
	REWIND_CLICKED,
	FORWARD_CLICKED,
	SEARCH_ACTIVATE,
	VOLUME_CHANGED,
	PROGRESS_CHANGED,
	LAST_SIGNAL
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint control_bar_signals[LAST_SIGNAL] = { 0 };

static GdkPixbuf *play_pixbuf       = NULL;
static GdkPixbuf *pause_pixbuf      = NULL;
static GdkPixbuf *stop_pixbuf       = NULL;
static GdkPixbuf *lcd_toggle_pixbuf = NULL;

/* object funtion prototypes */
static void control_bar_class_init (ControlBarClass *klass);
static void control_bar_init (ControlBar *p);
static void control_bar_finalize (GObject *object);

/* ui callbacks */
static void rewind_clicked_cb (GtkWidget *button, ControlBar *p);
static void play_clicked_cb (GtkWidget *button, ControlBar *p);
static void forward_clicked_cb (GtkWidget *button, ControlBar *p);
static void volume_adjustment_value_changed_cb (GtkAdjustment *adj, ControlBar *p);
static void search_entry_changed_cb (GtkEditable *entry, ControlBar *p);
static void seek_adjustment_value_changed_cb (GtkAdjustment *adj, ControlBar *p);
GtkWidget *control_bar_create_rewind_button (void);
GtkWidget *control_bar_create_play_button (void);
GtkWidget *control_bar_create_forward_button (void);

/* local function prototypes */
static void colorify_lcd (GdkPixbuf *pixbuf, GtkWidget *stylewidget);
static char *build_string_for_display (const char *t, const char *a, const char *uri_text);

/**
 * control_bar_get_type: get the GObject type of the playbar 
 */
GType
control_bar_get_type (void)
{
	static GType control_bar_type = 0;

  	if (control_bar_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (ControlBarClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) control_bar_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (ControlBar),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) control_bar_init
      		};

      		control_bar_type = g_type_register_static (G_TYPE_OBJECT,
                				           "ControlBar",
                                           	           &our_info, 0);
    	}

	return control_bar_type;
}

/**
 * control_bar_class_init: initialize the ControlBar class
 */
static void
control_bar_class_init (ControlBarClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = control_bar_finalize;

	/* init signals */
	control_bar_signals[PLAY_CLICKED] =
   		g_signal_new ("play_clicked",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (ControlBarClass, play_clicked),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__INT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_INT);
	control_bar_signals[REWIND_CLICKED] =
   		g_signal_new ("rewind_clicked",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (ControlBarClass, rewind_clicked),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	control_bar_signals[FORWARD_CLICKED] =
   		g_signal_new ("forward_clicked",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (ControlBarClass, forward_clicked),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	control_bar_signals[SEARCH_ACTIVATE] =
   		g_signal_new ("search_activate",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (ControlBarClass, search_activate),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_STRING);
	control_bar_signals[VOLUME_CHANGED] =
   		g_signal_new ("volume_changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (ControlBarClass, volume_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__FLOAT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_FLOAT);
	control_bar_signals[PROGRESS_CHANGED] =
   		g_signal_new ("progress_changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (ControlBarClass, progress_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__FLOAT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_FLOAT);
}

/**
 * control_bar_init: intialize the playbar object
 */
static void
control_bar_init (ControlBar *p)
{
	p->priv = g_new0 (ControlBarPrivate, 1);

}

/**
 * control_bar_finalize: finalize the playbar object
 */
static void
control_bar_finalize (GObject *object)
{
	ControlBar *p;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_CONTROL_BAR (object));
	
   	p = CONTROL_BAR (object);

	g_return_if_fail (p->priv != NULL);

	g_free (p->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * control_bar_new: create a new playbar object
 */
ControlBar *
control_bar_new (GladeXML *gxml)
{
 	ControlBar *p;

	p = CONTROL_BAR (g_object_new (TYPE_CONTROL_BAR, NULL));

	control_bar_init_interface (p, gxml);

	g_return_val_if_fail (p->priv != NULL, NULL);

	return p;
}

void
control_bar_set_ui (ControlBar *t, UI *ui)
{
	t->ui = ui;
}

UI *
control_bar_get_ui (ControlBar *t)
{
	return t->ui;
}

/**
 * control_bar_merge_widgets: add a list of widgets to the bar,
 * discarding the old set
 */
void
control_bar_merge_widgets (ControlBar *t,
			   GList *widgets)
{
	GList *l, *children;

	children = gtk_container_get_children (GTK_CONTAINER (t->priv->merged_buttons));

	for (l = children; l != NULL; l = g_list_next (l))
	{
		GtkWidget *widget = l->data;
		gtk_container_remove (GTK_CONTAINER (t->priv->merged_buttons), widget);
	}

	for (l = widgets; l != NULL; l = g_list_next (l))
	{
		GtkWidget *widget = l->data;
		gtk_widget_show_all (widget);
		gtk_box_pack_start (GTK_BOX (t->priv->merged_buttons),
			            widget, FALSE, FALSE, 0);
	}
}

/**
 * control_bar_new_action_button: create a new action button
 */
GtkWidget *
control_bar_new_action_button (ControlBar *t,
			       const char *label,
			       const char *tooltip,
			       GdkPixbuf *icon,
			       gboolean toggle)
{
	GtkWidget *button = toggle ? gtk_toggle_button_new () : gtk_button_new ();
	//GtkWidget *hbox = gtk_vbox_new (FALSE, FALSE);
	//GtkWidget *glabel = gtk_label_new_with_mnemonic (label);
	GtkWidget *image = gtk_image_new_from_pixbuf (icon);

	//gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
	//gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 0);

	gtk_container_add (GTK_CONTAINER (button), image);

	gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);

	gtk_tooltips_set_tip (t->priv->tooltips, button, tooltip, NULL);

	return button;
}

/**
 * control_bar_set_lcd_song: set the text of the upper row of text in
 * the lcd
 */
void
control_bar_set_lcd_song (ControlBar *t, const gchar *title, const char *artist, const char *uri)
{
	char *song = NULL;

	song = build_string_for_display (title, artist, uri);

	if (song != NULL)
	{
		char *tmp;
		if (strlen (song) > 40)
		{
			tmp = g_strdup_printf ("<big>%.19s...%s</big>",
					       song, song + strlen (song) - 8);
		}
		else
		{
			tmp = g_strdup_printf ("<big>%s</big>", song);
		}
		gnome_canvas_item_set (t->priv->lcd_text,
				       "markup", tmp, NULL);
		gnome_canvas_item_set (t->priv->lcd_elapsed,
				       "text", "", NULL);
		gnome_canvas_item_show (t->priv->lcd_text);
		gnome_canvas_item_show (t->priv->lcd_elapsed);
		gtk_widget_show (t->priv->lcd_seek);
		gnome_canvas_item_hide (t->priv->lcd_welcome);
		g_free (tmp);
	}
	else
	{
		gnome_canvas_item_set (t->priv->lcd_text,
				       "markup", "", NULL);
		gnome_canvas_item_set (t->priv->lcd_elapsed,
				       "text", "", NULL);
		gnome_canvas_item_hide (t->priv->lcd_text);
		gnome_canvas_item_hide (t->priv->lcd_elapsed);
		gtk_widget_hide (t->priv->lcd_seek);
		gnome_canvas_item_show (t->priv->lcd_welcome);
	}

	g_free (song);
}

/**
 * control_bar_set_lcd_time: update the time bit of the lcd 
 */
void
control_bar_set_lcd_time (ControlBar *t, long elapsed, float progress)
{
	char *time;
	int curseconds = (int) elapsed; // FIXME

	if (curseconds == -1)
	{
		time = g_strdup ("");
	}
	else
	{
		gint minutes = curseconds / 60;
		gint seconds = curseconds % 60;
		time = g_strdup_printf (_("Elapsed Time: %d:%.2d"), minutes, seconds);
	}

	/* update elapsed time label */
	gnome_canvas_item_set (t->priv->lcd_elapsed,
			       "text", time, NULL);

	/* update silder */
	g_signal_handlers_block_by_func (G_OBJECT (t->priv->seek_adj),
				         G_CALLBACK (seek_adjustment_value_changed_cb),
				         t);
	gtk_adjustment_set_value (GTK_ADJUSTMENT (t->priv->seek_adj),
				  progress);
	g_signal_handlers_unblock_by_func (G_OBJECT (t->priv->seek_adj),
				           G_CALLBACK (seek_adjustment_value_changed_cb),
				           t);

	g_free (time);
}

/**
 * control_bar_set_state: set the playbar to this state
 */
void
control_bar_set_state (ControlBar *p, UIState state)
{
	p->state = state;

	switch (state)
	{
	case UI_STATE_PLAYING:
		rb_image_button_set_normal_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						   pause_pixbuf);
		rb_image_button_set_depressed_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						      pause_pixbuf);
		break;
	case UI_STATE_PLAYING_WITH_STOP:
		rb_image_button_set_normal_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						   stop_pixbuf);
		rb_image_button_set_depressed_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						      stop_pixbuf);
		break;
	case UI_STATE_PAUSED:
		rb_image_button_set_normal_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						   play_pixbuf);
		rb_image_button_set_depressed_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						      play_pixbuf);
		break;
	default:
		rb_image_button_set_normal_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						   play_pixbuf);
		rb_image_button_set_depressed_pixbuf (RB_IMAGE_BUTTON (p->priv->play),
						      play_pixbuf);
		control_bar_set_lcd_song (p, NULL, NULL, NULL);
		break;
	}
}

/**
 * control_bar_get_state: return the control_bar's state
 */
UIState
control_bar_get_state (ControlBar *t)
{
	return t->state;
}

/**
 * control_bar_set_rewind_sensitive: set the rewind button sensitive
 */
void
control_bar_set_rewind_sensitive (ControlBar *t, gboolean sensitive)
{
	gtk_widget_set_sensitive (t->priv->rewind, sensitive);
}

/**
 * control_bar_get_rewind_sensitive: is the rewind button sensitive?
 */
gboolean
control_bar_get_rewind_sensitive (ControlBar *t)
{
	return GTK_WIDGET_SENSITIVE (t->priv->rewind);
}

/**
 * control_bar_set_forward_sensitive: set the forward button sensitive
 */
void
control_bar_set_forward_sensitive (ControlBar *t, gboolean sensitive)
{
	gtk_widget_set_sensitive (t->priv->forward, sensitive);
}

/**
 * control_bar_get_forward_sensitive: is the forward button sensitive?
 */
gboolean
control_bar_get_forward_sensitive (ControlBar *t)
{
	return GTK_WIDGET_SENSITIVE (t->priv->forward);
}

/**
 * control_bar_set_volume: change the volume in the ui
 */
void
control_bar_set_volume (ControlBar *t, gfloat volume)
{
	gtk_adjustment_set_value (GTK_ADJUSTMENT (t->priv->vol_adj), volume);
}

/**
 * control_bar_get_volume: return the volume currently set in the ui
 */
gfloat
control_bar_get_volume (ControlBar *t)
{
	return gtk_adjustment_get_value (GTK_ADJUSTMENT (t->priv->vol_adj));
}

/**
 * rewind_clicked_cb: rewind clicked
 */
static void
rewind_clicked_cb (GtkWidget *button, ControlBar *p)
{
	g_signal_emit (G_OBJECT (p), control_bar_signals[REWIND_CLICKED], 0);
}

/**
 * play_clicked_cb: play clicked
 */
static void
play_clicked_cb (GtkWidget *button, ControlBar *p)
{
	ControlBarPlayAction action = CONTROL_BAR_PLAY_ACTION_PLAY;
	switch (p->state)
	{
	case UI_STATE_PLAYING:
		action = CONTROL_BAR_PLAY_ACTION_PAUSE;
		break;
	case UI_STATE_PLAYING_WITH_STOP:
		action = CONTROL_BAR_PLAY_ACTION_STOP;
		break;
	case UI_STATE_PAUSED:
	case UI_STATE_STOPPED:
	case UI_STATE_PROCESSING:
		action = CONTROL_BAR_PLAY_ACTION_PLAY;
		break;
	default:
		break;
	}
	g_signal_emit (G_OBJECT (p), control_bar_signals[PLAY_CLICKED], 0,
		       action);
}

/**
 * forward_clicked_cb: forward clicked
 */
static void
forward_clicked_cb (GtkWidget *button, ControlBar *p)
{
	g_signal_emit (G_OBJECT (p), control_bar_signals[FORWARD_CLICKED], 0);
}

/**
 * volume_adjustment_value_changed_cb: value in the volume slider changed
 */
static void
volume_adjustment_value_changed_cb (GtkAdjustment *adj, ControlBar *p)
{
	g_signal_emit (G_OBJECT (p), control_bar_signals[VOLUME_CHANGED], 0,
		       gtk_adjustment_get_value (adj));
}

/**
 * seek_adjustment_value_changed_cb: value in the seek slider changed
 */
static void
seek_adjustment_value_changed_cb (GtkAdjustment *adj, ControlBar *p)
{
	g_signal_emit (G_OBJECT (p), control_bar_signals[PROGRESS_CHANGED], 0,
		       gtk_adjustment_get_value (adj));
}

/**
 * search_entry_changed_cb: search entry changed
 */
static void
search_entry_changed_cb (GtkEditable *entry, ControlBar *p)
{
	gboolean val = FALSE;
	gchar *str = gtk_editable_get_chars (entry, 0, -1);
	g_signal_emit (G_OBJECT (p), control_bar_signals[SEARCH_ACTIVATE], 0,
		       str, &val);
	g_free (str);
}

/**
 * control_bar_clear_search_entry: empty search entry
 */
void
control_bar_clear_search_entry (ControlBar *p)
{
	g_signal_handlers_block_by_func (G_OBJECT (p->priv->entry),
				         G_CALLBACK (search_entry_changed_cb),
				         p);
	gtk_editable_delete_text (GTK_EDITABLE (p->priv->entry), 0, -1);
	g_signal_handlers_unblock_by_func (G_OBJECT (p->priv->entry),
				           G_CALLBACK (search_entry_changed_cb),
				           p);
}

/**
 * colorify_lcd: substitute purple color in @pixbuf with the theme color
 */
static void
colorify_lcd (GdkPixbuf *pixbuf, GtkWidget *stylewidget)
{
	gint width, height, rowstride;
	guchar r, g, b, a;
	guchar purple_r, purple_g, purple_b;
	guchar shadow_black_r, shadow_black_g, shadow_black_b;
	guchar shadow_white_r, shadow_white_g, shadow_white_b;
	guchar *buffer;
	gint x, y, i;
	GdkColor tmp;
	GtkStyle *rcstyle;

	/* get pixbuf properties */
	width  = gdk_pixbuf_get_width (pixbuf);
	height = gdk_pixbuf_get_height (pixbuf);
	buffer = gdk_pixbuf_get_pixels (pixbuf);
	rowstride = gdk_pixbuf_get_rowstride (pixbuf);

	/* get replacement colors */
	rcstyle = gtk_rc_get_style (stylewidget);
	if (rcstyle == NULL)
		rcstyle = stylewidget->style;
	tmp = rcstyle->base[GTK_STATE_ACTIVE];
	purple_r       = tmp.red   >> 8;
	purple_g       = tmp.green >> 8;
	purple_b       = tmp.blue  >> 8;
	tmp = rcstyle->dark[GTK_STATE_NORMAL];
	shadow_black_r = tmp.red   >> 8;
	shadow_black_g = tmp.green >> 8;
	shadow_black_b = tmp.blue  >> 8;
	tmp = rcstyle->light[GTK_STATE_NORMAL];
	shadow_white_r = tmp.red   >> 8;
	shadow_white_g = tmp.green >> 8;
	shadow_white_b = tmp.blue  >> 8;
	
	/* recolor image */
	for (i = 0, y = 0; y < height; y++, i += rowstride - (width * 4))
	{
		for (x = 0; x < width; x++, i += 4)
		{
			/* get pixel */
			r = buffer[i + 0];
			g = buffer[i + 1];
			b = buffer[i + 2];
			a = buffer[i + 3];

			if (r == 255 && g == 0 && b == 255)
			{
				/* purple, replace */
				r = purple_r;
				g = purple_g;
				b = purple_b;
			}

			if (r == 0 && g == 0 && b == 0)
			{
				/* black, replace */
				r = shadow_black_r;
				g = shadow_black_g;
				b = shadow_black_b;
			}

			if (r == 255 && g == 255 && b == 255)
			{
				/* white, replace */
				r = shadow_white_r;
				g = shadow_white_g;
				b = shadow_white_b;
			}

			/* writeback pixel */
			buffer[i + 0] = r;
			buffer[i + 1] = g;
			buffer[i + 2] = b;
			buffer[i + 3] = a;
		}
	}
}

static char *
build_string_for_display (const char *t, const char *a, const char *uri_text)
{
	char *tmp = NULL;
	GnomeVFSURI *uri;

	if (t != NULL && a != NULL)
	{
		tmp = g_strdup_printf ("%s - %s", a, t);
	}

	if (tmp != NULL && strlen (tmp) > 40)
	{
		g_free (tmp);
		tmp = g_strdup (t);
	}

	if (t != NULL && a == NULL)
	{
		tmp = g_strdup (t);
	}

	if (a != NULL && t == NULL)
	{
		tmp = g_strdup (a);
	}

	if (a == NULL && t == NULL && uri_text != NULL)
	{
		uri = gnome_vfs_uri_new (uri_text);
		tmp = gnome_vfs_uri_extract_short_name (uri);
		gnome_vfs_uri_unref (uri);
	}

	return tmp;
}

/**
 * render_stock_pixbufs_if_necessary: render stock icons with widget if they aren't already
 */
static void
render_stock_pixbufs_if_necessary (GtkWidget *widget)
{
	/* load images */
	if (!play_pixbuf) 
	{
		play_pixbuf = gtk_widget_render_icon (widget,
						      RHYTHMBOX_STOCK_PLAY,
						      RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						      NULL);
	}

	/* FIXME unref these */
	if (!pause_pixbuf)
	{
		pause_pixbuf = gtk_widget_render_icon (widget,
						       RHYTHMBOX_STOCK_PAUSE,
						       RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						       NULL);
	}

	if (!stop_pixbuf)
	{
		stop_pixbuf = gtk_widget_render_icon (widget,
						      RHYTHMBOX_STOCK_STOP,
						      RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						      NULL);
	}

	if (!lcd_toggle_pixbuf) 
	{
		lcd_toggle_pixbuf = gtk_widget_render_icon (widget,
						            RHYTHMBOX_STOCK_LCD_TOGGLE,
						            RHYTHMBOX_ICON_SIZE_LCD,
						            NULL);
	}
}

void
control_bar_init_interface (ControlBar *p, GladeXML *gxml)
{
	GtkWidget *control_bar;
	GtkImage *image;
	GtkWidget *lcd_box;
	GdkPixbuf *lcd;
	char *font;
	GtkStyle *rcstyle;

	char *fn;
	GnomeCanvasGroup *root;

	control_bar = glade_xml_get_widget (gxml, "control_bar_box");
	render_stock_pixbufs_if_necessary (control_bar);

	/* Setup images in the glade file */
	image = GTK_IMAGE (glade_xml_get_widget (gxml, "no_volume_image"));
	gtk_image_set_from_stock (image, RHYTHMBOX_STOCK_VOLUME_OFF, GTK_ICON_SIZE_MENU);
	
	image = GTK_IMAGE (glade_xml_get_widget (gxml, "full_volume_image"));
	gtk_image_set_from_stock (image, RHYTHMBOX_STOCK_VOLUME, GTK_ICON_SIZE_MENU);

	/* Get references to the buttons */
	p->priv->rewind  = glade_xml_get_widget (gxml, "rewind_button");
	p->priv->forward = glade_xml_get_widget (gxml, "forward_button");
	p->priv->play    = glade_xml_get_widget (gxml, "play_button");

	/* Connect to button signals */
	g_signal_connect (G_OBJECT (p->priv->rewind), "clicked",
			  G_CALLBACK (rewind_clicked_cb), p);
	g_signal_connect (G_OBJECT (p->priv->forward), "clicked",
			  G_CALLBACK (forward_clicked_cb), p);
	g_signal_connect (G_OBJECT (p->priv->play), "clicked",
			  G_CALLBACK (play_clicked_cb), p);

	/* Get the volume slider and its adjustment */
	p->priv->slider  = glade_xml_get_widget (gxml, "volume_slider");
	p->priv->vol_adj = GTK_OBJECT (gtk_range_get_adjustment (GTK_RANGE (p->priv->slider)));
	g_signal_connect (G_OBJECT (p->priv->vol_adj), "value_changed",
			  G_CALLBACK (volume_adjustment_value_changed_cb), p);

	/* create lcd canvas */
	fn = user_file ("lcd.png", FALSE);
	lcd = gdk_pixbuf_new_from_file (fn, NULL);
	colorify_lcd (lcd, control_bar);
	g_free (fn);
	
	gtk_widget_push_colormap (gdk_rgb_get_cmap ());
	p->priv->lcd = gnome_canvas_new_aa ();
	gtk_widget_set_usize (p->priv->lcd, 336, 57);
	gnome_canvas_set_center_scroll_region (GNOME_CANVAS (p->priv->lcd), FALSE);
	root = gnome_canvas_root (GNOME_CANVAS (p->priv->lcd));
	gnome_canvas_set_pixels_per_unit (GNOME_CANVAS (p->priv->lcd), 1.0);
	gnome_canvas_item_new (root, 
			       gnome_canvas_pixbuf_get_type (),
			       "pixbuf", lcd,
			       "x", 0.0,
			       "y", 0,0,
			       "anchor", GTK_ANCHOR_NW,
			       NULL);

	rcstyle = gtk_rc_get_style (p->priv->lcd);
	if (rcstyle == NULL)
		rcstyle = p->priv->lcd->style;

	font = pango_font_description_to_string (rcstyle->font_desc);

	p->priv->lcd_text = gnome_canvas_item_new (root,
			       gnome_canvas_text_get_type (),
			       "text", "",
			       "x", 168.0,
			       "y", 13.0,
			       "font", font,
			       "anchor", GTK_ANCHOR_CENTER,
			       "justification", GTK_JUSTIFY_CENTER,
			       "fill_color", "black",
			       NULL);

	p->priv->lcd_elapsed = gnome_canvas_item_new (root,
			       gnome_canvas_text_get_type (),
			       "text", "",
			       "x", 168.0,
			       "y", 33.0,
			       "font", font,
			       "anchor", GTK_ANCHOR_CENTER,
			       "justification", GTK_JUSTIFY_CENTER,
			       "fill_color", "black",
			       NULL);

	p->priv->lcd_welcome = gnome_canvas_item_new (root,
			       gnome_canvas_text_get_type (),
			       "markup", _("<big>Welcome to RhythmBox !</big>"),
			       "x", 168.0,
			       "y", 28.0,
			       "font", font,
			       "anchor", GTK_ANCHOR_CENTER,
			       "justification", GTK_JUSTIFY_CENTER,
			       "fill_color", "black",
			       NULL);

	g_free (font);

	p->priv->lcd_state = gnome_canvas_item_new (root,
			       gnome_canvas_pixbuf_get_type (),
			       //"pixbuf", lcd_toggle_pixbuf,
			       "x", 7.0,
			       "y", 24.0,
			       NULL);

	p->priv->seek_adj = gtk_adjustment_new (0, 0, 1.0, 0.01, 0.1, 0.0);
	g_signal_connect (G_OBJECT (p->priv->seek_adj), "value_changed",
			  G_CALLBACK (seek_adjustment_value_changed_cb), p);
	p->priv->lcd_seek = lcd_hscale_new (GTK_ADJUSTMENT (p->priv->seek_adj));
	gnome_canvas_item_new (root,
			       gnome_canvas_widget_get_type (),
			       "widget", p->priv->lcd_seek,
			       "x", 168.0,
			       "y", 47.0,
			       "width", 200.0,
			       "height", 7.0,
			       "anchor", GTK_ANCHOR_CENTER,
			       NULL);

	gtk_widget_pop_colormap ();
	
	g_object_unref (G_OBJECT (lcd));

	/* pack the lcd */
	lcd_box = glade_xml_get_widget (gxml, "lcd_box");
	gtk_container_add (GTK_CONTAINER (lcd_box), p->priv->lcd);

	/* create search entry */
	p->priv->entry = glade_xml_get_widget (gxml, "search_entry");
	g_signal_connect (G_OBJECT (p->priv->entry), "changed",
			  G_CALLBACK (search_entry_changed_cb), p);

	/* prev/next insensitive by default */
	control_bar_set_forward_sensitive (p, FALSE);
	control_bar_set_rewind_sensitive (p, FALSE);

	/* current action button */
	p->priv->merged_buttons = glade_xml_get_widget (gxml, "action_button_box");
	p->priv->tooltips = gtk_tooltips_new ();
}

/**
 * control_bar_create_rewind_button: feed the rewind button to glade
 */
GtkWidget *
control_bar_create_rewind_button (void)
{
	GtkWidget *widget = gtk_label_new ("dummy");
	GdkPixbuf *normal = gtk_widget_render_icon (widget,
						    RHYTHMBOX_STOCK_REWIND,
						    RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						    NULL);
	GdkPixbuf *depressed = gtk_widget_render_icon (widget,
						       RHYTHMBOX_STOCK_REWIND,
						       RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						       NULL);
	gtk_widget_destroy (widget);
	return rb_image_button_new (normal, depressed);
}

/**
 * control_bar_create_play_button: feed the play button to glade
 */
GtkWidget *
control_bar_create_play_button (void)
{
	GtkWidget *widget = gtk_label_new ("dummy");
	GdkPixbuf *normal = gtk_widget_render_icon (widget,
						    RHYTHMBOX_STOCK_PLAY,
						    RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						    NULL);
	GdkPixbuf *depressed = gtk_widget_render_icon (widget,
						       RHYTHMBOX_STOCK_PLAY,
						       RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						       NULL);
	gtk_widget_destroy (widget);
	return rb_image_button_new (normal, depressed);
}

/**
 * control_bar_create_forward_button: feed the forward button to glade
 */
GtkWidget *
control_bar_create_forward_button (void)
{
	GtkWidget *widget = gtk_label_new ("dummy");
	GdkPixbuf *normal = gtk_widget_render_icon (widget,
						    RHYTHMBOX_STOCK_FORWARD,
						    RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						    NULL);
	GdkPixbuf *depressed = gtk_widget_render_icon (widget,
						       RHYTHMBOX_STOCK_FORWARD,
						       RHYTHMBOX_ICON_SIZE_CONTROL_BAR,
						       NULL);
	gtk_widget_destroy (widget);
	return rb_image_button_new (normal, depressed);
}
