/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgnome/gnome-util.h>
#include <libgnome/gnome-i18n.h>
#include <bonobo/bonobo-dock-item.h>
#include <gdk/gdkkeysyms.h>

#include "misc.h"
#include "prefs.h"
#include "ui.h"
#include "glade.h"
#include "file-watcher.h"
#include "songs-view.h"
#include "eel-gconf-extensions.h"

struct _PrefsPrivate
{
	GtkWidget *window;

	/* general widgets */
	GtkWidget *help;
	GtkWidget *cancel;
	GtkWidget *ok;
	GtkWidget *notebook;

	/* dirs page */
	GtkWidget *dir_entry;
	GtkWidget *dir_add;
	GtkWidget *dir_remove;
	GtkWidget *dir_view;
	GtkListStore *dir_store;

	/* columns page */
	GtkWidget *columns_view;
	GtkListStore *columns_store;
};

typedef enum
{
	DIR_COL_URI   = 0,
	DIR_COL_COUNT = 1
} DirColumn;

typedef enum
{
	COL_COL_CHECKED = 0,
	COL_COL_NAME    = 1,
	COL_COL_COUNT   = 2
} ColColumn;

/* globals */
static GObjectClass *parent_class = NULL;

/* from songs-view.c */
extern char *column_strings[];

/* object funtion prototypes */
static void prefs_class_init (PrefsClass *klass);
static void prefs_init (Prefs *s);
static void prefs_finalize (GObject *object);

/* ui callbacks */
gboolean prefs_delete_event_cb (GtkWidget *widget, GdkEventAny *event, Prefs *p);
void prefs_ok_clicked_cb (GtkWidget *widget, Prefs *p);
void prefs_cancel_clicked_cb (GtkWidget *widget, Prefs *p);
void prefs_help_clicked_cb (GtkWidget *widget, Prefs *p);
void prefs_add_clicked_cb (GtkWidget *widget, Prefs *p);
void prefs_remove_clicked_cb (GtkWidget *widget, Prefs *p);
void prefs_dir_entry_activate_cb (GtkWidget *w, Prefs *p);
gboolean prefs_dir_view_key_press_event_cb (GtkWidget *w, GdkEventKey *event, Prefs *p);
static void dir_view_selection_changed_cb (GtkTreeSelection *selection, Prefs *p);
static void columns_row_toggled_cb (GtkCellRendererToggle *toggle, const gchar *path_str,
				    Prefs *p);

/* local function prototypes */
static void prefs_close (Prefs *p);
static void list_iters (GtkTreeModel *model, GtkTreePath *path,
			GtkTreeIter *iter, void **data);
static void prefs_save (Prefs *p);
static gboolean list_uris (GtkTreeModel *model, GtkTreePath *path,
			   GtkTreeIter *iter, void **data);
static gboolean list_cols (GtkTreeModel *model, GtkTreePath *path,
			   GtkTreeIter *iter, void **data);

/**
 * prefs_get_type: get the GObject type of the prefs
 */
GType
prefs_get_type (void)
{
	static GType prefs_type = 0;

  	if (prefs_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (PrefsClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) prefs_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (Prefs),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) prefs_init
      		};

      		prefs_type = g_type_register_static (G_TYPE_OBJECT,
                				     "Prefs",
                                           	     &our_info, 0);
    	}

	return prefs_type;
}

/**
 * prefs_class_init: initialize the Prefs class
 */
static void
prefs_class_init (PrefsClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = prefs_finalize;
}

/**
 * prefs_init: intialize the prefs object
 */
static void
prefs_init (Prefs *s)
{
	GladeXML *gxml;
	GtkCellRenderer *renderer;
	
	s->priv = g_new0 (PrefsPrivate, 1);

	/* get the glade file */
	gxml = glade_widget_new ("prefs.glade",
				 "prefs",
				 &(s->priv->window), s);
	
	s->priv->help       = glade_xml_get_widget (gxml, "help_button");
	s->priv->cancel     = glade_xml_get_widget (gxml, "cancel_button");
	s->priv->ok         = glade_xml_get_widget (gxml, "ok_button");
	s->priv->notebook   = glade_xml_get_widget (gxml, "notebook");

	s->priv->dir_entry  = glade_xml_get_widget (gxml, "dir_entry");
	s->priv->dir_add    = glade_xml_get_widget (gxml, "add_button");
	s->priv->dir_remove = glade_xml_get_widget (gxml, "remove_button");
	s->priv->dir_view   = glade_xml_get_widget (gxml, "dir_view");

	s->priv->columns_view = glade_xml_get_widget (gxml, "columns_view");

	/* set up dirs treeview */
	s->priv->dir_store = gtk_list_store_new (DIR_COL_COUNT, G_TYPE_STRING);
	gtk_tree_view_set_model (GTK_TREE_VIEW (s->priv->dir_view),
				 GTK_TREE_MODEL (s->priv->dir_store));
	gtk_tree_selection_set_mode
		(gtk_tree_view_get_selection (GTK_TREE_VIEW (s->priv->dir_view)),
		 GTK_SELECTION_MULTIPLE);
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (s->priv->dir_view),
						     DIR_COL_URI, _("Location"),
						     renderer,
						     "text", DIR_COL_URI,
						     NULL);
	g_signal_connect
		(G_OBJECT (gtk_tree_view_get_selection (GTK_TREE_VIEW (s->priv->dir_view))),
		 "changed", G_CALLBACK (dir_view_selection_changed_cb), s);
	gtk_widget_set_sensitive (s->priv->dir_remove, FALSE);

	/* set up columns treeview */
	s->priv->columns_store = gtk_list_store_new (COL_COL_COUNT,
						     G_TYPE_BOOLEAN,
						     G_TYPE_STRING);
	gtk_tree_view_set_model (GTK_TREE_VIEW (s->priv->columns_view),
				 GTK_TREE_MODEL (s->priv->columns_store));
	gtk_tree_selection_set_mode
		(gtk_tree_view_get_selection (GTK_TREE_VIEW (s->priv->columns_view)),
		 GTK_SELECTION_BROWSE);
	renderer = gtk_cell_renderer_toggle_new ();
	gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (s->priv->columns_view),
						     COL_COL_CHECKED, NULL,
						     renderer,
						     "active", COL_COL_CHECKED,
						     NULL);
	g_signal_connect (G_OBJECT (renderer), "toggled",
			  G_CALLBACK (columns_row_toggled_cb), s);
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (s->priv->columns_view),
						     COL_COL_NAME, NULL,
						     renderer,
						     "text", COL_COL_NAME,
						     NULL);

	/* cleanup */
	g_object_unref (G_OBJECT (gxml));
}

/**
 * prefs_finalize: finalize the prefs object
 */
static void
prefs_finalize (GObject *object)
{
	Prefs *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_PREFS (object));
	
   	s = PREFS (object);

	g_return_if_fail (s->priv != NULL);

	gtk_widget_destroy (s->priv->window);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * prefs_new: create a new prefs object
 */
Prefs *
prefs_new (void)
{
 	Prefs *s;

	s = PREFS (g_object_new (TYPE_PREFS, NULL));

	g_return_val_if_fail (s->priv != NULL, NULL);

	return s;
}

/**
 * prefs_show: show the prefswindow
 */
void
prefs_show (Prefs *p)
{
	GSList *uris, *l, *cols;
	gint i;

	if (GTK_WIDGET_VISIBLE (p->priv->window)) return; 

	/* load watched dirs */
	uris = eel_gconf_get_string_list (CONF_FILE_WATCHER_URIS);

	for (l = uris; l != NULL; l = g_slist_next (l))
	{
		gchar *uri = l->data;
		GtkTreeIter iter;

		gtk_list_store_append (p->priv->dir_store, &iter);
		gtk_list_store_set (p->priv->dir_store, &iter,
				    DIR_COL_URI, uri,
				    -1);
		
		g_free (uri);
	}

	g_slist_free (uris);

	cols = eel_gconf_get_integer_list (CONF_SONGS_VIEW_COLUMNS);

	/* load columns */
	/* note that we start from one, we dont want to have
	 * the song column configurable */
	for (i = 1; column_strings[i] != NULL; i++)
	{
		GtkTreeIter iter;
		l = g_slist_nth (cols, i);

		gtk_list_store_append (p->priv->columns_store, &iter);
		gtk_list_store_set (p->priv->columns_store, &iter,
				    COL_COL_CHECKED, GPOINTER_TO_INT (l->data),
				    COL_COL_NAME, _(column_strings[i]),
				    -1);
	}

	g_slist_free (cols);
	
	/* show window */
	gtk_widget_show (p->priv->window);
}

/**
 * prefs_delete_event_cb: dialgo has been closed
 */
gboolean
prefs_delete_event_cb (GtkWidget *widget, GdkEventAny *event, Prefs *p)
{
	prefs_close (p);
	return TRUE;
}

/**
 * prefs_close: hide the prefswin
 */
static void
prefs_close (Prefs *p)
{
	eel_gconf_suggest_sync ();
	
	gtk_widget_hide (p->priv->window);

	gtk_list_store_clear (p->priv->dir_store);
	gtk_list_store_clear (p->priv->columns_store);
}

/**
 * prefs_ok_clicked_cb: ok clicked
 */
void
prefs_ok_clicked_cb (GtkWidget *widget, Prefs *p)
{
	prefs_close (p);
}

/**
 * prefs_cancel_clicked_cb: cancel clicked
 */
void
prefs_cancel_clicked_cb (GtkWidget *widget, Prefs *p)
{
	/* FIXME should discard changes */
	prefs_close (p);
}

/**
 * prefs_help_clicked_cb: help button clicked
 */
void
prefs_help_clicked_cb (GtkWidget *widget, Prefs *p)
{
	/* not yet implemented */
}

/**
 * prefs_add_clicked_cb: add the uri from the entry
 */
void
prefs_add_clicked_cb (GtkWidget *widget, Prefs *p)
{
	gchar *uri = gtk_editable_get_chars (GTK_EDITABLE (p->priv->dir_entry), 0, -1);
	GtkTreeIter iter;

	if (uri == NULL || strlen (uri) < 1) return;

	gtk_list_store_append (p->priv->dir_store, &iter);
	gtk_list_store_set (p->priv->dir_store, &iter,
			    DIR_COL_URI, uri,
			    -1);
	g_free (uri);
	prefs_save (p);
}

/**
 * prefs_remove_clicked_cb: remove selected uris
 */
void
prefs_remove_clicked_cb (GtkWidget *widget, Prefs *p)
{
	GList *list = NULL, *l;
	gtk_tree_selection_selected_foreach
		(gtk_tree_view_get_selection (GTK_TREE_VIEW (p->priv->dir_view)),
		 (GtkTreeSelectionForeachFunc) list_iters, (void **) &list); 
	for (l = list; l != NULL; l = g_list_next (l))
	{
		GtkTreeRowReference *r = l->data;
		GtkTreeIter iter;
		gtk_tree_model_get_iter (GTK_TREE_MODEL (p->priv->dir_store),
					 &iter, gtk_tree_row_reference_get_path (r));
		gtk_list_store_remove (p->priv->dir_store, &iter);
		gtk_tree_row_reference_free (r);
	}
	prefs_save (p);
}

/**
 * list_iters: removes a row from the view
 */
static void
list_iters (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data)
{
	GList **ret = (GList **) data;
	*ret = g_list_append (*ret, gtk_tree_row_reference_new (model, path)); 
}

/**
 * prefs_save: save changes
 */
static void
prefs_save (Prefs *p)
{
	GSList *list = NULL;

	/* save watched dirs */
	gtk_tree_model_foreach (GTK_TREE_MODEL (p->priv->dir_store),
				(GtkTreeModelForeachFunc) list_uris, (void **) &list);
	eel_gconf_set_string_list (CONF_FILE_WATCHER_URIS, list);
	g_slist_free (list);

	/* save columns */
	list = NULL;
	list = g_slist_append (list, GINT_TO_POINTER (TRUE)); /* the songs column is always shown */
	gtk_tree_model_foreach (GTK_TREE_MODEL (p->priv->columns_store),
				(GtkTreeModelForeachFunc) list_cols, (void **) &list);
	eel_gconf_set_integer_list (CONF_SONGS_VIEW_COLUMNS, list);
	g_slist_free (list);
}

/**
 * list_uris: list all uris in the treeview
 */
static gboolean
list_uris (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data)
{
	GSList **ret = (GSList **) data;
	GValue val = {0, };
	gtk_tree_model_get_value (model, iter, DIR_COL_URI, &val);
	*ret = g_slist_append (*ret, g_value_get_string (&val));
	return FALSE;
}

/**
 * list_cols: list all cols in the treeview
 */
static gboolean
list_cols (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data)
{
	GSList **ret = (GSList **) data;
	GValue val = {0, };
	gtk_tree_model_get_value (model, iter, COL_COL_CHECKED, &val);
	*ret = g_slist_append (*ret, GINT_TO_POINTER (g_value_get_boolean (&val)));
	return FALSE;
}

/**
 * prefs_dir_entry_activate_cb: dir entry activated
 */
void
prefs_dir_entry_activate_cb (GtkWidget *w, Prefs *p)
{
	prefs_add_clicked_cb (w, p);
}

/**
 * prefs_dir_view_key_press_event_cb: key pressed in the dirview, if delete
 * remove the selected rows
 */
gboolean
prefs_dir_view_key_press_event_cb (GtkWidget *w, GdkEventKey *event, Prefs *p)
{
	if (event->keyval != GDK_Delete) return TRUE;

	prefs_remove_clicked_cb (w, p);

	return FALSE;
}

/**
 * dir_view_selection_changed: selection changed in the dir view
 */
static void
dir_view_selection_changed_cb (GtkTreeSelection *selection, Prefs *p)
{
	GList *list = NULL, *l;
	gtk_tree_selection_selected_foreach
		(gtk_tree_view_get_selection (GTK_TREE_VIEW (p->priv->dir_view)),
		 (GtkTreeSelectionForeachFunc) list_iters, (void **) &list); 
	for (l = list; l != NULL; l = g_list_next (l))
	{
		GtkTreeRowReference *r = l->data;
		gtk_tree_row_reference_free (r);
	}
	gtk_widget_set_sensitive (p->priv->dir_remove, (g_list_length (list) > 0));
	g_list_free (list);
}

/**
 * columns_row_toggled_cb: columns configurator row toggled
 */
static void
columns_row_toggled_cb (GtkCellRendererToggle *toggle, const gchar *path_str, Prefs *p)
{
	GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
	GtkTreeIter iter;

	/* get iter */
	gtk_tree_model_get_iter (GTK_TREE_MODEL (p->priv->columns_store), &iter, path);
	
	/* toggle the toggle */
	gtk_list_store_set (p->priv->columns_store, &iter,
			    COL_COL_CHECKED, !toggle->active,
			    -1);

	/* save */
	gtk_tree_path_free (path);
	
	/* apply changes */
	prefs_save (p);
}

/**
 * prefs_set_page: switch to page @page
 */
void
prefs_set_page (Prefs *p, PrefsPage page)
{
	gtk_notebook_set_current_page (GTK_NOTEBOOK (p->priv->notebook), page);
}
