/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgnome/gnome-util.h>
#include <libgnome/gnome-i18n.h>

#include "misc.h"
#include "sourcelist.h"
#include "ui.h"
#include "state.h"
#include "rhythmbox-marshal.h"
#include "treeview-utils.h"

struct _SourcelistPrivate
{
	GtkWidget *treeview;
	GtkListStore *liststore;
	GHashTable *id_to_ref;
	GHashTable *id_to_child;
};

/* all the different signals */
enum
{
	SOURCE_SELECTED,
	SOURCE_DELETED,
	SOURCE_EDITED,
	LAST_SIGNAL
};

/* columns of the lists */
typedef enum
{
	COL_NAME     = 0,
	COL_PIXBUF   = 1,
	COL_ID       = 2,
	COL_EDITABLE = 3,
	COL_WIDGET   = 4,
	COL_PRIORITY = 5,
	COL_STYLE    = 6,
	COL_COUNT    = 7
} Column;

/* globals */
static GObjectClass *parent_class = NULL;

static guint sourcelist_signals[LAST_SIGNAL] = { 0 };

static gint id_counter = 0;

/* object funtion prototypes */
static void sourcelist_class_init (SourcelistClass *klass);
static void sourcelist_init (Sourcelist *w);
static void sourcelist_finalize (GObject *object);

/* local function prototypes */
static gint treeview_get_selection_single (GtkTreeView *treeview, gint column);

/* ui callbacks */
static void selection_changed_cb (GtkTreeSelection *selection, Sourcelist *s);
static gboolean key_press_event_cb (GtkWidget *widget, GdkEventKey *event,
			            Sourcelist *s);
static void edited_cb (GtkCellRendererText *cell, gchar *path_string, gchar *new_text,
	               Sourcelist *s);

/**
 * sourcelist_get_type: get the GObject type of the sourcelist UI
 */
GType
sourcelist_get_type (void)
{
	static GType sourcelist_type = 0;

  	if (sourcelist_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (SourcelistClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) sourcelist_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (Sourcelist),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) sourcelist_init
      		};

      		sourcelist_type = g_type_register_static (G_TYPE_OBJECT,
                					  "Sourcelist",
                                           	   	   &our_info, 0);
    	}

	return sourcelist_type;
}

/**
 * sourcelist_class_init: initialize the Sourcelist class
 */
static void
sourcelist_class_init (SourcelistClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = sourcelist_finalize;

	/* init signals */
	sourcelist_signals[SOURCE_SELECTED] =
   		g_signal_new ("source_selected",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SourcelistClass, source_selected),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__INT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_INT);
	sourcelist_signals[SOURCE_DELETED] =
   		g_signal_new ("source_deleted",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SourcelistClass, source_deleted),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__INT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_INT);
	sourcelist_signals[SOURCE_EDITED] =
   		g_signal_new ("source_edited",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (SourcelistClass, source_edited),
			      NULL, NULL,
			      rhythmbox_marshal_VOID__INT_STRING,
			      G_TYPE_NONE,
			      2,
			      G_TYPE_INT,
			      G_TYPE_STRING);
}

/**
 * sourcelist_init: intialize the sourcelist object
 */
static void
sourcelist_init (Sourcelist *s)
{
	s->priv = g_new0 (SourcelistPrivate, 1);


}

/**
 * sourcelist_finalize: finalize the sourcelist object
 */
static void
sourcelist_finalize (GObject *object)
{
	Sourcelist *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_SOURCELIST (object));
	
   	s = SOURCELIST (object);

	g_return_if_fail (s->priv != NULL);

	g_hash_table_destroy (s->priv->id_to_ref);
	g_hash_table_destroy (s->priv->id_to_child);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * sourcelist_new: create a new sourcelist object
 */
Sourcelist*
sourcelist_new (GladeXML *gxml)
{
 	Sourcelist *s;

	s = SOURCELIST (g_object_new (TYPE_SOURCELIST, NULL));

	sourcelist_init_interface (s, gxml);

	g_return_val_if_fail (s->priv != NULL, NULL);

	return s;
}

/**
 * sourcelist_get_selected_source: return the selected source node
 */
gint
sourcelist_get_selected_source (Sourcelist *s)
{
	return treeview_get_selection_single (GTK_TREE_VIEW (s->priv->treeview),
				              COL_ID);
}

/**
 * sourcelist_set_ui: set associated ui
 */
void
sourcelist_set_ui (Sourcelist *s, UI *ui)
{
	s->ui = ui;
}

/**
 * sourcelist_get_ui: get associated ui
 */
UI *
sourcelist_get_ui (Sourcelist *s)
{
	return s->ui;
}

/**
 * sourcelist_append_source: add a source to the treeview
 */
gint
sourcelist_append_source (Sourcelist *s, GtkWidget *source)
{
	GtkTreeIter iter;
	GtkTreeRowReference *ref;
	gint id = id_counter;

	g_assert (s->ui != NULL);
	
	gtk_list_store_append (s->priv->liststore, &iter);
	gtk_list_store_set (s->priv->liststore, &iter,
			    COL_ID, id,
			    COL_WIDGET, source,
			    COL_PRIORITY, 0,
			    -1);
	ref = gtk_tree_row_reference_new
		(GTK_TREE_MODEL (s->priv->liststore),
		 gtk_tree_model_get_path
		 	(GTK_TREE_MODEL (s->priv->liststore), &iter));

	gtk_notebook_append_page (GTK_NOTEBOOK (s->ui->notebook),
				  source, NULL);
	g_hash_table_insert (s->priv->id_to_child, GINT_TO_POINTER (id), source);

	g_hash_table_insert (s->priv->id_to_ref, GINT_TO_POINTER (id), ref);

	id_counter++;
	
	return id;
}

/**
 * sourcelist_select_source: select a source
 */
void
sourcelist_select_source (Sourcelist *s, gint id)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));
	GtkTreeIter iter;

	gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore), &iter,
				 gtk_tree_row_reference_get_path (ref));
	gtk_tree_selection_select_iter (gtk_tree_view_get_selection
					(GTK_TREE_VIEW (s->priv->treeview)), &iter);
}

/**
 * sourcelist_remove_source: remove a sourcelist from the treeview
 */
void
sourcelist_remove_source (Sourcelist *s, gint id)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));
	GtkTreeIter iter;
	GtkWidget *child;
	gint page;
	GtkTreePath *path;

	g_assert (s->ui != NULL);

	g_hash_table_remove (s->priv->id_to_ref, GINT_TO_POINTER (id));

	child = g_hash_table_lookup (s->priv->id_to_child, GINT_TO_POINTER (id));
	page = gtk_notebook_page_num (GTK_NOTEBOOK (s->ui->notebook), child);

	gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore), &iter,
				 gtk_tree_row_reference_get_path (ref));

	/* store path */
	path = gtk_tree_model_get_path (GTK_TREE_MODEL (s->priv->liststore), &iter);
	
	/* remove row */
	gtk_list_store_remove (s->priv->liststore, &iter);
	gtk_tree_row_reference_free (ref);
	g_hash_table_remove (s->priv->id_to_child, GINT_TO_POINTER (id));

	/* select next or prev row */
	if (!gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore),
				      &iter, path))
	{
		gtk_tree_path_prev (path);
	}
	gtk_tree_selection_select_path (gtk_tree_view_get_selection
					(GTK_TREE_VIEW (s->priv->treeview)), path);
}

/**
 * sourcelist_source_set_name: set the name of a source
 */
void
sourcelist_source_set_name (Sourcelist *s, gint id, const gchar *name)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));
	GtkTreeIter iter;

	gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore), &iter,
				 gtk_tree_row_reference_get_path (ref));
	gtk_list_store_set (s->priv->liststore, &iter,
			    COL_NAME, name, -1);
}

/**
 * sourcelist_source_set_priority: set the sorting priority of a source
 */
void
sourcelist_source_set_priority (Sourcelist *s, gint id, gint priority)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));
	GtkTreeIter iter;

	gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore), &iter,
				 gtk_tree_row_reference_get_path (ref));
	gtk_list_store_set (s->priv->liststore, &iter,
			    COL_PRIORITY, priority, -1);
}

/**
 * sourcelist_source_set_icon: set the name of a source
 */
void
sourcelist_source_set_icon (Sourcelist *s, gint id, GdkPixbuf *icon)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));
	GtkTreeIter iter;

	gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore), &iter,
				 gtk_tree_row_reference_get_path (ref));
	gtk_list_store_set (s->priv->liststore, &iter,
			    COL_PIXBUF, icon, -1);
}

/**
 * sourcelist_source_set_editable: set a source's editability
 */
void
sourcelist_source_set_editable (Sourcelist *s, gint id, gboolean editable)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));
	GtkTreeIter iter;

	gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore), &iter,
				 gtk_tree_row_reference_get_path (ref));
	gtk_list_store_set (s->priv->liststore, &iter,
			    COL_EDITABLE, editable, -1);
}

/**
 * sourcelist_source_start_editing: start editing the cell of this source
 */
void
sourcelist_source_start_editing (Sourcelist *s, int id)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));

	gtk_tree_view_set_cursor (GTK_TREE_VIEW (s->priv->treeview),
				  gtk_tree_row_reference_get_path (ref), 
				  gtk_tree_view_get_column (GTK_TREE_VIEW (s->priv->treeview), 0), TRUE);
				  
}

/**
 * sourcelist_source_set_playing: make text bold to indicate that the source is currently playing
 */
void
sourcelist_source_set_playing (Sourcelist *s, int id, gboolean playing)
{
	GtkTreeRowReference *ref = g_hash_table_lookup (s->priv->id_to_ref,
							GINT_TO_POINTER (id));
	GtkTreeIter iter;

	gtk_tree_model_get_iter (GTK_TREE_MODEL (s->priv->liststore), &iter,
				 gtk_tree_row_reference_get_path (ref));
	gtk_list_store_set (s->priv->liststore, &iter,
			    COL_STYLE, playing, -1);
}

/**
 * selection_changed_cb: selection changed in the treeview
 */
static void
selection_changed_cb (GtkTreeSelection *selection, Sourcelist *s)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GValue val = {0, };
	gint id;
	GtkWidget *child;
	gint page;

	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) return;
	gtk_tree_model_get_value (model, &iter, COL_ID, &val);
	id = g_value_get_int (&val);

	child = g_hash_table_lookup (s->priv->id_to_child, GINT_TO_POINTER (id));
	page = gtk_notebook_page_num (GTK_NOTEBOOK (s->ui->notebook), child);
	gtk_notebook_set_current_page (GTK_NOTEBOOK (s->ui->notebook), page);
	
	g_signal_emit (G_OBJECT (s), sourcelist_signals[SOURCE_SELECTED], 0, id);
}

/**
 * key_press_event_cb: key pressed in the treeview, if key == delete remove
 * this source
 */
static gboolean
key_press_event_cb (GtkWidget *widget, GdkEventKey *event, Sourcelist *s)
{
	gint id;

	if (event->keyval != GDK_Delete) return FALSE;

	id = treeview_get_selection_single (GTK_TREE_VIEW (s->priv->treeview), COL_ID);

	g_signal_emit (G_OBJECT (s), sourcelist_signals[SOURCE_DELETED], 0,
       		       id);

	return TRUE;
}

/**
 * edited_cb: row edited
 */
static void
edited_cb (GtkCellRendererText *cell, gchar *path_string, gchar *new_text,
	   Sourcelist *s)
{
	GtkTreeModel *model = GTK_TREE_MODEL (s->priv->liststore);
	GtkTreeIter iter;
	GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
	GValue val = {0, };
	gint id;

	gtk_tree_model_get_iter (model, &iter, path);
	gtk_list_store_set (GTK_LIST_STORE (model), &iter, COL_NAME, new_text, -1);
	gtk_tree_model_get_value (model, &iter, COL_ID, &val);
	id = g_value_get_int (&val);

	gtk_tree_path_free (path);

	g_signal_emit (G_OBJECT (s), sourcelist_signals[SOURCE_EDITED], 0,
       		       id, g_strdup (new_text));
}

/**
 * treeview_get_selection_single: get the id associated to the
 * selected row of a GtkTreeView. selection type should be single or browse.
 */
static gint
treeview_get_selection_single (GtkTreeView *treeview, gint column)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GValue val = {0, };
	GtkTreeSelection *selection = gtk_tree_view_get_selection (treeview);

	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) return -1;
	gtk_tree_model_get_value (model, &iter, column, &val);
	return g_value_get_int (&val);
}

/**
 * sourcelist_init_interface: setup the interface from a glade file
 */
void
sourcelist_init_interface (Sourcelist *s, GladeXML *gxml)
{
	GtkTreeView *current;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	

	s->priv->treeview = glade_xml_get_widget (gxml, "source_list");

	/* setup sources view */
	current = GTK_TREE_VIEW (s->priv->treeview);
	s->priv->liststore = gtk_list_store_new (COL_COUNT,
						 G_TYPE_STRING,
						 GDK_TYPE_PIXBUF,
						 G_TYPE_INT,
						 G_TYPE_BOOLEAN,
						 G_TYPE_POINTER,
						 G_TYPE_INT,
						 G_TYPE_BOOLEAN);

	s->priv->id_to_ref = g_hash_table_new (g_direct_hash, g_direct_equal);
	s->priv->id_to_child = g_hash_table_new (g_direct_hash, g_direct_equal);
	gtk_tree_view_set_model (current, GTK_TREE_MODEL (s->priv->liststore));
	gtk_tree_view_set_rules_hint (current, FALSE);
	gtk_tree_view_set_headers_visible (current, TRUE);

	column = gtk_tree_view_column_new ();
	gtk_tree_view_column_set_title (column, _("Source"));
	renderer = gtk_cell_renderer_pixbuf_new ();
	gtk_tree_view_column_pack_start (column, renderer, FALSE);
	gtk_tree_view_column_add_attribute (column, renderer,
					    "pixbuf", COL_PIXBUF);
	renderer = gtk_cell_renderer_text_new ();
	g_object_set (G_OBJECT (renderer),
		      "weight", PANGO_WEIGHT_BOLD,
		      NULL);
	g_signal_connect (G_OBJECT (renderer), "edited",
		          G_CALLBACK (edited_cb), s);
	gtk_tree_view_column_pack_start (column, renderer, FALSE);
	gtk_tree_view_column_add_attribute (column, renderer,
					    "text", COL_NAME);
	gtk_tree_view_column_add_attribute (column, renderer,
					    "editable", COL_EDITABLE);
	gtk_tree_view_column_add_attribute (column, renderer,
					    "weight_set", COL_STYLE);
	g_object_set_data (G_OBJECT (column), "original_column", 0);
	gtk_tree_view_append_column (current, column);
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (s->priv->liststore),
					 COL_NAME,
					 priority_sort_func,
					 GINT_TO_POINTER (COL_PRIORITY),
					 NULL);
	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (s->priv->liststore),
					      COL_NAME,
					      GTK_SORT_ASCENDING);
	gtk_tree_view_column_set_clickable (column, FALSE);
	selection = gtk_tree_view_get_selection (current);
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
	g_signal_connect (G_OBJECT (selection), "changed",
			  G_CALLBACK (selection_changed_cb), s);
	g_signal_connect (G_OBJECT (current), "key_press_event",
			  G_CALLBACK (key_press_event_cb), s);
}
