/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgnome/gnome-util.h>
#include <libgnome/gnome-i18n.h>
#include <bonobo/bonobo-dock-item.h>

#include "misc.h"
#include "statusbar.h"
#include "ui.h"
#include "icons.h"

struct _StatusbarPrivate
{
	GtkWidget *hbox;
	GtkWidget *status;
	GtkWidget *shuffle;
	GtkWidget *repeat;
};

/* all the different signals */
enum
{
	SHUFFLE_CLICKED,
	REPEAT_CLICKED,
	NEW_PLAYLIST_CLICKED,
	LAST_SIGNAL
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint statusbar_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void statusbar_class_init (StatusbarClass *klass);
static void statusbar_init (Statusbar *s);
static void statusbar_finalize (GObject *object);

/* ui callbacks */
static void shuffle_clicked_cb (GtkWidget *button, Statusbar *s);
static void repeat_clicked_cb (GtkWidget *button, Statusbar *s);
#if 0
static void new_playlist_clicked_cb (GtkWidget *button, Statusbar *s);
#endif

/**
 * statusbar_get_type: get the GObject type of the statusbar
 */
GType
statusbar_get_type (void)
{
	static GType statusbar_type = 0;

  	if (statusbar_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (StatusbarClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) statusbar_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (Statusbar),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) statusbar_init
      		};

      		statusbar_type = g_type_register_static (G_TYPE_OBJECT,
                					        "Statusbar",
                                           	   	        &our_info, 0);
    	}

	return statusbar_type;
}

/**
 * statusbar_class_init: initialize the Statusbar class
 */
static void
statusbar_class_init (StatusbarClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = statusbar_finalize;

	/* init signals */
	statusbar_signals[SHUFFLE_CLICKED] =
   		g_signal_new ("shuffle_clicked",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (StatusbarClass, shuffle_clicked),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	statusbar_signals[REPEAT_CLICKED] =
   		g_signal_new ("repeat_clicked",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (StatusbarClass, repeat_clicked),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	statusbar_signals[NEW_PLAYLIST_CLICKED] =
   		g_signal_new ("new_playlist_clicked",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (StatusbarClass, new_playlist_clicked),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
}

/**
 * statusbar_init: intialize the statusbar object
 */
static void
statusbar_init (Statusbar *s)
{
	s->priv = g_new0 (StatusbarPrivate, 1);
}

/**
 * statusbar_finalize: finalize the statusbar object
 */
static void
statusbar_finalize (GObject *object)
{
	Statusbar *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_STATUSBAR (object));
	
   	s = STATUSBAR (object);

	g_return_if_fail (s->priv != NULL);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * statusbar_new: create a new statusbar object
 */
Statusbar *
statusbar_new (GladeXML *gxml)
{
 	Statusbar *s;

	s = STATUSBAR (g_object_new (TYPE_STATUSBAR, NULL));
	g_return_val_if_fail (s->priv != NULL, NULL);

	statusbar_init_interface (s, gxml);

	return s;
}

/**
 * statusbar_set_ui: set associated ui
 */
void
statusbar_set_ui (Statusbar *s, UI *ui)
{
	s->ui = ui;
}

/**
 * statusbar_get_ui: get associated ui
 */
UI *
statusbar_get_ui (Statusbar *s)
{
	return s->ui;
}

/**
 * statusbar_set_message: set statusbar message
 */
void
statusbar_set_message (Statusbar *s, const gchar *message)
{
	gtk_label_set_text (GTK_LABEL (s->priv->status), message);
}

/**
 * statusbar_get_message: return statusbar message
 */
const gchar *
statusbar_get_message (Statusbar *s)
{
	return gtk_label_get_text (GTK_LABEL (s->priv->status));
}

/**
 * statusbar_set_shuffle: set the shuffle button state
 */
void
statusbar_set_shuffle (Statusbar *s, gboolean shuffle)
{
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (s->priv->shuffle), shuffle);
}

/**
 * statusbar_get_shuffle: get shuffle button state
 */
gboolean
statusbar_get_shuffle (Statusbar *s)
{
	return GTK_TOGGLE_BUTTON (s->priv->shuffle)->active;
}

/**
 * statusbar_set_repeat: set the repeat button state
 */
void
statusbar_set_repeat (Statusbar *s, gboolean repeat)
{
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (s->priv->repeat), repeat);
}

/**
 * statusbar_get_repeat: get repeat button state
 */
gboolean
statusbar_get_repeat (Statusbar *s)
{
	return GTK_TOGGLE_BUTTON (s->priv->repeat)->active;
}

/**
 * shuffle_clicked_cb: shuffle clicked
 */
static void
shuffle_clicked_cb (GtkWidget *button, Statusbar *s)
{
	g_signal_emit (G_OBJECT (s), statusbar_signals[SHUFFLE_CLICKED], 0,
		       GTK_TOGGLE_BUTTON (button)->active);
}

/**
 * repeat_clicked_cb: repeat clicked
 */
static void
repeat_clicked_cb (GtkWidget *button, Statusbar *s)
{
	g_signal_emit (G_OBJECT (s), statusbar_signals[REPEAT_CLICKED], 0,
		       GTK_TOGGLE_BUTTON (button)->active);
}

#if 0
/**
 * new_playlist_clicked_cb: "new playlist" clicked
 */
static void
new_playlist_clicked_cb (GtkWidget *button, Statusbar *s)
{
	g_signal_emit (G_OBJECT (s), statusbar_signals[NEW_PLAYLIST_CLICKED], 0);
}
#endif

/**
 * statusbar_set_interface: sets the glade file used to load the interface from
 */
void 
statusbar_init_interface (Statusbar *s, GladeXML *gxml)
{
	GtkImage *image;

	s->priv->status  = glade_xml_get_widget (gxml, "song_status_label");
	s->priv->shuffle = glade_xml_get_widget (gxml, "shuffle_button");
	s->priv->repeat  = glade_xml_get_widget (gxml, "repeat_button");

	image = GTK_IMAGE (glade_xml_get_widget (gxml, "playlist_image"));
	gtk_image_set_from_stock (image, RHYTHMBOX_STOCK_PLAYLIST, GTK_ICON_SIZE_MENU);

	image = GTK_IMAGE (glade_xml_get_widget (gxml, "shuffle_image"));
	gtk_image_set_from_stock (image, RHYTHMBOX_STOCK_SHUFFLE, GTK_ICON_SIZE_MENU);

	image = GTK_IMAGE (glade_xml_get_widget (gxml, "repeat_image"));
	gtk_image_set_from_stock (image, RHYTHMBOX_STOCK_REPEAT, GTK_ICON_SIZE_MENU);

	g_signal_connect (G_OBJECT (s->priv->shuffle), "clicked",
			  G_CALLBACK (shuffle_clicked_cb), s);
	g_signal_connect (G_OBJECT (s->priv->repeat), "clicked",
			  G_CALLBACK (repeat_clicked_cb), s);
}
