/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: ui.c,v 1.41 2002/04/13 14:51:54 jbaayen Exp $
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgnomeui/gnome-about.h>
#include <libgnomeui/gnome-dialog.h>
#include <libgnome/gnome-util.h>
#include <libgnome/gnome-i18n.h>
#include <bonobo/bonobo-dock.h>
#include <bonobo/bonobo-dock-item.h>

#include "prefs.h"
#include "glade.h"
#include "ui.h"
#include "main.h"
#include "misc.h"
#include "state.h"
#include "eel-gconf-extensions.h"
#include "rhythmbox-marshal.h"

#define CONF_MENUBARS_DETACHABLE "/desktop/gnome/interface/menubar_detachable"

struct _UIPrivate
{
	/* general widgets */
	GtkWidget *window;
	GtkWidget *hpaned;
	GtkWidget *dock;
	GtkWidget *menudock;
	GtkWidget *window_contents;

	/* menuitems */
	GtkWidget *cut;
	GtkWidget *copy;
	GtkWidget *paste;
	GtkWidget *shuffle;
	GtkWidget *repeat;
	GtkWidget *play;
	GtkWidget *rewind;
	GtkWidget *forward;
	GtkWidget *vol_up;
	GtkWidget *vol_down;
	GtkWidget *mute;

	gboolean muted;
	float volume_backup;

	char *current_song;
	char *current_artist;
	char *current_uri;

	long elapsed;
	float progress;

	Prefs *prefs;
};

/* all the different signals */
enum {
	PLAY,
	PAUSE,
	STOP,
	REWIND,
	FORWARD,
	VOLUME_CHANGED,
	PROGRESS_CHANGED,
	NEW_PLAYLIST,
	SET_SHUFFLE,
	SET_REPEAT,
	NEW_SONG,
	QUERY,
	SOURCE_SELECTED,
	SOURCE_DELETED,
	SOURCE_EDITED,
	CUT,
	COPY,
	PASTE,
	CLEAR,
	SELECT_ALL,
	SELECT_NONE,
	LAST_SIGNAL
};

/* globals */
static GObjectClass *parent_class = NULL;

static guint ui_signals[LAST_SIGNAL] = { 0 };

/* object funtion prototypes */
static void ui_class_init (UIClass *klass);
static void ui_init (UI *w);
static void ui_finalize (GObject *object);

/* glade callbacks */
gboolean window_delete_event_cb (GtkWidget *widget, GdkEventAny *event, UI *w);
void window_new_playlist_cb (GtkWidget *widget, UI *w);
void window_new_playlist_from_sel_cb (GtkWidget *widget, UI *w);
void window_add_to_lib_cb (GtkWidget *widget, UI *w);
void window_lib_dirs_cb (GtkWidget *widget, UI *w);
void window_song_info_cb (GtkWidget *widget, UI *w);
void window_exit_cb (GtkWidget *widget, UI *w);
void window_cut_cb (GtkWidget *widget, UI *ui);
void window_copy_cb (GtkWidget *widget, UI *ui);
void window_paste_cb (GtkWidget *widget, UI *ui);
void window_clear_cb (GtkWidget *widget, UI *ui);
void window_select_all_cb (GtkWidget *widget, UI *w);
void window_select_none_cb (GtkWidget *widget, UI *w);
void window_vol_up_cb (GtkWidget *widget, UI *w);
void window_vol_down_cb (GtkWidget *widget, UI *w);
void window_mute_cb (GtkWidget *widget, UI *w);
void window_small_cb (GtkWidget *widget, UI *w);
void window_large_cb (GtkWidget *widget, UI *w);
void window_prefs_cb (GtkWidget *widget, UI *w);
void window_about_cb (GtkWidget *widget, UI *w);
void window_play_cb (GtkWidget *widget, UI *ui);
void window_forward_cb (GtkWidget *widget, UI *ui);
void window_rewind_cb (GtkWidget *widget, UI *ui);

/* local ui callbacks */
static void add_to_lib_ok_cb (GtkWidget *widget, UI *w);
static void add_to_lib_cancel_cb (GtkWidget *widget, UI *w);
static void add_playlist (UI *ui, gboolean sel);

/* redirection callbacks */
static void control_bar_play_clicked_cb (ControlBar *t, ControlBarPlayAction action, UI *ui);
static void control_bar_rewind_clicked_cb (ControlBar *t, UI *ui);
static void control_bar_forward_clicked_cb (ControlBar *t, UI *ui);
static void control_bar_search_activate_cb (ControlBar *t, gchar *text, UI *ui);
static void control_bar_volume_changed_cb (ControlBar *t, gfloat vol, UI *ui);
static void control_bar_progress_changed_cb (ControlBar *t, gfloat progress, UI *ui);
static void statusbar_shuffle_clicked_cb (ControlBar *t, gboolean shuffle, UI *ui);
static void statusbar_repeat_clicked_cb (ControlBar *t, gboolean repeat, UI *ui);
static void statusbar_new_playlist_clicked_cb (ControlBar *t, UI *ui);
static void sourcelist_source_selected_cb (ControlBar *t, int id, UI *ui);
static void sourcelist_source_deleted_cb (ControlBar *t, int id, UI *ui);
static void sourcelist_source_edited_cb (ControlBar *t, int id, gchar *text, UI *ui);

/**
 * ui_get_type: get the GObject type of the player
 */
GType
ui_get_type (void)
{
	static GType ui_type = 0;

  	if (ui_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (UIClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) ui_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (UI),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) ui_init
      		};

      		ui_type = g_type_register_static (G_TYPE_OBJECT,
                				  "UserInterface",
                                           	  &our_info, 0);
    	}

	return ui_type;
}

/**
 * ui_class_init: initialize the UI class
 */
static void
ui_class_init (UIClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = ui_finalize;

	/* init signals */
	ui_signals[PLAY] =
   		g_signal_new ("play",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, play),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[PAUSE] =
   		g_signal_new ("pause",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, pause),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[STOP] =
   		g_signal_new ("stop",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, stop),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[REWIND] =
   		g_signal_new ("rewind",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, rewind),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[FORWARD] =
   		g_signal_new ("forward",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, forward),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[VOLUME_CHANGED] =
   		g_signal_new ("volume_changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, volume_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__FLOAT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_FLOAT);
	ui_signals[PROGRESS_CHANGED] =
   		g_signal_new ("progress_changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, progress_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__FLOAT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_FLOAT);
	ui_signals[NEW_PLAYLIST] =
   		g_signal_new ("new_playlist",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, new_playlist),
			      NULL, NULL,
			      rhythmbox_marshal_VOID__STRING_BOOLEAN,
			      G_TYPE_NONE,
			      2,
			      G_TYPE_STRING,
			      G_TYPE_BOOLEAN);
	ui_signals[SET_SHUFFLE] =
   		g_signal_new ("set_shuffle",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, set_shuffle),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	ui_signals[SET_REPEAT] =
   		g_signal_new ("set_repeat",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, set_repeat),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	ui_signals[NEW_SONG] =
   		g_signal_new ("new_song",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, new_song),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_STRING);
	ui_signals[QUERY] =
   		g_signal_new ("query",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, query),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_BOOLEAN,
			      1,
			      G_TYPE_STRING);
	ui_signals[SOURCE_SELECTED] =
   		g_signal_new ("source_selected",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, source_selected),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__INT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_INT);
	ui_signals[SOURCE_DELETED] =
   		g_signal_new ("source_deleted",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, source_deleted),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__INT,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_INT);
	ui_signals[SOURCE_EDITED] =
   		g_signal_new ("source_edited",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, source_edited),
			      NULL, NULL,
			      rhythmbox_marshal_VOID__INT_STRING,
			      G_TYPE_NONE,
			      2,
			      G_TYPE_INT,
			      G_TYPE_STRING);
	ui_signals[CUT] =
   		g_signal_new ("cut",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, cut),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[COPY] =
   		g_signal_new ("copy",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, copy),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[PASTE] =
   		g_signal_new ("paste",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, paste),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[CLEAR] =
   		g_signal_new ("clear",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, clear),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[SELECT_ALL] =
   		g_signal_new ("select_all",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, select_all),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
	ui_signals[SELECT_NONE] =
   		g_signal_new ("select_none",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (UIClass, select_none),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE,
			      0);
}

static void
ui_init (UI *w)
{
	BonoboDockItemBehavior behv = BONOBO_DOCK_ITEM_BEH_NEVER_VERTICAL |
				      BONOBO_DOCK_ITEM_BEH_EXCLUSIVE;
	GladeXML *gxml;
	GtkWidget *song_box;

	w->priv = g_new0 (UIPrivate, 1);

	gxml = glade_widget_new ("rhythmbox.glade",
				 "rhythmbox",
				 &(w->priv->window), w);

	w->priv->prefs           = NULL;
	w->priv->muted           = FALSE;

	w->priv->current_song    = NULL;
	w->priv->current_artist  = NULL;
	w->priv->current_uri     = NULL;

	w->priv->elapsed         = 0;
	w->priv->progress        = 0.0;

	/* general widgets */
	w->priv->hpaned          = glade_xml_get_widget (gxml, "sp_hpaned");
	w->priv->dock            = glade_xml_get_widget (gxml, "dock");
	w->priv->window_contents = glade_xml_get_widget (gxml, "window-contents");

	/* menuitems */
	w->priv->cut      = glade_xml_get_widget (gxml, "cut");
	w->priv->copy     = glade_xml_get_widget (gxml, "copy");
	w->priv->paste    = glade_xml_get_widget (gxml, "paste");
	w->priv->shuffle  = glade_xml_get_widget (gxml, "shuffle");
	w->priv->repeat   = glade_xml_get_widget (gxml, "repeat");
	w->priv->play     = glade_xml_get_widget (gxml, "play");
	w->priv->rewind   = glade_xml_get_widget (gxml, "rewind");
	w->priv->forward  = glade_xml_get_widget (gxml, "forward");
	w->priv->vol_up   = glade_xml_get_widget (gxml, "vol_up");
	w->priv->vol_down = glade_xml_get_widget (gxml, "vol_down");
	w->priv->mute     = glade_xml_get_widget (gxml, "mute");

	/* setup menubar */
	w->priv->menudock = glade_xml_get_widget (gxml, "menudock");
	if (!eel_gconf_get_boolean (CONF_MENUBARS_DETACHABLE))
		behv = behv | BONOBO_DOCK_ITEM_BEH_LOCKED;
	BONOBO_DOCK_ITEM (w->priv->menudock)->behavior = behv;

	/* add statusbar */
	w->statusbar = statusbar_new (gxml);
	statusbar_set_ui (w->statusbar, w);
	g_signal_connect (G_OBJECT (w->statusbar),
			  "shuffle_clicked",
			  G_CALLBACK (statusbar_shuffle_clicked_cb), w);
	g_signal_connect (G_OBJECT (w->statusbar),
			  "repeat_clicked",
			  G_CALLBACK (statusbar_repeat_clicked_cb), w);
	g_signal_connect (G_OBJECT (w->statusbar),
			  "new_playlist_clicked",
			  G_CALLBACK (statusbar_new_playlist_clicked_cb), w);

	/* add sources list */
	w->sourcelist = sourcelist_new (gxml);
	sourcelist_set_ui (w->sourcelist, w);
	g_signal_connect (G_OBJECT (w->sourcelist),
			  "source_selected",
			  G_CALLBACK (sourcelist_source_selected_cb), w);
	g_signal_connect (G_OBJECT (w->sourcelist),
			  "source_deleted",
			  G_CALLBACK (sourcelist_source_deleted_cb), w);
	g_signal_connect (G_OBJECT (w->sourcelist),
			  "source_edited",
			  G_CALLBACK (sourcelist_source_edited_cb), w);

	/* add notebook */
	song_box = glade_xml_get_widget (gxml, "song_box");
	w->notebook = gtk_notebook_new ();
	gtk_notebook_set_show_border (GTK_NOTEBOOK (w->notebook), FALSE);
	gtk_notebook_set_show_tabs (GTK_NOTEBOOK (w->notebook), FALSE);
	gtk_container_add (GTK_CONTAINER (song_box), w->notebook);
	gtk_widget_show_all (w->notebook);

	/* add control_bar */
	w->control_bar = control_bar_new (gxml);
	control_bar_set_ui (w->control_bar, w);
	g_signal_connect (G_OBJECT (w->control_bar),
			  "play_clicked",
			  G_CALLBACK (control_bar_play_clicked_cb), w);
	g_signal_connect (G_OBJECT (w->control_bar),
			  "rewind_clicked",
			  G_CALLBACK (control_bar_rewind_clicked_cb), w);
	g_signal_connect (G_OBJECT (w->control_bar),
			  "forward_clicked",
			  G_CALLBACK (control_bar_forward_clicked_cb), w);
	g_signal_connect (G_OBJECT (w->control_bar),
			  "search_activate",
			  G_CALLBACK (control_bar_search_activate_cb), w);
	g_signal_connect (G_OBJECT (w->control_bar),
			  "volume_changed",
			  G_CALLBACK (control_bar_volume_changed_cb), w);
	g_signal_connect (G_OBJECT (w->control_bar),
			  "progress_changed",
			  G_CALLBACK (control_bar_progress_changed_cb), w);

	/* load state */
	state_load_window_state (w->priv->window, "player");
	state_load_dock_layout (w->priv->window, "player");
	state_load_pane_pos (w->priv->hpaned, "main_paned");

	/* cleanup */
	g_object_unref (G_OBJECT (gxml));

	/* show contents */
	gtk_widget_show_all (GTK_BIN (w->priv->window)->child);
}

static void
ui_finalize (GObject *object)
{
	UI *w;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_UI (object));
	
   	w = UI (object);

	g_return_if_fail (w->priv != NULL);

	/* save state */
	state_save_window_state (w->priv->window, "player");
	state_save_dock_layout (w->priv->window, "player");
	state_save_pane_pos (w->priv->hpaned, "main_paned");

	g_object_unref (w->control_bar);
	g_object_unref (w->sourcelist);
	g_object_unref (w->statusbar);
	if (w->priv->prefs != NULL)
		g_object_unref (w->priv->prefs);

	gtk_widget_destroy (w->priv->window);

	g_free (w->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

UI*
ui_new (void)
{
 	UI *w;

	w = UI (g_object_new (TYPE_UI, NULL));

	g_return_val_if_fail (w->priv != NULL, NULL);

	return w;
}

gboolean
window_delete_event_cb (GtkWidget *widget, GdkEventAny *event, UI *w)
{
	rhythmbox_exit (rhythmbox);

	return TRUE;
}

void
window_new_playlist_cb (GtkWidget *widget, UI *ui)
{
	add_playlist (ui, FALSE);
}

void
window_new_playlist_from_sel_cb (GtkWidget *widget, UI *ui)
{
	add_playlist (ui, TRUE);
}

static void
add_to_lib_ok_cb (GtkWidget *widget, UI *w)
{
	GtkWidget *filesel;
	const gchar *filename;

	filesel = gtk_object_get_data (GTK_OBJECT (widget), "filesel");
	filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (filesel));

	if (g_file_test (filename, G_FILE_TEST_EXISTS))
	{
		g_signal_emit (G_OBJECT (w), ui_signals[NEW_SONG], 0,
	       		       g_strdup (filename));
	}

	gtk_widget_destroy (filesel);
}

static void
add_to_lib_cancel_cb (GtkWidget *widget, UI *w)
{
	GtkWidget *filesel;

	filesel = gtk_object_get_data (GTK_OBJECT (widget), "filesel");

	gtk_widget_destroy (filesel);
}

void
window_add_to_lib_cb (GtkWidget *widget, UI *w)
{
	GtkWidget *filesel;

	filesel = gtk_file_selection_new ("Add to library");

	g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
                          "clicked", G_CALLBACK (add_to_lib_ok_cb), w);
	g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button),
                          "clicked", G_CALLBACK (add_to_lib_cancel_cb), w);

	gtk_object_set_data (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button), 
			     "filesel", filesel);
	gtk_object_set_data (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button), 
			     "filesel", filesel);

	gtk_widget_show (filesel);
}

void
window_lib_dirs_cb (GtkWidget *widget, UI *w)
{
	if (w->priv->prefs == NULL)
		w->priv->prefs = prefs_new ();
	prefs_set_page (w->priv->prefs, PREFS_PAGE_LIBRARY_FOLDERS);
	prefs_show (w->priv->prefs);
}

void
window_song_info_cb (GtkWidget *widget, UI *w)
{
	/* get selected song, and cast that back via a signal */
}

void
window_exit_cb (GtkWidget *widget, UI *w)
{
	rhythmbox_exit (rhythmbox);
}

void
window_cut_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[CUT], 0);
}

void
window_copy_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[COPY], 0);
}

void
window_paste_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[PASTE], 0);
}

void
window_clear_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[CLEAR], 0);
}

void
window_select_all_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[SELECT_ALL], 0);
}

void
window_select_none_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[SELECT_NONE], 0);
}

void
window_shuffle_cb (GtkWidget *widget, UI *ui)
{
	gboolean shuffle = GTK_CHECK_MENU_ITEM (widget)->active;
	ui_set_shuffle (ui, shuffle);
}

void
window_repeat_cb (GtkWidget *widget, UI *ui)
{
	gboolean repeat = GTK_CHECK_MENU_ITEM (widget)->active;
	ui_set_repeat (ui, repeat);
}

void
window_play_cb (GtkWidget *widget, UI *ui)
{
	switch (ui->state)
	{
	case UI_STATE_PLAYING:
		g_signal_emit (G_OBJECT (ui), ui_signals[PAUSE], 0);
		break;
	case UI_STATE_PLAYING_WITH_STOP:
		g_signal_emit (G_OBJECT (ui), ui_signals[STOP], 0);
		break;
	case UI_STATE_PAUSED:
	case UI_STATE_STOPPED:
	case UI_STATE_PROCESSING:
		g_signal_emit (G_OBJECT (ui), ui_signals[PLAY], 0);
		break;
	default:
		break;
	}
}

void
window_forward_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[FORWARD], 0);
}

void
window_rewind_cb (GtkWidget *widget, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[REWIND], 0);
}

void
window_vol_up_cb (GtkWidget *widget, UI *ui)
{
	float newvol = control_bar_get_volume (CONTROL_BAR (ui->control_bar)) + 0.1;
	if (newvol > 1.0) newvol = 1.0;
	control_bar_set_volume (CONTROL_BAR (ui->control_bar), newvol);
}

void
window_vol_down_cb (GtkWidget *widget, UI *ui)
{
	float newvol = control_bar_get_volume (CONTROL_BAR (ui->control_bar)) - 0.1;
	if (newvol < 0.0) newvol = 0.0;
	control_bar_set_volume (CONTROL_BAR (ui->control_bar), newvol);
}

void
window_mute_cb (GtkWidget *widget, UI *ui)
{
	float newvol = 0.0;

	if (GTK_CHECK_MENU_ITEM (widget)->active == FALSE)
	{
		newvol = ui->priv->volume_backup;
	}
	else
	{
		ui->priv->volume_backup = control_bar_get_volume (CONTROL_BAR (ui->control_bar));
	}
	
	control_bar_set_volume (CONTROL_BAR (ui->control_bar), newvol);

	ui->priv->muted = GTK_CHECK_MENU_ITEM (widget)->active;
}

void
window_small_cb (GtkWidget *widget, UI *ui)
{

}

void
window_large_cb (GtkWidget *widget, UI *ui)
{

}

void
window_prefs_cb (GtkWidget *widget, UI *w)
{
	if (w->priv->prefs == NULL)
		w->priv->prefs = prefs_new ();
	prefs_show (w->priv->prefs);
}

void
window_about_cb (GtkWidget *widget, UI *ui)
{
	static GtkWidget *about;
	GdkPixbuf *pixbuf;

	const char *authors[] = {
		"Jorn Baayen <jorn@nl.linux.org>",
		"Marco Pesenti Gritti <marco@it.gnome.org>",
		"Bastien Nocera <hadess@hadess.net>",
		"Seth Nickell <snickell@stanford.edu>",
		"Olivier Martin <omartin@ifrance.com>",
		NULL};

	if (about != NULL) 
	{
                gdk_window_raise (about->window);
                gdk_window_show (about->window);
                return;
        }

	pixbuf = gdk_pixbuf_new_from_file (PIXMAP_DIR "/rhythmbox.png", NULL);
	about = gnome_about_new ("RhythmBox Music Player", VERSION,
				 "(C) 2002 The RhythmBox Team",
				 "A music management package",
				 (const char **) authors,
				 NULL, NULL, pixbuf);
	if (pixbuf != NULL)
		g_object_unref (G_OBJECT (pixbuf));

	gtk_window_set_wmclass (GTK_WINDOW (about), "dialog_about",
				"rhythmbox");

	g_signal_connect (G_OBJECT (about), "destroy",
			  G_CALLBACK (gtk_widget_destroyed), &about);

	gtk_window_set_transient_for (GTK_WINDOW (about), GTK_WINDOW (ui->priv->window));
	
        gtk_widget_show (about);
}

void
ui_set_state (UI *ui, UIState state)
{
	GtkWidget *play_label = GTK_BIN (ui->priv->play)->child;
		
	ui->state = state;

	/* update controlbar */
	control_bar_set_state (CONTROL_BAR (ui->control_bar), state);

	/* update play menuitem */
	switch (state)
	{
	case UI_STATE_PLAYING:
		gtk_label_set_text_with_mnemonic (GTK_LABEL (play_label), _("_Pause"));
		break;
	case UI_STATE_PLAYING_WITH_STOP:
		gtk_label_set_text_with_mnemonic (GTK_LABEL (play_label), _("_Stop"));
		break;
	default:
		gtk_label_set_text_with_mnemonic (GTK_LABEL (play_label), _("_Play"));
		break;
	}
}

void
ui_set_can_rewind (UI *ui, gboolean can)
{
	control_bar_set_rewind_sensitive (CONTROL_BAR (ui->control_bar), can);
	gtk_widget_set_sensitive (ui->priv->rewind, can);
}

void
ui_set_can_forward (UI *ui, gboolean can)
{
	control_bar_set_forward_sensitive (CONTROL_BAR (ui->control_bar), can);
	gtk_widget_set_sensitive (ui->priv->forward, can);
}

void
ui_merge_widgets (UI *ui, GList *widgets)
{
	control_bar_merge_widgets (CONTROL_BAR (ui->control_bar), widgets);
}

GtkWidget *
ui_new_action_button (UI *ui,
		      const char *label,
		      const char *tooltip,
		      GdkPixbuf *icon,
		      gboolean toggle)
{
	return control_bar_new_action_button (CONTROL_BAR (ui->control_bar),
				              label, tooltip, icon, toggle);
}

void
ui_set_volume (UI *ui, gfloat volume)
{
	control_bar_set_volume (CONTROL_BAR (ui->control_bar), volume);
}

float
ui_get_volume (UI *ui)
{
	return control_bar_get_volume (CONTROL_BAR (ui->control_bar));
}

void
ui_set_song (UI *ui, const char *title, const char *artist, const char *uri)
{
	char *song;

	g_free (ui->priv->current_song);
	g_free (ui->priv->current_artist);
	g_free (ui->priv->current_uri);

	ui->priv->current_song   = g_strdup (title);
	ui->priv->current_artist = g_strdup (artist);
	ui->priv->current_uri    = g_strdup (uri);
 
	if (title != NULL && artist != NULL)
	{
		song = g_strdup_printf ("%s - %s", artist, title);
	}
	else if (title != NULL)
	{
		song = g_strdup (title);
	}
	else
	{
		song = g_strdup (uri);
	}

	gtk_window_set_title (GTK_WINDOW (ui->priv->window), song != NULL ? song : _("RhythmBox"));
	control_bar_set_lcd_song (CONTROL_BAR (ui->control_bar), title, artist, uri);
	g_free (song);
}

void
ui_set_elapsed_time (UI *ui, long elapsed, float progress)
{
	ui->priv->elapsed = elapsed;
	ui->priv->progress = progress;
	control_bar_set_lcd_time (CONTROL_BAR (ui->control_bar), elapsed,
				  progress);
}

void
ui_set_status (UI *ui, const gchar *status)
{
	statusbar_set_message (STATUSBAR (ui->statusbar), status);
}

int
ui_append_source (UI *ui, GtkWidget *source)
{
	return sourcelist_append_source (SOURCELIST (ui->sourcelist), source);
}

void
ui_select_source (UI *ui, int id)
{
	sourcelist_select_source (SOURCELIST (ui->sourcelist), id);
}

void
ui_remove_source (UI *ui, int id)
{
	sourcelist_remove_source (SOURCELIST (ui->sourcelist), id);
}

void
ui_source_set_name (UI *ui, int id, const gchar *name)
{
	sourcelist_source_set_name (SOURCELIST (ui->sourcelist), id, name);
}

void
ui_source_set_priority (UI *ui, int id, int priority)
{
	sourcelist_source_set_priority (SOURCELIST (ui->sourcelist), id, priority);
}

void
ui_source_set_icon (UI *ui, int id, GdkPixbuf *icon)
{
	sourcelist_source_set_icon (SOURCELIST (ui->sourcelist), id, icon);
}

void
ui_source_set_editable (UI *ui, int id, gboolean editable)
{
	sourcelist_source_set_editable (SOURCELIST (ui->sourcelist), id, editable);
}

void
ui_source_start_editing (UI *ui, int id)
{
	sourcelist_source_start_editing (SOURCELIST (ui->sourcelist), id);
}

void
ui_source_set_playing (UI *ui, int id, gboolean playing)
{
	sourcelist_source_set_playing (SOURCELIST (ui->sourcelist), id, playing);
}

int
ui_get_selected_source (UI *ui)
{
	return sourcelist_get_selected_source (SOURCELIST (ui->sourcelist));
}

static void
control_bar_play_clicked_cb (ControlBar *t, ControlBarPlayAction action, UI *ui)
{
	switch (action)
	{
	case CONTROL_BAR_PLAY_ACTION_PAUSE:
		g_signal_emit (G_OBJECT (ui), ui_signals[PAUSE], 0);
		break;
	case CONTROL_BAR_PLAY_ACTION_PLAY:
		g_signal_emit (G_OBJECT (ui), ui_signals[PLAY], 0);
		break;
	case CONTROL_BAR_PLAY_ACTION_STOP:
		g_signal_emit (G_OBJECT (ui), ui_signals[STOP], 0);
		break;
	default:
		break;
	}
}

static void
control_bar_rewind_clicked_cb (ControlBar *t, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[REWIND], 0);
}

static void
control_bar_forward_clicked_cb (ControlBar *t, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[FORWARD], 0);
}

static void
control_bar_search_activate_cb (ControlBar *t, gchar *text, UI *ui)
{
	gboolean val = FALSE;
	g_signal_emit (G_OBJECT (ui), ui_signals[QUERY], 0,
		       text, &val);
}

static void
control_bar_volume_changed_cb (ControlBar *t, gfloat vol, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[VOLUME_CHANGED], 0,
       		       vol);
	gtk_widget_set_sensitive (ui->priv->vol_up,   (vol < 1.0));
	gtk_widget_set_sensitive (ui->priv->vol_down, (vol > 0.0));
	if (ui->priv->muted == TRUE)
	{
		g_signal_handlers_block_by_func (G_OBJECT (ui->priv->mute),
					         G_CALLBACK (window_mute_cb),
					         ui);
		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (ui->priv->mute), FALSE);
		ui->priv->muted = FALSE;
		g_signal_handlers_unblock_by_func (G_OBJECT (ui->priv->mute),
					           G_CALLBACK (window_mute_cb),
					           ui);
	}
}

static void
control_bar_progress_changed_cb (ControlBar *t, gfloat progress, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[PROGRESS_CHANGED], 0,
       		       progress);
}

static void
statusbar_shuffle_clicked_cb (ControlBar *t, gboolean shuffle, UI *ui)
{
	ui_set_shuffle (ui, shuffle);
}

static void
statusbar_repeat_clicked_cb (ControlBar *t, gboolean repeat, UI *ui)
{
	ui_set_repeat (ui, repeat);
}

static void
statusbar_new_playlist_clicked_cb (ControlBar *t, UI *ui)
{
	add_playlist (ui, FALSE);
}

static void
sourcelist_source_selected_cb (ControlBar *t, int id, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[SOURCE_SELECTED], 0,
       		       id);
}

static void
sourcelist_source_deleted_cb (ControlBar *t, int id, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[SOURCE_DELETED], 0,
       		       id);
}

static void
sourcelist_source_edited_cb (ControlBar *t, int id, gchar *text, UI *ui)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[SOURCE_EDITED], 0,
       		       id, text);
}

void
ui_clear_search_entry (UI *ui)
{
	control_bar_clear_search_entry (ui->control_bar);
}

/**
 * ui_set_visible: show ui
 */
void
ui_set_visible (UI *ui, gboolean visible)
{
	if (visible == TRUE)
	{
		gtk_widget_show (ui->priv->window);
		gdk_window_raise (ui->priv->window->window);
	}
	else
	{
		gtk_widget_hide (ui->priv->window);
	}
}

/**
 * ui_get_visible: get visibility
 */
gboolean
ui_get_visible (UI *ui)
{
	return GTK_WIDGET_VISIBLE (ui->priv->window);
}

/**
 * ui_set_can_cut: should Cut menuitem be sensitive ?
 */
void
ui_set_can_cut (UI *ui, gboolean can)
{
	gtk_widget_set_sensitive (ui->priv->cut, can);
}

/**
 * ui_set_can_copy: should Copy menuitem be sensitive ?
 */
void
ui_set_can_copy (UI *ui, gboolean can)
{
	gtk_widget_set_sensitive (ui->priv->copy, can);
}

/**
 * ui_set_can_paste: should Paste menuitem be sensitive ?
 */
void
ui_set_can_paste (UI *ui, gboolean can)
{
	gtk_widget_set_sensitive (ui->priv->paste, can);
}

/**
 * ui_set_shuffle: set the ui to reflect shuffle state
 */
void
ui_set_shuffle (UI *ui, gboolean shuffle)
{
	g_signal_handlers_block_by_func (G_OBJECT (ui->priv->shuffle),
				         G_CALLBACK (window_shuffle_cb),
				         ui);
	g_signal_handlers_block_by_func (G_OBJECT (ui->statusbar),
				         G_CALLBACK (statusbar_shuffle_clicked_cb),
				         ui);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (ui->priv->shuffle), shuffle);
	statusbar_set_shuffle (ui->statusbar, shuffle);
	g_signal_handlers_unblock_by_func (G_OBJECT (ui->priv->shuffle),
				           G_CALLBACK (window_shuffle_cb),
				           ui);
	g_signal_handlers_unblock_by_func (G_OBJECT (ui->statusbar),
				           G_CALLBACK (statusbar_shuffle_clicked_cb),
				           ui);
	g_signal_emit (G_OBJECT (ui), ui_signals[SET_SHUFFLE], 0, shuffle);
}

/**
 * ui_get_shuffle: check whether we are in shuffle mode
 */
gboolean
ui_get_shuffle (UI *ui)
{
	return GTK_CHECK_MENU_ITEM (ui->priv->shuffle)->active;
}

/**
 * ui_set_repeat: set the ui to reflect repeat state
 */
void
ui_set_repeat (UI *ui, gboolean repeat)
{
	g_signal_handlers_block_by_func (G_OBJECT (ui->priv->repeat),
				         G_CALLBACK (window_repeat_cb),
				         ui);
	g_signal_handlers_block_by_func (G_OBJECT (ui->statusbar),
				         G_CALLBACK (statusbar_repeat_clicked_cb),
				         ui);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (ui->priv->repeat), repeat);
	statusbar_set_repeat (ui->statusbar, repeat);
	g_signal_handlers_unblock_by_func (G_OBJECT (ui->priv->repeat),
				           G_CALLBACK (window_repeat_cb),
				           ui);
	g_signal_handlers_unblock_by_func (G_OBJECT (ui->statusbar),
				           G_CALLBACK (statusbar_repeat_clicked_cb),
				           ui);
	g_signal_emit (G_OBJECT (ui), ui_signals[SET_REPEAT], 0, repeat);
}

/**
 * ui_get_repeat: check whether we are in repeat mode
 */
gboolean
ui_get_repeat (UI *ui)
{
	return GTK_CHECK_MENU_ITEM (ui->priv->repeat)->active;
}

/**
 * add_playlist: request a playlist name, and emit a playlist created signal
 */
static void
add_playlist (UI *ui, gboolean sel)
{
	g_signal_emit (G_OBJECT (ui), ui_signals[NEW_PLAYLIST], 0, _("Untitled"), sel);
}

/**
 * ui_play: set the ui to be playing and emit a play signal
 */
void
ui_play (UI *ui)
{
	if (ui->state == UI_STATE_PAUSED ||
	    ui->state == UI_STATE_STOPPED ||
	    ui->state == UI_STATE_PROCESSING)
		g_signal_emit (G_OBJECT (ui), ui_signals[PLAY], 0);
}

/**
 * ui_pause: set the ui to be paused and emit a pause signal
 */
void
ui_pause (UI *ui)
{
	if (ui->state == UI_STATE_PLAYING ||
	    ui->state == UI_STATE_PLAYING_WITH_STOP)
		g_signal_emit (G_OBJECT (ui), ui_signals[PAUSE], 0);
}

/**
 * ui_rewind: goto prev song
 */
void
ui_rewind (UI *ui)
{
	if (GTK_WIDGET_SENSITIVE (ui->priv->rewind))
		g_signal_emit (G_OBJECT (ui), ui_signals[REWIND], 0);
}

/**
 * ui_forward: goto next song
 */
void
ui_forward (UI *ui)
{
	if (GTK_WIDGET_SENSITIVE (ui->priv->forward))
		g_signal_emit (G_OBJECT (ui), ui_signals[FORWARD], 0);
}
