/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __UI_H
#define __UI_H

#include <glib-object.h>
#include <gtk/gtkwidget.h>

#define TYPE_UI             	(ui_get_type ())
#define UI(obj)		        (GTK_CHECK_CAST ((obj), TYPE_UI, UI))
#define UI_CLASS(klass)	        (GTK_CHECK_CLASS_CAST ((klass), TYPE_UI, UIClass))
#define IS_UI(obj)		(GTK_CHECK_TYPE ((obj), TYPE_UI))
#define IS_UI_CLASS(klass)  	(GTK_CHECK_CLASS_TYPE ((klass), TYPE_UI))
#define UI_GET_CLASS(obj)  	(GTK_CHECK_GET_CLASS ((obj), TYPE_UI, UIClass))

typedef struct _UI        	UI;
typedef struct _UIClass 	UIClass;

typedef struct _UIPrivate 	UIPrivate;

typedef enum
{
	UI_STATE_STOPPED,
	UI_STATE_PLAYING,
	UI_STATE_PAUSED,
	UI_STATE_PLAYING_WITH_STOP,
	UI_STATE_PROCESSING
} UIState;

#include "control-bar.h"
#include "statusbar.h"
#include "sourcelist.h"

struct _UI
{
	GObject base;
	
	UIPrivate *priv;

	GtkWidget *notebook;

	UIState state;

	ControlBar *control_bar;
	Statusbar *statusbar;
	Sourcelist *sourcelist;
};

struct _UIClass
{
	GObjectClass parent_class;

	/* play controls */
	void (*play)             (UI *ui);
	void (*pause)            (UI *ui);
	void (*stop)             (UI *ui);
	void (*rewind)           (UI *ui);
	void (*forward)          (UI *ui);

	/* volume signals */
	void (*volume_changed)   (UI *ui, gfloat volume);

	/* seeking signals */
	void (*progress_changed) (UI *ui, gfloat progress);

	/* statusbar functions */
	void (*new_playlist)     (UI *ui, gchar *name, gboolean from_sel);
	void (*set_shuffle)      (UI *ui, gboolean shuffle);
	void (*set_repeat)       (UI *ui, gboolean repeat);

	/* a single song added to the library via a file selector */
	void (*new_song)         (UI *ui, gchar *fn);

	/* query signals */
	void (*query)            (UI *ui, gchar *text);
	
	/* sources view signals */
	void (*source_selected)  (UI *ui, int id);
	void (*source_deleted)   (UI *ui, int id);
	void (*source_edited)    (UI *ui, int id, gchar *text);

	/* menubar signals */
	void (*cut)              (UI *ui);
	void (*copy)             (UI *ui);
	void (*paste)            (UI *ui);
	void (*clear)            (UI *ui);
	void (*select_all)       (UI *ui);
	void (*select_none)      (UI *ui);
};

/* object functions */
GType      ui_get_type            (void);

UI        *ui_new                 (void);

/* playing controls */
void       ui_set_state           (UI *ui, UIState state);
void       ui_set_can_rewind      (UI *ui, gboolean can);
void       ui_set_can_forward     (UI *ui, gboolean can);

/* shuffle/repeat */
void       ui_set_shuffle         (UI *ui, gboolean shuffle);
gboolean   ui_get_shuffle         (UI *ui);
void       ui_set_repeat          (UI *ui, gboolean repeat);
gboolean   ui_get_repeat          (UI *ui);

/* action buttons */
void       ui_merge_widgets       (UI *ui, GList *widgets); 
GtkWidget *ui_new_action_button   (UI *ui,
			           const char *label,
			           const char *tooltip,
			           GdkPixbuf *icon,
			           gboolean toggle);

/* search entry functions */
void       ui_clear_search_entry  (UI *ui);

/* volume functions */
void       ui_set_volume          (UI *ui, float volume);
float      ui_get_volume          (UI *ui);

/* lcd functions */
void       ui_set_offset          (UI *ui, gfloat offset);
void       ui_set_song            (UI *ui, const char *title, const char *artist, const char *uri);
void       ui_set_elapsed_time    (UI *ui, long elapsed, float progress);

/* statusbar functions */
void       ui_set_status          (UI *ui, const gchar *status);

/* sources functions */
int        ui_append_source        (UI *ui, GtkWidget *source);
void       ui_select_source        (UI *ui, int id);
void       ui_remove_source        (UI *ui, int id);
void       ui_source_set_name      (UI *ui, int id, const gchar *name);
void       ui_source_set_priority  (UI *ui, int id, int priority);
void       ui_source_set_icon      (UI *ui, int id, GdkPixbuf *icon);
void       ui_source_set_editable  (UI *ui, int id, gboolean editable);
void       ui_source_start_editing (UI *ui, int id);
void       ui_source_set_playing   (UI *ui, int id, gboolean playing);
int        ui_get_selected_source  (UI *ui);

/* menubar functions */
void       ui_set_can_cut         (UI *ui, gboolean can);
void       ui_set_can_copy        (UI *ui, gboolean can);
void       ui_set_can_paste       (UI *ui, gboolean can);

/* visibility */
void       ui_set_visible         (UI *ui, gboolean visible);
gboolean   ui_get_visible         (UI *ui);

/* remote ui control */
void       ui_play                (UI *ui);
void       ui_pause               (UI *ui);
void       ui_rewind              (UI *ui);
void       ui_forward             (UI *ui);

#endif /* __UI_H */
