/*  
 *  Copyright (C) 2002 James Willcox <jwillcox@cs.indiana.edu>
 *                     Bastien Nocera <hadess@hadess.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: rb-image-button.c,v 1.5 2002/04/01 14:49:23 jbaayen Exp $
 */

#include <gtk/gtk.h>
#include <math.h>

#include "rb-image-button.h"

struct _RBImageButtonPrivate
{
	GdkPixbuf *normal;
	GdkPixbuf *over;
	GdkPixbuf *depressed;
	GdkPixbuf *insensitive;
};


/* GObject properties */
enum {
	PROP_BOGUS,
	PROP_IMAGE_NORMAL,
	PROP_IMAGE_DEPRESSED
};

/* globals */
static GObjectClass *parent_class = NULL;

/* object function prototypes */
static void rb_image_button_class_init (RBImageButtonClass *klass);
static void rb_image_button_init (RBImageButton *s);
static void rb_image_button_finalize (GObject *object);

/* colorshift a pixbuf */
static void
do_colorshift(GdkPixbuf * dest, GdkPixbuf * src, int shift)
{
	gint i, j;
	gint width, height, has_alpha, srcrowstride, destrowstride;
	guchar *target_pixels;
	guchar *original_pixels;
	guchar *pixsrc;
	guchar *pixdest;
	int val;
	guchar r, g, b;

	has_alpha = gdk_pixbuf_get_has_alpha(src);
	width = gdk_pixbuf_get_width(src);
	height = gdk_pixbuf_get_height(src);
	srcrowstride = gdk_pixbuf_get_rowstride(src);
	destrowstride = gdk_pixbuf_get_rowstride(dest);
	target_pixels = gdk_pixbuf_get_pixels(dest);
	original_pixels = gdk_pixbuf_get_pixels(src);

	for (i = 0; i < height; i++) {
		pixdest = target_pixels + i * destrowstride;
		pixsrc = original_pixels + i * srcrowstride;
		for (j = 0; j < width; j++) {
			r = *(pixsrc++);
			g = *(pixsrc++);
			b = *(pixsrc++);
			val = r + shift;
			*(pixdest++) = CLAMP(val, 0, 255);
			val = g + shift;
			*(pixdest++) = CLAMP(val, 0, 255);
			val = b + shift;
			*(pixdest++) = CLAMP(val, 0, 255);
			if (has_alpha)
				*(pixdest++) = *(pixsrc++);
		}
	}
}

static GdkPixbuf *
make_prelight_pixbuf(GdkPixbuf * pb)
{
	GdkPixbuf *new;
	if (!pb)
		return NULL;

	new = gdk_pixbuf_new(gdk_pixbuf_get_colorspace(pb),
			     gdk_pixbuf_get_has_alpha(pb),
			     gdk_pixbuf_get_bits_per_sample(pb),
			     gdk_pixbuf_get_width(pb),
			     gdk_pixbuf_get_height(pb));
	do_colorshift(new, pb, 30);

	return new;
}

static GdkPixbuf *
make_insensitive_pixbuf(GdkPixbuf * pb)
{
	GdkPixbuf *new;
	if (!pb)
		return NULL;

	new = gdk_pixbuf_new(gdk_pixbuf_get_colorspace(pb),
			     gdk_pixbuf_get_has_alpha(pb),
			     gdk_pixbuf_get_bits_per_sample(pb),
			     gdk_pixbuf_get_width(pb),
			     gdk_pixbuf_get_height(pb));
	gdk_pixbuf_saturate_and_pixelate (pb, new, 0.8, TRUE);

	return new;
}

static void
rb_image_button_paint (RBImageButton *button,
			  GdkRectangle *area,
			  GtkStateType state_type)
{
	GdkWindow *window = GTK_WIDGET (button)->window;
	GtkWidget *widget = GTK_WIDGET (button);
	GdkGC *gc = gdk_gc_new (GDK_DRAWABLE (window));
	GdkPixbuf *pixbuf;
	int width;
	int height;
	int x;
	int y;

	if (state_type == GTK_STATE_NORMAL)
		pixbuf = button->priv->normal;
	else if (state_type == GTK_STATE_PRELIGHT)
		pixbuf = button->priv->over;
	else if (state_type == GTK_STATE_ACTIVE)
		pixbuf = button->priv->depressed;
	else if (state_type == GTK_STATE_INSENSITIVE)
		pixbuf = button->priv->insensitive;
	else
		pixbuf = button->priv->normal;

	width = gdk_pixbuf_get_width (pixbuf);
	height = gdk_pixbuf_get_height (pixbuf);

	x = floor (widget->allocation.x
		 + ((widget->allocation.width - widget->requisition.width))
		 + 0.5);
	y = floor (widget->allocation.y 
		 + ((widget->allocation.height - widget->requisition.height))
		 + 0.5);

	gdk_pixbuf_render_to_drawable (pixbuf,
				       GDK_DRAWABLE (window),
				       gc,
				       0, 0, x, y,
				       width, height,
				       GDK_RGB_DITHER_NORMAL,
				       0, 0);
}

static gboolean
rb_image_button_expose (GtkWidget      *widget,
			   GdkEventExpose *event)
{
	if (GTK_WIDGET_DRAWABLE (widget))
	{
		RBImageButton *button = RB_IMAGE_BUTTON (widget);

		rb_image_button_paint (button,
					  &event->area,
					  GTK_WIDGET_STATE (widget));

	}

	return FALSE;
}

static void
rb_image_button_size_request (GtkWidget      *widget,
				 GtkRequisition *requisition)
{
	RBImageButton *button = (RBImageButton *) widget;

	requisition->width = (GTK_CONTAINER (widget)->border_width +
			      GTK_WIDGET (widget)->style->xthickness) * 2;

	requisition->height = (GTK_CONTAINER (widget)->border_width +
			       GTK_WIDGET (widget)->style->ythickness) * 2;

	requisition->width += gdk_pixbuf_get_width (button->priv->normal);
	requisition->height += gdk_pixbuf_get_height (button->priv->normal);
}

#if 0
static void
rb_image_button_size_allocate (GtkWidget     *widget,
				  GtkAllocation *allocation)
{
	/*
	GtkButton *button;
	int border_width = GTK_CONTAINER (widget)->border_width;
	int xthickness = GTK_WIDGET (widget)->style->xthickness;
	int ythickness = GTK_WIDGET (widget)->style->ythickness;
	*/

	widget->allocation = *allocation;

	/*
	if (GTK_WIDGET_REALIZED (widget))
		gdk_window_move_resize (button->event_window,
				widget->allocation.x + border_width,
				widget->allocation.y + border_width,
				widget->allocation.width - border_width * 2,
				widget->allocation.height - border_width * 2);
	*/


}
#endif

static GdkPixbuf *
rb_image_button_get_normal_pixbuf (RBImageButton *button)
{
	return button->priv->normal;
}

static GdkPixbuf *
rb_image_button_get_depressed_pixbuf (RBImageButton *button)
{
	return button->priv->depressed;
}

void
rb_image_button_set_normal_pixbuf (RBImageButton *button, GdkPixbuf *pixbuf)
{
	button->priv->normal = pixbuf;
	button->priv->over = make_prelight_pixbuf (pixbuf);
	button->priv->insensitive = make_insensitive_pixbuf (pixbuf);

	gtk_widget_queue_draw (GTK_WIDGET (button));
}

void
rb_image_button_set_depressed_pixbuf (RBImageButton *button, GdkPixbuf *pixbuf)
{
	button->priv->depressed = pixbuf;

	gtk_widget_queue_draw (GTK_WIDGET (button));
}

static void
rb_image_button_set_property (GObject *object,
			   guint prop_id,
			   const GValue *value,
			   GParamSpec *pspec)
{
	RBImageButton *button = (RBImageButton *)object;
	GdkPixbuf *pixbuf;

	switch (prop_id)
	{
		case PROP_IMAGE_NORMAL:
			pixbuf = g_value_get_object (value);
			rb_image_button_set_normal_pixbuf (button, pixbuf);
		break;
		case PROP_IMAGE_DEPRESSED:
			pixbuf = g_value_get_object (value);
			rb_image_button_set_depressed_pixbuf (button, pixbuf);
		break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
rb_image_button_get_property (GObject *object,
			   guint prop_id,
			   GValue *value,
			   GParamSpec *pspec)
{
	RBImageButton *button = (RBImageButton *)object;

	switch (prop_id)
	{
		case PROP_IMAGE_NORMAL:
			g_value_set_object (value, rb_image_button_get_normal_pixbuf (button));
		break;
		case PROP_IMAGE_DEPRESSED:
			g_value_set_object (value, rb_image_button_get_depressed_pixbuf (button));
		break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

/**
 * rb_image_button_get_type: get the GObject type of the RBImageButton 
 */
GType
rb_image_button_get_type (void)
{
	static GType rb_image_button_type = 0;

  	if (rb_image_button_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (RBImageButtonClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) rb_image_button_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (RBImageButton),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) rb_image_button_init
      		};

      		rb_image_button_type = g_type_register_static (GTK_TYPE_BUTTON,
                				             "RBImageButton",
                                           	             &our_info, 0);
    	}

	return rb_image_button_type;
}

/**
 * rb_image_button_class_init: initialize the RBImageButton class
 */
static void
rb_image_button_class_init (RBImageButtonClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GtkWidgetClass *widget_class = (GtkWidgetClass *) klass;

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = rb_image_button_finalize;
	widget_class->expose_event = rb_image_button_expose;
	widget_class->size_request = rb_image_button_size_request;
	//widget_class->size_allocate = rb_image_button_size_allocate;

	object_class->set_property = rb_image_button_set_property;
	object_class->get_property = rb_image_button_get_property;


	g_object_class_install_property (object_class,
					 PROP_IMAGE_NORMAL,
					 g_param_spec_object ("image-normal",
						 	      "Image normally displayed",
							      "Image normally displayed",
							      gdk_pixbuf_get_type (),
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_IMAGE_DEPRESSED,
					 g_param_spec_object ("image-depressed",
						 	      "Image displayed when depressed",
							      "Image displayed when depressed",
							      gdk_pixbuf_get_type (),
							      G_PARAM_READWRITE));
}

/**
 * rb_image_button_init: intialize the RBImageButton object
 */
static void
rb_image_button_init (RBImageButton *w)
{
	w->priv = g_new0 (RBImageButtonPrivate, 1);
}

/**
 * rb_image_button_finalize: finalize the RBImageButton object
 */
static void
rb_image_button_finalize (GObject *object)
{
	RBImageButton *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_RB_IMAGE_BUTTON (object));
	
   	s = RB_IMAGE_BUTTON (object);

	g_return_if_fail (s->priv != NULL);

	g_free (s->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * rb_image_button_new: create a new RBImageButton object
 */
GtkWidget *
rb_image_button_new (GdkPixbuf *normal, GdkPixbuf *depressed)
{
 	GtkWidget *s;

	s = GTK_WIDGET (g_object_new (TYPE_RB_IMAGE_BUTTON,
				      "image-normal",
				      normal,
				      "image-depressed",
				      depressed, NULL));

	g_return_val_if_fail (RB_IMAGE_BUTTON (s)->priv != NULL, NULL);

	return s;
}
