/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: rbtreemodelfilter.c,v 1.6 2002/04/11 14:18:41 jbaayen Exp $
 */

#include <gtk/gtktreeview.h>
#include <gtk/gtktreemodel.h>
#include <gtk/gtkliststore.h>
#include <libgnome/gnome-i18n.h>
#include <stdlib.h>

#include "rbtreemodelfilter.h"

/* GObject properties */
enum
{
	PROP_0,
	PROP_VISIBLE_ROWS
};

/* globals */
static GObjectClass *parent_class = NULL;

/* object funtion prototypes */
static void rb_tree_model_filter_class_init (RBTreeModelFilterClass *klass);
static void rb_tree_model_filter_init (RBTreeModelFilter *l);
static void rb_tree_model_filter_finalize (GObject *object);
static void rb_tree_model_filter_set_property (GObject *object,
			                       guint prop_id,
			                       const GValue *value,
			                       GParamSpec *pspec);
static void rb_tree_model_filter_get_property (GObject *object,
			                       guint prop_id,
			                       GValue *value,
			                       GParamSpec *pspec);

/* callback prototypes */
static gboolean list_iters (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data);
static void iter_hide (GtkTreeIter *iter, RBTreeModelFilter *f);
static void iter_show (GtkTreeIter *iter, RBTreeModelFilter *f);

/**
 * rb_tree_model_filter_get_type: get the GObject type of the RBTreeModelFilter
 */
GType
rb_tree_model_filter_get_type (void)
{
	static GType rb_tree_model_filter_type = 0;

  	if (rb_tree_model_filter_type == 0)
    	{
      		static const GTypeInfo our_info =
      		{
        		sizeof (RBTreeModelFilterClass),
        		NULL, /* base_init */
        		NULL, /* base_finalize */
        		(GClassInitFunc) rb_tree_model_filter_class_init,
        		NULL, /* class_finalize */
        		NULL, /* class_data */
        		sizeof (RBTreeModelFilter),
        		0,    /* n_preallocs */
        		(GInstanceInitFunc) rb_tree_model_filter_init
      		};

      		rb_tree_model_filter_type = g_type_register_static (GTK_TYPE_TREE_MODEL_FILTER,
                				                    "RBTreeModelFilter",
                                           	                    &our_info, 0);
    	}

	return rb_tree_model_filter_type;
}

/**
 * rb_tree_model_filter_class_init: initialize the RBTreeModelFilter class
 */
static void
rb_tree_model_filter_class_init (RBTreeModelFilterClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

  	parent_class = g_type_class_peek_parent (klass);

  	object_class->finalize = rb_tree_model_filter_finalize;

	object_class->set_property = rb_tree_model_filter_set_property;
	object_class->get_property = rb_tree_model_filter_get_property;

	/* init properties */
	g_object_class_install_property (object_class,
					 PROP_VISIBLE_ROWS,
					 g_param_spec_pointer ("visible-rows",
						 	       "Visible rows",
							       "Visible GtkTreeIters",
							       G_PARAM_WRITABLE));
}

/**
 * rb_tree_model_filter_init: intialize the RBTreeModelFilter object
 */
static void
rb_tree_model_filter_init (RBTreeModelFilter *l)
{
}

/**
 * rb_tree_model_filter_finalize: finalize the RBTreeModelFilter object
 */
static void
rb_tree_model_filter_finalize (GObject *object)
{
	RBTreeModelFilter *s;

	g_return_if_fail (object != NULL);
	g_return_if_fail (RB_IS_TREE_MODEL_FILTER (object));
	
   	s = RB_TREE_MODEL_FILTER (object);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * rb_tree_model_filter_new_with_model: create a new RBTreeModelFilter object
 */
GtkTreeModel *
rb_tree_model_filter_new_with_model (GtkTreeModel *child_model, int filter_column, GtkTreePath *virtual_root)
{
 	RBTreeModelFilter *l;

	l = RB_TREE_MODEL_FILTER (g_object_new (RB_TYPE_TREE_MODEL_FILTER,
						"model", child_model,
						"filter_column", filter_column,
						"virtual_root", virtual_root,
						NULL));

	return GTK_TREE_MODEL (l);
}

/**
 * rb_tree_model_filter_show_iters: show a list of gtktreeiters, hiding the previously
 * visible ones
 */
void
rb_tree_model_filter_show_iters (RBTreeModelFilter *f, GList *iters)
{
	GList *list = NULL;

	/* hide old nodes */
	gtk_tree_model_foreach (GTK_TREE_MODEL (f),
			        (GtkTreeModelForeachFunc) list_iters, (void **) &list);
	g_list_foreach (list, (GFunc) iter_hide, f);
	g_list_free (list);
	
	/* show new nodes */
	g_list_foreach (iters, (GFunc) iter_show, f);
}

/**
 * list_iters: list gtktreeiters
 */
static gboolean
list_iters (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, void **data)
{
	GList **list = (GList **) data;
	GtkTreeIter *child_iter = g_new0 (GtkTreeIter, 1);
	GtkTreePath *child_path;
	GtkTreeModelFilter *gf = GTK_TREE_MODEL_FILTER (model);

	gtk_tree_model_filter_convert_iter_to_child_iter (gf, child_iter, iter);
	child_path = gtk_tree_model_get_path (gf->child_model, child_iter);

	*list = g_list_append (*list, child_iter);
	
	/* go on */
	return FALSE;
}

/**
 * iter_show: show an iter
 */
static void
iter_show (GtkTreeIter *iter, RBTreeModelFilter *f)
{
	GtkTreeModelFilter *gf = GTK_TREE_MODEL_FILTER (f);

	if (iter == NULL) return;
	
	gtk_list_store_set (GTK_LIST_STORE (gf->child_model), iter,
			    gf->filter_column, TRUE, -1);
}

/**
 * iter_hide: show an iter
 */
static void
iter_hide (GtkTreeIter *iter, RBTreeModelFilter *f)
{
	GtkTreeModelFilter *gf = GTK_TREE_MODEL_FILTER (f);

	gtk_list_store_set (GTK_LIST_STORE (gf->child_model), iter,
			    gf->filter_column, FALSE, -1);

	gtk_tree_iter_free (iter);
}

/**
 * rb_tree_model_filter_set_property: set a RBTreeModelFilter property
 */
static void
rb_tree_model_filter_set_property (GObject *object,
			           guint prop_id,
			           const GValue *value,
			           GParamSpec *pspec)
{
	RBTreeModelFilter *f = RB_TREE_MODEL_FILTER (object);

	switch (prop_id)
	{
		case PROP_VISIBLE_ROWS:
			rb_tree_model_filter_show_iters (f, g_value_get_pointer (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/**
 * rb_tree_model_filter_get_property: get a RBTreeModelFilter property
 */
static void
rb_tree_model_filter_get_property (GObject *object,
			           guint prop_id,
			           GValue *value,
			           GParamSpec *pspec)
{
	switch (prop_id)
	{
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}
