/*  RhythmBox
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *                     Marco Pesenti Gritti <marco@it.gnome.org>
 *                     Bastien Nocera <hadess@hadess.net>
 *                     Seth Nickell <snickell@stanford.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: treeview-utils.c,v 1.24 2002/04/13 21:54:29 kristian Exp $
 */

#include <libgnome/gnome-util.h>
#include <libgnome/gnome-i18n.h>
#include <gtk/gtk.h>
#include <string.h>
#include <unistd.h>

#include "misc.h"
#include "treeview-utils.h"
#include "gtktreemodelfilter.h"

/* local function prototypes */
static void free_iter (gpointer key, GtkTreeIter *val, gpointer data);

/**
 * store_path_to_sort_path: convert a GtkTreePath from the root store to a
 * GtkTreePath in the sortmodel
 */
GtkTreePath *
store_path_to_sort_path (GtkTreeModel *filtermodel, GtkTreeModel *sortmodel,
			 GtkTreePath *path)
{
	GtkTreePath *path1, *path2;

	path1 = gtk_tree_model_filter_convert_child_path_to_path
		(GTK_TREE_MODEL_FILTER (filtermodel), path);
	path2 = gtk_tree_model_sort_convert_child_path_to_path
		(GTK_TREE_MODEL_SORT (sortmodel), path1);
	gtk_tree_path_free (path1);

	return path2;
}

/**
 * store_iter_to_sort_iter: convert a GtkTreeIter from the root store to
 * a GtkTreeIter in the sortmodel
 */
void
store_iter_to_sort_iter (GtkTreeModel *filtermodel,
			 GtkTreeModel *sortmodel,
			 GtkTreeIter *converted_iter,
			 GtkTreeIter *iter)
{
	GtkTreeIter temp;

	gtk_tree_model_filter_convert_child_iter_to_iter
		(GTK_TREE_MODEL_FILTER (filtermodel), &temp, iter);
	gtk_tree_model_sort_convert_child_iter_to_iter
		(GTK_TREE_MODEL_SORT (sortmodel), converted_iter, &temp);
}

/**
 * priority_sort_func: treeview sorting function taking care of priorities
 * (assumes the to-be sorted string column is 0)
 */
gint
priority_sort_func (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		    gpointer user_data)
{
	gint priority_column = GPOINTER_TO_INT (user_data);
	GValue aval = {0, }, bval = {0, };
	gint ia, ib;
	const gchar *sa, *sb;
	gboolean ret;

	gtk_tree_model_get_value (model, a, priority_column, &aval);
	ia = g_value_get_int (&aval);
	g_value_unset (&aval);

	gtk_tree_model_get_value (model, b, priority_column, &bval);
	ib = g_value_get_int (&bval);
	g_value_unset (&bval);

	/* check priorities */
	if (ia < ib) return 1;
	if (ia > ib) return -1;

	/* no priorities - do some strcmping */
	/* (assumes 0 is the sort column) */
	gtk_tree_model_get_value (model, a, 0, &aval);
	sa = g_value_get_string (&aval);
	if (sa == NULL) sa = "";
	sa = g_utf8_casefold (sa, g_utf8_strlen (sa, -1));

	gtk_tree_model_get_value (model, b, 0, &bval);
	sb = g_value_get_string (&bval);
	if (sb == NULL) sb = "";
	sb = g_utf8_casefold (sb, g_utf8_strlen (sb, -1));

	ret = g_utf8_collate (sa, sb);

	g_free (sa);
	g_free (sb);

	g_value_unset (&aval);
	g_value_unset (&bval);

	return ret;
}

/**
 * free_node_to_iter_hash: free a Node to GtkTreeIter hashtable
 */
void
free_node_to_iter_hash (GHashTable *hash)
{
	g_hash_table_foreach (hash, (GHFunc) free_iter, NULL);
	g_hash_table_destroy (hash);
}

/**
 * free_iter: free_node_to_iter_hash helper
 */
static void
free_iter (gpointer key, GtkTreeIter *val, gpointer data)
{
	gtk_tree_iter_free (val);
}
